@@ -9,8 +9,11 @@ using System.Reflection;
 using System.Reflection.Emit;
 using System.Diagnostics;
 using System.Threading;
+using System.Threading.Tasks;
 using System.Collections.Generic;
 using System.Linq;
+using System.Net.Sockets;
+using MonoTests.Helpers;
 
 public class TestsBase
 {
@@ -311,6 +314,10 @@ public class Tests : TestsBase, ITest2
 			wait_one ();
 			return 0;
 		}
+		if (args.Length >0 && args [0] == "threadpool-io") {
+			threadpool_io ();
+			return 0;
+		}
 		breakpoints ();
 		single_stepping ();
 		arguments ();
@@ -419,7 +426,11 @@ public class Tests : TestsBase, ITest2
 		ss_step_through ();
 		ss_non_user_code ();
 		ss_recursive (1);
+		ss_recursive2 (1);
+		ss_recursive2 (1);
+		ss_recursive_chaotic ();
 		ss_fp_clobber ();
+		ss_no_frames ();
 	}
 
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
@@ -568,6 +579,92 @@ public class Tests : TestsBase, ITest2
 		ss_recursive (n + 1);
 	}
 
+	// Breakpoint will be placed here
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive2_trap ()
+	{
+	}
+
+	public static void ss_recursive2_at (string s)
+	{
+		// Console.WriteLine (s);
+	}
+
+	// This method is used both for a step over and step out test.
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive2 (int x)
+	{
+		ss_recursive2_at ( "ss_recursive2 in " + x);
+		if (x < 5) {
+			int next = x + 1;
+			ss_recursive2_at ("ss_recursive2 descend " + x);
+			ss_recursive2_trap ();
+			ss_recursive2 (next);
+		}
+		ss_recursive2_at ("ss_recursive2 out " + x);
+	}
+
+	// Breakpoint will be placed here
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive_chaotic_trap ()
+	{
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive_chaotic_at (bool exiting, string at, int n)
+	{
+//		string indent = "";
+//		for (int count = 5 - n; count > 0; count--)
+//			indent += "\t";
+//		Console.WriteLine (indent + (exiting ? "<--" : "-->") + " " + at + " " + n);
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive_chaotic_fizz (int n)
+	{
+		ss_recursive_chaotic_at (false, "fizz", n);
+		if (n > 0) {
+			int next = n - 1;
+			ss_recursive_chaotic_buzz (next);
+			ss_recursive_chaotic_fizzbuzz (next);
+		} else {
+			ss_recursive_chaotic_trap ();
+		}
+		ss_recursive_chaotic_at (true, "fizz", n);
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive_chaotic_buzz (int n)
+	{
+		ss_recursive_chaotic_at (false, "buzz", n);
+		if (n > 0) {
+			int next = n - 1;
+			ss_recursive_chaotic_fizz (next);
+			ss_recursive_chaotic_fizzbuzz (next);
+		}
+		ss_recursive_chaotic_at (true, "buzz", n);
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive_chaotic_fizzbuzz (int n)
+	{
+		ss_recursive_chaotic_at (false, "fizzbuzz", n);
+		if (n > 0) {
+			int next = n - 1;
+			ss_recursive_chaotic_fizz (next);
+			ss_recursive_chaotic_buzz (next);
+			ss_recursive_chaotic_fizzbuzz (next);
+		}
+		ss_recursive_chaotic_at (true, "fizzbuzz", n);
+	}
+
+	// Call a complex tree of recursive calls that has tripped up "step out" in the past.
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_recursive_chaotic ()
+	{
+		ss_recursive_chaotic_fizz (5);
+	}
+
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
 	public static void ss_fp_clobber () {
 		double v = ss_fp_clobber_1 (5.0);
@@ -583,6 +680,25 @@ public class Tests : TestsBase, ITest2
 	public static void ss_fp_clobber_2 (double d) {
 	}
 
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_no_frames () {
+		Action a = ss_no_frames_2;
+		var ar = a.BeginInvoke (null, null);
+		ar.AsyncWaitHandle.WaitOne ();
+		// Avoid waiting every time this runs
+		if (static_i == 56)
+			Thread.Sleep (200);
+		ss_no_frames_3 ();
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_no_frames_2 () {
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void ss_no_frames_3 () {
+	}
+
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
 	public static bool is_even (int i) {
 		return i % 2 == 0;
@@ -733,9 +849,7 @@ public class Tests : TestsBase, ITest2
 	}
 
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
-#if NET_4_5
 	[StateMachine (typeof (int))]
-#endif
 	public static void locals2<T> (string[] args, int arg, T t, ref string rs, ref AStruct astruct) {
 		long i = 42;
 		string s = "AB";
@@ -1088,12 +1202,10 @@ public class Tests : TestsBase, ITest2
 
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
 	public static void unhandled_exception_user () {
-#if NET_4_5
 		System.Threading.Tasks.Task.Factory.StartNew (() => {
 				Throw ();
 			});
 		Thread.Sleep (10000);
-#endif
 	}
 
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
@@ -1200,6 +1312,8 @@ public class Tests : TestsBase, ITest2
 		CrossDomain o = (CrossDomain)domain.CreateInstanceAndUnwrap (
 				   typeof (CrossDomain).Assembly.FullName, "CrossDomain");
 
+		domains_print_across (o);
+
 		domains_2 (o, new CrossDomain ());
 
 		o.invoke_2 ();
@@ -1208,6 +1322,8 @@ public class Tests : TestsBase, ITest2
 
 		o.invoke_2 ();
 
+		o.assembly_load ();
+
 		AppDomain.Unload (domain);
 
 		domains_3 ();
@@ -1223,6 +1339,10 @@ public class Tests : TestsBase, ITest2
 	public static void domains_2 (object o, object o2) {
 	}
 
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void domains_print_across (object o) {
+	}
+
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
 	public static void domains_3 () {
 	}
@@ -1235,6 +1355,11 @@ public class Tests : TestsBase, ITest2
 	public static void invoke_in_domain_2 () {
 	}
 
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void assembly_load_in_domain () {
+		Assembly.Load ("System.Transactions");
+	}
+
 	[MethodImplAttribute (MethodImplOptions.NoInlining)]
 	public static void dynamic_methods () {
 		var m = new DynamicMethod ("dyn_method", typeof (void), new Type []  { typeof (int) }, typeof (object).Module);
@@ -1446,6 +1571,49 @@ public class Tests : TestsBase, ITest2
 	public override string virtual_method () {
 		return "V2";
 	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void threadpool_bp () { }
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static void threadpool_io () {
+		// Start a threadpool task that blocks on I/O.
+		// Regression test for #42625
+		const int nbytes = 16;
+		var bsOut = new byte[nbytes];
+		for (int i = 0; i < nbytes; i++) {
+			bsOut[i] = (byte)i;
+		}
+		var endPoint = NetworkHelpers.LocalEphemeralEndPoint ();
+		var l = new TcpListener (endPoint);
+		l.Start ();
+		Task<byte[]> t = Task.Run (async () => {
+			var c = new TcpClient ();
+			await c.ConnectAsync (endPoint.Address, endPoint.Port);
+			var streamIn = c.GetStream ();
+			var bs = new byte[nbytes];
+			int nread = 0;
+			int nremain = nbytes;
+			while (nread < nbytes) {
+				int r = await streamIn.ReadAsync (bs, nread, nremain);
+				nread += r;
+				nremain -= r;
+			}
+			streamIn.Close ();
+			return bs;
+			});
+		var s = l.AcceptTcpClient ();
+		l.Stop ();
+		// write bytes in two groups so that the task blocks on the ReadAsync
+		var streamOut = s.GetStream ();
+		var nbytesFirst = nbytes / 2;
+		var nbytesRest = nbytes - nbytesFirst;
+		streamOut.Write (bsOut, 0, nbytesFirst);
+		threadpool_bp ();
+		streamOut.Write (bsOut, nbytesFirst, nbytesRest);
+		streamOut.Close ();
+		var bsIn = t.Result;
+	}
 }
 
 class TypeLoadClass {
@@ -1454,8 +1622,13 @@ class TypeLoadClass {
 class TypeLoadClass2 {
 }
 
+public class SentinelClass : MarshalByRefObject {
+}
+
 public class CrossDomain : MarshalByRefObject
 {
+	SentinelClass printMe = new SentinelClass ();
+
 	public void invoke () {
 		Tests.invoke_in_domain ();
 	}
@@ -1467,6 +1640,10 @@ public class CrossDomain : MarshalByRefObject
 	public int invoke_3 () {
 		return 42;
 	}
+
+	public void assembly_load () {
+		Tests.assembly_load_in_domain ();
+	}
 }	
 
 public class Foo
