@@ -163,6 +163,13 @@ namespace System.Data.SqlClient
 			}
 		}
 	
+		[MonoTODO]
+		public Guid ClientConnectionId {
+			get {
+				throw new NotImplementedException ();
+			}
+		}
+
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public
 		override
@@ -254,7 +261,7 @@ namespace System.Data.SqlClient
 			set { statisticsEnabled = value; }
 		}
 
-		protected internal override DbProviderFactory DbProviderFactory {
+		protected override DbProviderFactory DbProviderFactory {
 			get {
 				return SqlClientFactory.Instance;
 			}
@@ -419,7 +426,10 @@ namespace System.Data.SqlClient
 		
 		private SqlInfoMessageEventArgs CreateSqlInfoMessageEvent (TdsInternalErrorCollection errors)
 		{
-			return new SqlInfoMessageEventArgs (errors);
+			foreach (TdsInternalError e in errors)
+				return new SqlInfoMessageEventArgs (new SqlException (e.Class, e.LineNumber, e.Message, e.Number, e.Procedure, e.Server, "Mono SqlClient Data Provider", e.State));
+
+			return null;
 		}
 
 		private StateChangeEventArgs CreateStateChangeEvent (ConnectionState originalState, ConnectionState currentState)
@@ -918,7 +928,11 @@ namespace System.Data.SqlClient
 
 				if (Client.Available <= 0)
 					return -1; // Error
-				IPEndPoint endpoint = new IPEndPoint (Dns.GetHostEntry ("localhost").AddressList [0], 0);
+
+				IPEndPoint endpoint = CreateLocalEndpoint ();
+				if (endpoint == null)
+					return -1;
+
 				Byte [] rawrs;
 
 				rawrs = Receive (ref endpoint);
@@ -943,6 +957,16 @@ namespace System.Data.SqlClient
 
 				return SqlServerTcpPort;
 			}
+
+			IPEndPoint CreateLocalEndpoint ()
+			{
+				foreach (var addr in Dns.GetHostEntry ("localhost").AddressList) {
+					if (addr.AddressFamily == Client.AddressFamily)
+						return new IPEndPoint (addr, 0);
+				}
+
+				return null;
+			}
 		}
 
 		struct ColumnInfo
@@ -1660,6 +1684,17 @@ namespace System.Data.SqlClient
 			}
 		}
 
+		[MonoTODO]
+		public void ResetStatistics ()
+		{
+			throw new NotImplementedException ();
+		}
+
+		[MonoTODO]
+		public IDictionary RetrieveStatistics ()
+		{
+			throw new NotImplementedException ();
+		}
 
 		#endregion // Methods
 
