@@ -3,9 +3,6 @@
 #if MONO_SECURITY_ALIAS
 extern alias MonoSecurity;
 #endif
-#if MONO_X509_ALIAS
-extern alias PrebuiltSystem;
-#endif
 
 #if MONO_SECURITY_ALIAS
 using MonoSecurity::Mono.Security.Interface;
@@ -16,13 +13,6 @@ using Mono.Security.Interface;
 using MSX = Mono.Security.X509;
 using Mono.Security.X509.Extensions;
 #endif
-#if MONO_X509_ALIAS
-using XX509CertificateCollection = PrebuiltSystem::System.Security.Cryptography.X509Certificates.X509CertificateCollection;
-using XX509Chain = PrebuiltSystem::System.Security.Cryptography.X509Certificates.X509Chain;
-#else
-using XX509CertificateCollection = System.Security.Cryptography.X509Certificates.X509CertificateCollection;
-using XX509Chain = System.Security.Cryptography.X509Certificates.X509Chain;
-#endif
 
 using System;
 using System.Net;
@@ -70,7 +60,7 @@ namespace Mono.Net.Security
 #endif
 		}
 
-		public static X509Chain CreateX509Chain (XX509CertificateCollection certs)
+		public static X509Chain CreateX509Chain (X509CertificateCollection certs)
 		{
 			var chain = new X509Chain ();
 			chain.ChainPolicy = new X509ChainPolicy ((X509CertificateCollection)(object)certs);
@@ -82,7 +72,7 @@ namespace Mono.Net.Security
 			return chain;
 		}
 
-		static bool BuildX509Chain (XX509CertificateCollection certs, X509Chain chain, ref SslPolicyErrors errors, ref int status11)
+		static bool BuildX509Chain (X509CertificateCollection certs, X509Chain chain, ref SslPolicyErrors errors, ref int status11)
 		{
 #if MOBILE
 			return false;
@@ -114,7 +104,7 @@ namespace Mono.Net.Security
 #endif
 		}
 
-		static bool CheckUsage (XX509CertificateCollection certs, string host, ref SslPolicyErrors errors, ref int status11)
+		static bool CheckUsage (X509CertificateCollection certs, string host, ref SslPolicyErrors errors, ref int status11)
 		{
 #if !MONOTOUCH
 			var leaf = certs[0] as X509Certificate2;
@@ -138,7 +128,7 @@ namespace Mono.Net.Security
 			return true;
 		}
 
-		static bool EvaluateSystem (XX509CertificateCollection certs, XX509CertificateCollection anchors, string host, X509Chain chain, ref SslPolicyErrors errors, ref int status11)
+		static bool EvaluateSystem (X509CertificateCollection certs, X509CertificateCollection anchors, string host, X509Chain chain, ref SslPolicyErrors errors, ref int status11)
 		{
 			var leaf = certs [0];
 			bool result;
@@ -199,7 +189,7 @@ namespace Mono.Net.Security
 		}
 
 		public static bool Evaluate (
-			MonoTlsSettings settings, string host, XX509CertificateCollection certs,
+			MonoTlsSettings settings, string host, X509CertificateCollection certs,
 			X509Chain chain, ref SslPolicyErrors errors, ref int status11)
 		{
 			if (!CheckUsage (certs, host, ref errors, ref status11))
