@@ -33,10 +33,11 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 
-#if !NET_2_1
+#if !MOBILE
 
 using System;
 using System.Collections;
+using System.Collections.Generic;
 using System.IO;
 using System.Runtime.InteropServices;
 using System.Security;
@@ -240,48 +241,66 @@ namespace Microsoft.Win32
 		public void SetValue (RegistryKey rkey, string name, object value, RegistryValueKind valueKind)
 		{
 			Type type = value.GetType ();
-			int result;
 			IntPtr handle = GetHandle (rkey);
 
-			if (valueKind == RegistryValueKind.QWord && type == typeof (long)) {
-				long rawValue = (long)value;
-				result = RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.QWord, ref rawValue, Int64ByteSize); 
-			} else if (valueKind == RegistryValueKind.DWord && type == typeof (int)) {
-				int rawValue = (int)value;
-				result = RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.DWord, ref rawValue, Int32ByteSize); 
-			} else if (valueKind == RegistryValueKind.Binary && type == typeof (byte[])) {
-				byte[] rawValue = (byte[]) value;
-				result = RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.Binary, rawValue, rawValue.Length);
-			} else if (valueKind == RegistryValueKind.MultiString && type == typeof (string[])) {
-				string[] vals = (string[]) value;
-				StringBuilder fullStringValue = new StringBuilder ();
-				foreach (string v in vals)
-				{
-					fullStringValue.Append (v);
-					fullStringValue.Append ('\0');
+			switch (valueKind) {
+			case RegistryValueKind.QWord:
+				try {
+					long rawValue = Convert.ToInt64 (value);
+					CheckResult (RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.QWord, ref rawValue, Int64ByteSize));
+					return;
+				} catch (OverflowException) {
 				}
-				fullStringValue.Append ('\0');
+				break;
+			case RegistryValueKind.DWord:
+				try {
+					int rawValue = Convert.ToInt32 (value);
+					CheckResult (RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.DWord, ref rawValue, Int32ByteSize));
+					return;
+				} catch (OverflowException) {
+				}
+				break;
+			case RegistryValueKind.Binary:
+				if (type == typeof (byte[])) {
+					byte[] rawValue = (byte[]) value;
+					CheckResult (RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.Binary, rawValue, rawValue.Length));
+					return;
+				}
+				break;
+			case RegistryValueKind.MultiString:
+				if (type == typeof (string[])) {
+					string[] vals = (string[]) value;
+					StringBuilder fullStringValue = new StringBuilder ();
+					foreach (string v in vals)
+					{
+						fullStringValue.Append (v);
+						fullStringValue.Append ('\0');
+					}
+					fullStringValue.Append ('\0');
 
-				byte[] rawValue = Encoding.Unicode.GetBytes (fullStringValue.ToString ());
+					byte[] rawValue = Encoding.Unicode.GetBytes (fullStringValue.ToString ());
 			
-				result = RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.MultiString, rawValue, rawValue.Length); 
-			} else if ((valueKind == RegistryValueKind.String || valueKind == RegistryValueKind.ExpandString) &&
-				   type == typeof (string)){
-				string rawValue = String.Format ("{0}{1}", value, '\0');
-				result = RegSetValueEx (handle, name, IntPtr.Zero, valueKind, rawValue,
-							rawValue.Length * NativeBytesPerCharacter);
-				
-			} else if (type.IsArray) {
-				throw new ArgumentException ("Only string and byte arrays can written as registry values");
-			} else {
-				throw new ArgumentException ("Type does not match the valueKind");
+					CheckResult (RegSetValueEx (handle, name, IntPtr.Zero, RegistryValueKind.MultiString, rawValue, rawValue.Length));
+					return;
+				}
+				break;
+			case RegistryValueKind.String:
+			case RegistryValueKind.ExpandString:
+				if (type == typeof (string)) {
+					string rawValue = String.Format ("{0}{1}", value, '\0');
+					CheckResult (RegSetValueEx (handle, name, IntPtr.Zero, valueKind, rawValue,
+								rawValue.Length * NativeBytesPerCharacter));
+					return;
+				}
+				break;
+			default:
+				if (type.IsArray) {
+					throw new ArgumentException ("Only string and byte arrays can written as registry values");
+				}
+				break;
 			}
 
-			// handle the result codes
-			if (result != Win32ResultCode.Success)
-			{
-				GenerateException (result);
-			}
+			throw new ArgumentException ("Type does not match the valueKind");
 		}
 	
 		public void SetValue (RegistryKey rkey, string name, object value)
@@ -317,9 +336,6 @@ namespace Microsoft.Win32
 							rawValue.Length * NativeBytesPerCharacter);
 			}
 
-			if (result == Win32ResultCode.MarkedForDeletion)
-				throw RegistryKey.CreateMarkedForDeletionException ();
-
 			// handle the result codes
 			if (result != Win32ResultCode.Success)
 			{
@@ -355,9 +371,6 @@ namespace Microsoft.Win32
 				int result = RegEnumKey (handle, index, stringBuffer,
 					stringBuffer.Capacity);
 
-				if (result == Win32ResultCode.MarkedForDeletion)
-					throw RegistryKey.CreateMarkedForDeletionException ();
-
 				if (result == Win32ResultCode.Success)
 					continue;
 				
@@ -385,9 +398,6 @@ namespace Microsoft.Win32
 						       IntPtr.Zero, ref type, 
 						       IntPtr.Zero, IntPtr.Zero);
 
-				if (result == Win32ResultCode.MarkedForDeletion)
-					throw RegistryKey.CreateMarkedForDeletionException ();
-
 				if (result == Win32ResultCode.Success || result == Win32ResultCode.MoreData)
 					continue;
 				
@@ -442,19 +452,16 @@ namespace Microsoft.Win32
 		{
 			if (!IsHandleValid (rkey))
 				return;
-#if NET_4_0
 			SafeRegistryHandle safe_handle = rkey.Handle;
 			if (safe_handle != null) {
 				// closes the unmanaged pointer for us.
 				safe_handle.Close ();
 				return;
 			}
-#endif
 			IntPtr handle = GetHandle (rkey);
 			RegCloseKey (handle);
 		}
 
-#if NET_4_0
 		public RegistryKey FromHandle (SafeRegistryHandle handle)
 		{
 			// At this point we can't tell whether the key is writable
@@ -462,7 +469,6 @@ namespace Microsoft.Win32
 			// .Net seems to do.
 			return new RegistryKey (handle.DangerousGetHandle (), String.Empty, true);
 		}
-#endif
 
 		public RegistryKey CreateSubKey (RegistryKey rkey, string keyName)
 		{
@@ -473,9 +479,6 @@ namespace Microsoft.Win32
 				RegOptionsNonVolatile,
 				OpenRegKeyRead | OpenRegKeyWrite, IntPtr.Zero, out subKeyHandle, out disposition);
 
-			if (result == Win32ResultCode.MarkedForDeletion)
-				throw RegistryKey.CreateMarkedForDeletionException ();
-
 			if (result != Win32ResultCode.Success) {
 				GenerateException (result);
 			}
@@ -484,7 +487,6 @@ namespace Microsoft.Win32
 				true);
 		}
 
-#if NET_4_0
 		public RegistryKey CreateSubKey (RegistryKey rkey, string keyName, RegistryOptions options)
 		{
 			IntPtr handle = GetHandle (rkey);
@@ -494,16 +496,12 @@ namespace Microsoft.Win32
 				options == RegistryOptions.Volatile ? RegOptionsVolatile : RegOptionsNonVolatile,
 				OpenRegKeyRead | OpenRegKeyWrite, IntPtr.Zero, out subKeyHandle, out disposition);
 
-			if (result == Win32ResultCode.MarkedForDeletion)
-				throw RegistryKey.CreateMarkedForDeletionException ();
-
 			if (result != Win32ResultCode.Success)
 				GenerateException (result);
 			
 			return new RegistryKey (subKeyHandle, CombineName (rkey, keyName),
 				true);
 		}
-#endif
 
 		public void DeleteKey (RegistryKey rkey, string keyName, bool shouldThrowWhenKeyMissing)
 		{
@@ -542,7 +540,7 @@ namespace Microsoft.Win32
 		{
 			IntPtr handle = GetHandle (rkey);
 			StringBuilder buffer = new StringBuilder (BufferMaxLength);
-			ArrayList keys = new ArrayList ();
+			var keys = new List<string> ();
 				
 			for (int index = 0; true; index ++) {
 				int result = RegEnumKey (handle, index, buffer, buffer.Capacity);
@@ -559,14 +557,14 @@ namespace Microsoft.Win32
 				// should not be here!
 				GenerateException (result);
 			}
-			return (string []) keys.ToArray (typeof(String));
+			return keys.ToArray ();
 		}
 
 
 		public string [] GetValueNames (RegistryKey rkey)
 		{
 			IntPtr handle = GetHandle (rkey);
-			ArrayList values = new ArrayList ();
+			var values = new List<string> ();
 			
 			for (int index = 0; true; index ++)
 			{
@@ -585,13 +583,17 @@ namespace Microsoft.Win32
 				if (result == Win32ResultCode.NoMoreEntries)
 					break;
 
-				if (result == Win32ResultCode.MarkedForDeletion)
-					throw RegistryKey.CreateMarkedForDeletionException ();
-
 				GenerateException (result);
 			}
 
-			return (string []) values.ToArray (typeof(String));
+			return values.ToArray ();
+		}
+
+		private void CheckResult (int result)
+		{
+			if (result != Win32ResultCode.Success) {
+				GenerateException (result);
+			}
 		}
 
 		/// <summary>
@@ -609,6 +611,10 @@ namespace Microsoft.Win32
 					throw new IOException ("The network path was not found.");
 				case Win32ResultCode.InvalidHandle:
 					throw new IOException ("Invalid handle.");
+				case Win32ResultCode.MarkedForDeletion:
+					throw RegistryKey.CreateMarkedForDeletionException ();
+				case Win32ResultCode.ChildMustBeVolatile:
+					throw new IOException ("Cannot create a stable subkey under a volatile parent key.");
 				default:
 					// unidentified system exception
 					throw new SystemException ();
@@ -631,5 +637,5 @@ namespace Microsoft.Win32
 	}
 }
 
-#endif // NET_2_1
+#endif // MOBILE
 
