@@ -41,14 +41,78 @@ using System.Runtime.Serialization;
 
 namespace System
 {
+	// Contains information about the type which is expensive to compute
+	[StructLayout (LayoutKind.Sequential)]
+	internal class MonoTypeInfo {
+		// this is the displayed form: special characters
+		// ,+*&*[]\ in the identifier portions of the names
+		// have been escaped with a leading backslash (\)
+		public string full_name;
+		public MonoCMethod default_ctor;
+	}
+
+	[StructLayout (LayoutKind.Sequential)]
 	partial class RuntimeType
 	{
+		[NonSerialized]
+		MonoTypeInfo type_info;
+
 		internal Object GenericCache;
 
-		internal virtual MonoCMethod GetDefaultConstructor ()
+		internal RuntimeType (Object obj)
 		{
-			// TODO: Requires MonoType
-			throw new NotSupportedException ();
+			throw new NotImplementedException ();
+		}
+
+		internal MonoCMethod GetDefaultConstructor ()
+		{
+			MonoCMethod ctor = null;
+
+			if (type_info == null)
+				type_info = new MonoTypeInfo ();
+			else
+				ctor = type_info.default_ctor;
+
+			if (ctor == null) {
+				var ctors = GetConstructors (BindingFlags.Public | BindingFlags.Instance | BindingFlags.NonPublic | BindingFlags.DeclaredOnly);
+
+				for (int i = 0; i < ctors.Length; ++i) {
+					if (ctors [i].GetParametersCount () == 0) {
+						type_info.default_ctor = ctor = (MonoCMethod) ctors [i];
+						break;
+					}
+				}
+			}
+
+			return ctor;
+		}
+
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		extern MethodInfo GetCorrespondingInflatedMethod (MethodInfo generic);
+
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		extern ConstructorInfo GetCorrespondingInflatedConstructor (ConstructorInfo generic);
+
+		internal override MethodInfo GetMethod (MethodInfo fromNoninstanciated)
+                {
+			if (fromNoninstanciated == null)
+				throw new ArgumentNullException ("fromNoninstanciated");
+                        return GetCorrespondingInflatedMethod (fromNoninstanciated);
+                }
+
+		internal override ConstructorInfo GetConstructor (ConstructorInfo fromNoninstanciated)
+		{
+			if (fromNoninstanciated == null)
+				throw new ArgumentNullException ("fromNoninstanciated");
+                        return GetCorrespondingInflatedConstructor (fromNoninstanciated);
+		}
+
+		internal override FieldInfo GetField (FieldInfo fromNoninstanciated)
+		{
+			/* create sensible flags from given FieldInfo */
+			BindingFlags flags = fromNoninstanciated.IsStatic ? BindingFlags.Static : BindingFlags.Instance;
+			flags |= fromNoninstanciated.IsPublic ? BindingFlags.Public : BindingFlags.NonPublic;
+			return GetField (fromNoninstanciated.Name, flags);
 		}
 
 		string GetDefaultMemberName ()
@@ -388,7 +452,8 @@ namespace System
 				throw new InvalidOperationException(Environment.GetResourceString("Arg_NotGenericParameter"));
 			Contract.EndContractBlock();
 
-			Type[] constraints = GetGenericParameterConstraints_impl ();
+			var paramInfo = new Mono.RuntimeGenericParamInfoHandle (RuntimeTypeHandle.GetGenericParameterInfo (this));
+			Type[] constraints = paramInfo.Constraints;
 
 			if (constraints == null)
 				constraints = EmptyArray<Type>.Value;
@@ -407,13 +472,57 @@ namespace System
 		static extern Type MakeGenericType (Type gt, Type [] types);
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		internal extern RuntimeMethodInfo[] GetMethodsByName (string name, BindingFlags bindingAttr, bool ignoreCase, Type reflected_type);
+		internal extern IntPtr GetMethodsByName_native (IntPtr namePtr, BindingFlags bindingAttr, bool ignoreCase);
+
+		internal RuntimeMethodInfo[] GetMethodsByName (string name, BindingFlags bindingAttr, bool ignoreCase, RuntimeType reflectedType)
+		{
+			var refh = new RuntimeTypeHandle (reflectedType);
+			using (var namePtr = new Mono.SafeStringMarshal (name))
+			using (var h = new Mono.SafeGPtrArrayHandle (GetMethodsByName_native (namePtr.Value, bindingAttr, ignoreCase))) {
+				var n = h.Length;
+				var a = new RuntimeMethodInfo [n];
+				for (int i = 0; i < n; i++) {
+					var mh = new RuntimeMethodHandle (h[i]);
+					a[i] = (RuntimeMethodInfo) MethodBase.GetMethodFromHandleNoGenericCheck (mh, refh);
+				}
+				return a;
+			}
+		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern RuntimePropertyInfo[] GetPropertiesByName (string name, BindingFlags bindingAttr, bool icase, Type reflected_type);		
+		extern IntPtr GetPropertiesByName_native (IntPtr name, BindingFlags bindingAttr, bool icase);		
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern RuntimeConstructorInfo[] GetConstructors_internal (BindingFlags bindingAttr, Type reflected_type);
+		extern IntPtr GetConstructors_native (BindingFlags bindingAttr);
+
+		RuntimeConstructorInfo[] GetConstructors_internal (BindingFlags bindingAttr, RuntimeType reflectedType)
+		{
+			var refh = new RuntimeTypeHandle (reflectedType);
+			using (var h = new Mono.SafeGPtrArrayHandle (GetConstructors_native (bindingAttr))) {
+				var n = h.Length;
+				var a = new RuntimeConstructorInfo [n];
+				for (int i = 0; i < n; i++) {
+					var mh = new RuntimeMethodHandle (h[i]);
+					a[i] = (RuntimeConstructorInfo) MethodBase.GetMethodFromHandleNoGenericCheck (mh, refh);
+				}
+				return a;
+			}
+		}
+
+		RuntimePropertyInfo[] GetPropertiesByName (string name, BindingFlags bindingAttr, bool icase, RuntimeType reflectedType)
+		{
+			var refh = new RuntimeTypeHandle (reflectedType);
+			using (var namePtr = new Mono.SafeStringMarshal (name))
+			using (var h = new Mono.SafeGPtrArrayHandle (GetPropertiesByName_native (namePtr.Value, bindingAttr, icase))) {
+				var n = h.Length;
+				var a = new RuntimePropertyInfo [n];
+				for (int i = 0; i < n; i++) {
+					var ph = new Mono.RuntimePropertyHandle (h[i]);
+					a[i] = (RuntimePropertyInfo) PropertyInfo.GetPropertyFromHandle (ph, refh);
+				}
+				return a;
+			}
+		}
 
 		public override InterfaceMapping GetInterfaceMap (Type ifaceType)
 		{
@@ -554,26 +663,71 @@ namespace System
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		extern Type[] GetGenericArgumentsInternal (bool runtimeArray);
 
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern GenericParameterAttributes GetGenericParameterAttributes ();
-
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern Type[] GetGenericParameterConstraints_impl ();
+		GenericParameterAttributes GetGenericParameterAttributes () {
+			return (new Mono.RuntimeGenericParamInfoHandle (RuntimeTypeHandle.GetGenericParameterInfo (this))).Attributes;
+		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		extern int GetGenericParameterPosition ();
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern RuntimeEventInfo[] GetEvents_internal (string name, BindingFlags bindingAttr, Type reflected_type);
+		extern IntPtr GetEvents_native (IntPtr name, BindingFlags bindingAttr);
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern RuntimeFieldInfo[] GetFields_internal (string name, BindingFlags bindingAttr, Type reflected_type);
+		extern IntPtr GetFields_native (IntPtr name, BindingFlags bindingAttr);
+
+		RuntimeFieldInfo[] GetFields_internal (string name, BindingFlags bindingAttr, RuntimeType reflectedType)
+		{
+			var refh = new RuntimeTypeHandle (reflectedType);
+			using (var namePtr = new Mono.SafeStringMarshal (name))
+			using (var h = new Mono.SafeGPtrArrayHandle (GetFields_native (namePtr.Value, bindingAttr))) {
+				int n = h.Length;
+				var a = new RuntimeFieldInfo[n];
+				for (int i = 0; i < n; i++) {
+					var fh = new RuntimeFieldHandle (h[i]);
+					a[i] = (RuntimeFieldInfo) FieldInfo.GetFieldFromHandle (fh, refh);
+				}
+				return a;
+			}
+		}
+
+		RuntimeEventInfo[] GetEvents_internal (string name, BindingFlags bindingAttr, RuntimeType reflectedType)
+		{
+			var refh = new RuntimeTypeHandle (reflectedType);
+			using (var namePtr = new Mono.SafeStringMarshal (name))
+			using (var h = new Mono.SafeGPtrArrayHandle (GetEvents_native (namePtr.Value, bindingAttr))) {
+				int n = h.Length;
+				var a = new RuntimeEventInfo[n];
+				for (int i = 0; i < n; i++) {
+					var eh = new Mono.RuntimeEventHandle (h[i]);
+					a[i] = (RuntimeEventInfo) EventInfo.GetEventFromHandle (eh, refh);
+				}
+				return a;
+			}
+		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern override Type[] GetInterfaces();
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern RuntimeType[] GetNestedTypes_internal (string name, BindingFlags bindingAttr);		
+		extern IntPtr GetNestedTypes_native (IntPtr name, BindingFlags bindingAttr);
+
+		RuntimeType[] GetNestedTypes_internal (string displayName, BindingFlags bindingAttr)
+		{
+			string internalName = null;
+			if (displayName != null)
+				internalName = TypeIdentifiers.FromDisplay (displayName).InternalName;
+			using (var namePtr = new Mono.SafeStringMarshal (internalName))
+			using (var h = new Mono.SafeGPtrArrayHandle (GetNestedTypes_native (namePtr.Value, bindingAttr))) {
+				int n = h.Length;
+				var a = new RuntimeType [n];
+				for (int i = 0; i < n; i++) {
+					var th = new RuntimeTypeHandle (h[i]);
+					a[i] = (RuntimeType) Type.GetTypeFromHandle (th);
+				}
+				return a;
+			}
+		}
 
 		public override string AssemblyQualifiedName {
 			get {
@@ -586,12 +740,6 @@ namespace System
 			get;
 		}
 
-		public override string FullName {
-			get {
-				throw new NotImplementedException ();
-			}
-		}
-
 		public extern override string Name {
 			[MethodImplAttribute(MethodImplOptions.InternalCall)]
 			get;
@@ -602,6 +750,12 @@ namespace System
 			get;
 		}
 
+#if MOBILE
+		static int get_core_clr_security_level ()
+		{
+			return 1;
+		}
+#else
 		//seclevel { transparent = 0, safe-critical = 1, critical = 2}
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern int get_core_clr_security_level ();
@@ -616,6 +770,34 @@ namespace System
 
 		public override bool IsSecuritySafeCritical {
 			get { return get_core_clr_security_level () == 1; }
-		}		
+		}
+#endif
+
+		public override int GetHashCode()
+		{
+			Type t = UnderlyingSystemType;
+			if (t != null && t != this)
+				return t.GetHashCode ();
+			return (int)_impl.Value;
+		}
+
+		public override string FullName {
+			get {
+				string fullName;
+				// This doesn't need locking
+				if (type_info == null)
+					type_info = new MonoTypeInfo ();
+				if ((fullName = type_info.full_name) == null)
+					fullName = type_info.full_name = getFullName (true, false);
+
+				return fullName;
+			}
+		}
+
+		internal override bool IsUserType {
+			get {
+				return false;
+			}
+		}
 	}
 }
