@@ -62,7 +62,7 @@ namespace System.IO {
 
 			FullPath = Path.GetFullPath (path);
 			if (simpleOriginalPath)
-				OriginalPath = Path.GetFileName (path);
+				OriginalPath = Path.GetFileName (FullPath);
 			else
 				OriginalPath = path;
 
@@ -102,12 +102,13 @@ namespace System.IO {
 
 		public override bool Exists {
 			get {
-				Refresh (false);
+				if (_dataInitialised == -1)
+					Refresh ();
 
-				if (stat.Attributes == MonoIO.InvalidFileAttributes)
+				if (_data.fileAttributes == MonoIO.InvalidFileAttributes)
 					return false;
 
-				if ((stat.Attributes & FileAttributes.Directory) == 0)
+				if ((_data.fileAttributes & FileAttributes.Directory) == 0)
 					return false;
 
 				return true;
@@ -426,20 +427,20 @@ namespace System.IO {
 		static internal IEnumerable<FileSystemInfo> EnumerateFileSystemInfos (string full, string searchPattern, SearchOption searchOption)
 		{
 			string path_with_pattern = Path.Combine (full, searchPattern);
-			IntPtr handle;
+			IntPtr handle = IntPtr.Zero;
 			MonoIOError error;
 			FileAttributes rattr;
 			bool subdirs = searchOption == SearchOption.AllDirectories;
 
 			Path.Validate (full);
 			
-			string s = MonoIO.FindFirst (full, path_with_pattern, out rattr, out error, out handle);
-			if (s == null)
-				yield break;
-			if (error != 0)
-				throw MonoIO.GetException (Path.GetDirectoryName (path_with_pattern), (MonoIOError) error);
-
 			try {
+				string s = MonoIO.FindFirst (full, path_with_pattern, out rattr, out error, out handle);
+				if (s == null)
+					yield break;
+				if (error != 0)
+					throw MonoIO.GetException (Path.GetDirectoryName (path_with_pattern), (MonoIOError) error);
+
 				do {
 					if (((rattr & FileAttributes.ReparsePoint) == 0)){
 						if ((rattr & FileAttributes.Directory) != 0)
@@ -454,10 +455,24 @@ namespace System.IO {
 
 				} while ((s = MonoIO.FindNext (handle, out rattr, out error)) != null);
 			} finally {
-				MonoIO.FindClose (handle);
+				if (handle != IntPtr.Zero)
+					MonoIO.FindClose (handle);
 			}
 		}
 		
-		
+		internal void CheckPath (string path)
+		{
+			if (path == null)
+				throw new ArgumentNullException ("path");
+			if (path.Length == 0)
+				throw new ArgumentException ("An empty file name is not valid.");
+			if (path.IndexOfAny (Path.InvalidPathChars) != -1)
+				throw new ArgumentException ("Illegal characters in path.");
+			if (Environment.IsRunningOnWindows) {
+				int idx = path.IndexOf (':');
+				if (idx >= 0 && idx != 1)
+					throw new ArgumentException ("path");
+			}
+		}
 	}
 }
