@@ -40,6 +40,7 @@ using System.Security.Permissions;
 using System.Runtime.CompilerServices;
 using System.Runtime.InteropServices;
 using System.Diagnostics.SymbolStore;
+using System.Collections.Generic;
 
 namespace System.Reflection.Emit
 {
@@ -268,6 +269,13 @@ namespace System.Reflection.Emit
 			}
 		}
 
+		public void SetMethodBody (byte[] il, int maxStack, byte[] localSignature,
+			IEnumerable<ExceptionHandler> exceptionHandlers, IEnumerable<int> tokenFixups)
+		{
+			var ilgen = GetILGenerator ();
+			ilgen.Init (il, maxStack, localSignature, exceptionHandlers, tokenFixups);
+		}
+
 		public override Object Invoke(Object obj, BindingFlags invokeAttr, Binder binder, Object[] parameters, CultureInfo culture)
 		{
 			throw NotSupported ();
@@ -356,7 +364,22 @@ namespace System.Reflection.Emit
 			if (ilgen != null)
 				ilgen.label_fixup (this);
 		}
-		
+
+		internal void ResolveUserTypes () {
+			rtype = TypeBuilder.ResolveUserType (rtype);
+			TypeBuilder.ResolveUserTypes (parameters);
+			TypeBuilder.ResolveUserTypes (returnModReq);
+			TypeBuilder.ResolveUserTypes (returnModOpt);
+			if (paramModReq != null) {
+				foreach (var types in paramModReq)
+					TypeBuilder.ResolveUserTypes (types);
+			}
+			if (paramModOpt != null) {
+				foreach (var types in paramModOpt)
+					TypeBuilder.ResolveUserTypes (types);
+			}
+		}
+
 		internal void GenerateDebugInfo (ISymbolWriter symbolWriter)
 		{
 			if (ilgen != null && ilgen.HasDebugInfo) {
@@ -470,7 +493,7 @@ namespace System.Reflection.Emit
 
 		public void AddDeclarativeSecurity (SecurityAction action, PermissionSet pset)
 		{
-#if !NET_2_1
+#if !MOBILE
 			if (pset == null)
 				throw new ArgumentNullException ("pset");
 			if ((action == SecurityAction.RequestMinimum) ||
@@ -655,7 +678,7 @@ namespace System.Reflection.Emit
 
 		public override Module Module {
 			get {
-				return base.Module;
+				return GetModule ();
 			}
 		}
 
