@@ -39,6 +39,9 @@ using System.Runtime.CompilerServices;
 using System.Runtime.InteropServices;
 using System.Collections.Generic;
 using System.Configuration.Assemblies;
+using System.Threading;
+using System.Text;
+using System.Diagnostics.Contracts;
 
 using Mono.Security;
 
@@ -50,11 +53,12 @@ namespace System.Reflection {
 	[ClassInterface(ClassInterfaceType.None)]
 	[StructLayout (LayoutKind.Sequential)]
 #if MOBILE
-	public partial class Assembly : ICustomAttributeProvider {
+	public partial class Assembly : ICustomAttributeProvider, ISerializable
 #else
-	public abstract class Assembly : ICustomAttributeProvider, _Assembly, IEvidenceFactory, ISerializable {
+	public abstract class Assembly : ICustomAttributeProvider, _Assembly, IEvidenceFactory, ISerializable
 #endif
-		internal class ResolveEventHolder {
+	{
+		internal class ResolveEventHolder {	
 			public event ModuleResolveEventHandler ModuleResolve;
 		}
 
@@ -70,7 +74,7 @@ namespace System.Reflection {
 
 			protected override void Dispose (bool disposing)
 			{
-				if (!closed) {
+				if (_isOpen) {
 					/* 
 					 * The returned pointer points inside metadata, so
 					 * we have to increase the refcount of the module, and decrease
@@ -85,7 +89,7 @@ namespace System.Reflection {
 
 		// Note: changes to fields must be reflected in _MonoReflectionAssembly struct (object-internals.h)
 #pragma warning disable 649
-		private IntPtr _mono_assembly;
+		internal IntPtr _mono_assembly;
 #pragma warning restore 649
 
 		private ResolveEventHolder resolve_event_holder;
@@ -112,7 +116,7 @@ namespace System.Reflection {
 		// We can't store the event directly in this class, since the
 		// compiler would silently insert the fields before _mono_assembly
 		//
-		public event ModuleResolveEventHandler ModuleResolve {
+		public virtual event ModuleResolveEventHandler ModuleResolve {
 			[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 			add {
 				resolve_event_holder.ModuleResolve += value;
@@ -135,11 +139,14 @@ namespace System.Reflection {
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern string InternalImageRuntimeVersion ();
 
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		static internal extern string GetAotId ();
+
 		// SECURITY: this should be the only caller to icall get_code_base
 		private string GetCodeBase (bool escaped)
 		{
 			string cb = get_code_base (escaped);
-#if !NET_2_1
+#if !MOBILE
 			if (SecurityManager.SecurityEnabled) {
 				// we cannot divulge local file informations
 				if (String.Compare ("FILE://", 0, cb, 0, 7, true, CultureInfo.InvariantCulture) == 0) {
@@ -209,7 +216,7 @@ namespace System.Reflection {
 					return String.Empty;
 
 				string loc = get_location ();
-#if !NET_2_1
+#if !MOBILE
 				if ((loc != String.Empty) && SecurityManager.SecurityEnabled) {
 					// we cannot divulge local file informations
 					new FileIOPermission (FileIOPermissionAccess.PathDiscovery, loc).Demand ();
@@ -337,19 +344,58 @@ namespace System.Reflection {
 
 		public virtual Stream GetManifestResourceStream (Type type, String name)
 		{
-			string ns;
-			if (type != null) {
-				ns = type.Namespace;
-			} else {
+			StackCrawlMark stackMark = StackCrawlMark.LookForMyCaller;
+			return GetManifestResourceStream(type, name, false, ref stackMark);
+		}
+
+		internal Stream GetManifestResourceStream (Type type, String name, bool skipSecurityCheck, ref StackCrawlMark stackMark)
+		{
+			StringBuilder sb = new StringBuilder ();
+			if (type == null) {
 				if (name == null)
-					throw new ArgumentNullException ("type");
-				ns = null;
+						throw new ArgumentNullException ("type");
+			} else {
+				String nameSpace = type.Namespace;
+				if (nameSpace != null) {
+					sb.Append (nameSpace);
+					if (name != null)
+						sb.Append (Type.Delimiter);
+				}
 			}
 
-			if (ns == null || ns.Length == 0)
-				return GetManifestResourceStream (name);
-			else
-				return GetManifestResourceStream (ns + "." + name);
+			if (name != null)
+				sb.Append(name);
+
+			return GetManifestResourceStream (sb.ToString());
+		}
+
+		internal unsafe Stream GetManifestResourceStream(String name, ref StackCrawlMark stackMark, bool skipSecurityCheck)
+		{
+			return GetManifestResourceStream (null, name, skipSecurityCheck, ref stackMark);
+		}
+
+		internal String GetSimpleName()
+		{
+			AssemblyName aname = GetName (true);
+			return aname.Name;
+		}
+
+		internal byte[] GetPublicKey()
+		{
+			AssemblyName aname = GetName (true);
+			return aname.GetPublicKey ();
+		}
+
+		internal Version GetVersion()
+		{
+			AssemblyName aname = GetName (true);
+			return aname.Version;
+		}
+
+		private AssemblyNameFlags GetFlags()
+		{
+			AssemblyName aname = GetName (true);
+			return aname.Flags;
 		}
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
@@ -380,19 +426,9 @@ namespace System.Reflection {
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		internal extern static void InternalGetAssemblyName (string assemblyFile, AssemblyName aname);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		static extern void FillName (Assembly ass, AssemblyName aname);
-
-		[MonoTODO ("copiedName == true is not supported")]
 		public virtual AssemblyName GetName (Boolean copiedName)
 		{
-#if !MOBILE
-			// CodeBase, which is restricted, will be copied into the AssemblyName object so...
-			if (SecurityManager.SecurityEnabled) {
-				GetCodeBase (true); // this will ensure the Demand is made
-			}
-#endif
-			return UnprotectedGetName ();
+			throw new NotImplementedException ();
 		}
 
 		public virtual AssemblyName GetName ()
@@ -400,14 +436,6 @@ namespace System.Reflection {
 			return GetName (false);
 		}
 
-		// the security runtime requires access to the assemblyname (e.g. to get the strongname)
-		internal virtual AssemblyName UnprotectedGetName ()
-		{
-			AssemblyName aname = new AssemblyName ();
-			FillName (this, aname);
-			return aname;
-		}
-
 		public override string ToString ()
 		{
 			// note: ToString work without requiring CodeBase (so no checks are needed)
@@ -435,40 +463,57 @@ namespace System.Reflection {
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		public static extern Assembly GetEntryAssembly();
 
-		internal Assembly GetSatelliteAssemblyNoThrow (CultureInfo culture, Version version)
-		{
-			return GetSatelliteAssembly (culture, version, false);
-		}
-
 		internal Assembly GetSatelliteAssembly (CultureInfo culture, Version version, bool throwOnError)
 		{
 			if (culture == null)
-				throw new ArgumentException ("culture");
+				throw new ArgumentNullException("culture");
+			Contract.EndContractBlock();
 
-			AssemblyName aname = GetName (true);
-			if (version != null)
-				aname.Version = version;
+			StackCrawlMark stackMark = StackCrawlMark.LookForMyCaller;
+			String name = GetSimpleName() + ".resources";
+			return InternalGetSatelliteAssembly(name, culture, version, true, ref stackMark);
+		}
+
+		internal RuntimeAssembly InternalGetSatelliteAssembly (String name, CultureInfo culture, Version version, bool throwOnFileNotFound, ref StackCrawlMark stackMark)
+		{
+			AssemblyName an = new AssemblyName ();
+
+			an.SetPublicKey (GetPublicKey ());
+			an.Flags = GetFlags () | AssemblyNameFlags.PublicKey;
+
+			if (version == null)
+				an.Version = GetVersion ();
+			else
+				an.Version = version;
+
+			an.CultureInfo = culture;
+			an.Name = name;
 
-			aname.CultureInfo = culture;
-			aname.Name = aname.Name + ".resources";
 			Assembly assembly;
 
 			try {
-				assembly = AppDomain.CurrentDomain.LoadSatellite (aname, false);
+				assembly = AppDomain.CurrentDomain.LoadSatellite (an, false);
 				if (assembly != null)
-					return assembly;
+					return (RuntimeAssembly)assembly;
 			} catch (FileNotFoundException) {
 				assembly = null;
 				// ignore
 			}
 
-			// Try the assembly directory
-			string location = Path.GetDirectoryName (Location);
-			string fullName = Path.Combine (location, Path.Combine (culture.Name, aname.Name + ".dll"));
-			if (!throwOnError && !File.Exists (fullName))
+			if (String.IsNullOrEmpty (Location))
 				return null;
 
-			return LoadFrom (fullName);
+			// Try the assembly directory
+			string location = Path.GetDirectoryName (Location);
+			string fullName = Path.Combine (location, Path.Combine (culture.Name, an.Name + ".dll"));
+
+			try {
+				return (RuntimeAssembly)LoadFrom (fullName);
+			} catch {
+				if (!throwOnFileNotFound && !File.Exists (fullName))
+					return null;
+				throw;
+			}
 		}
 
 #if !MOBILE
@@ -491,7 +536,7 @@ namespace System.Reflection {
 		public static Assembly LoadFrom (String assemblyFile, Evidence securityEvidence)
 		{
 			Assembly a = LoadFrom (assemblyFile, false);
-#if !NET_2_1
+#if !MOBILE
 			if ((a != null) && (securityEvidence != null)) {
 				// merge evidence (i.e. replace defaults with provided evidences)
 				a.Evidence.Merge (securityEvidence);
@@ -716,7 +761,7 @@ namespace System.Reflection {
 				throw new ArgumentNullException ("resourceName");
 			if (resourceName.Length == 0)
 				throw new ArgumentException ("String cannot have zero length.");
-			ManifestResourceInfo result = new ManifestResourceInfo ();
+			ManifestResourceInfo result = new ManifestResourceInfo (null, null, 0);
 			bool found = GetManifestResourceInfoInternal (resourceName, result);
 			if (found)
 				return result;
@@ -763,7 +808,7 @@ namespace System.Reflection {
 			return other._mono_assembly == _mono_assembly;
 		}
 
-#if !NET_2_1
+#if !MOBILE
 		// Code Access Security
 
 		internal void Resolve () 
@@ -848,13 +893,12 @@ namespace System.Reflection {
 		public virtual PermissionSet PermissionSet {
 			get { return this.GrantedPermissionSet; }
 		}
-		
+#endif
+
 		public virtual SecurityRuleSet SecurityRuleSet {
 			get { throw CreateNIE (); }
 		}
 
-#endif
-
 		static Exception CreateNIE ()
 		{
 			return new NotImplementedException ("Derived classes must implement it");
