@@ -37,10 +37,10 @@ using System;
 using System.Security;
 using System.Reflection;
 using System.Threading;
+using System.Runtime.InteropServices.ComTypes;
 
 using System.Runtime.ConstrainedExecution;
 #if !FULL_AOT_RUNTIME
-using System.Runtime.InteropServices.ComTypes;
 using Mono.Interop;
 #endif
 
@@ -68,6 +68,19 @@ namespace System.Runtime.InteropServices
 #endif
 		}
 
+		[MonoTODO]
+		public static bool AreComObjectsAvailableForCleanup ()
+		{
+			return false;
+		}
+
+		[MonoTODO]
+		public static void CleanupUnusedObjectsInCurrentContext ()
+		{
+			if (Environment.IsRunningOnWindows)
+				throw new PlatformNotSupportedException ();
+		}
+
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static IntPtr AllocCoTaskMem (int cb);
 
@@ -191,9 +204,11 @@ namespace System.Runtime.InteropServices
 			return CreateAggregatedObject (pOuter, (object)o);
 		}
 
-#if !FULL_AOT_RUNTIME
 		public static object CreateWrapperOfType (object o, Type t)
 		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			__ComObject co = o as __ComObject;
 			if (co == null)
 				throw new ArgumentException ("o must derive from __ComObject", "o");
@@ -207,12 +222,12 @@ namespace System.Runtime.InteropServices
 			}
 
 			return ComInteropProxy.GetProxy (co.IUnknown, t).GetTransparentProxy ();
+#endif
 		}
 
 		public static TWrapper CreateWrapperOfType<T, TWrapper> (T o) {
 			return (TWrapper)CreateWrapperOfType ((object)o, typeof (TWrapper));
 		}
-#endif
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		[ComVisible (true)]
@@ -329,22 +344,30 @@ namespace System.Runtime.InteropServices
 				return GetCCW (o, T);
 		}
 #endif
+#endif // !FULL_AOT_RUNTIME
 
 		public static IntPtr GetComInterfaceForObject (object o, Type T)
 		{
-#if !MOBILE
+#if MOBILE
+			throw new PlatformNotSupportedException ();
+#else
 			IntPtr pItf = GetComInterfaceForObjectInternal (o, T);
 			AddRef (pItf);
 			return pItf;
-#else
-			throw new NotImplementedException ();
 #endif
 		}
 
+		[MonoTODO]
+		public static IntPtr GetComInterfaceForObject (object o, Type T, CustomQueryInterfaceMode mode)
+		{
+			throw new NotImplementedException ();
+		}
+
 		public static IntPtr GetComInterfaceForObject<T, TInterface> (T o) {
 			return GetComInterfaceForObject ((object)o, typeof (T));
 		}
 
+#if !FULL_AOT_RUNTIME
 		[MonoTODO]
 		public static IntPtr GetComInterfaceForObjectInContext (object o, Type t)
 		{
@@ -383,12 +406,6 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
-		[MonoTODO]
-		public static int GetExceptionCode()
-		{
-			throw new NotImplementedException ();
-		}
-
 		[MonoTODO]
 		[ComVisible (true)]
 		public static IntPtr GetExceptionPointers()
@@ -405,22 +422,35 @@ namespace System.Runtime.InteropServices
 		}
 #endif // !FULL_AOT_RUNTIME
 
-#if !FULL_AOT_RUNTIME
+		public static int GetExceptionCode ()
+		{
+			throw new PlatformNotSupportedException ();
+		}
+
 		public static int GetHRForException (Exception e)
 		{
+			if (e == null) return 0;
+
+#if FEATURE_COMINTEROP
 			var errorInfo = new ManagedErrorInfo(e);
 			SetErrorInfo (0, errorInfo);
+#endif
 
-			return e.hresult;
+			return e._HResult;
 		}
 
 		[MonoTODO]
 		[ReliabilityContract (Consistency.WillNotCorruptState, Cer.Success)]
 		public static int GetHRForLastWin32Error()
 		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			throw new NotImplementedException ();
+#endif
 		}
 
+#if !FULL_AOT_RUNTIME
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern static IntPtr GetIDispatchForObjectInternal (object o);
 
@@ -444,17 +474,6 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		private extern static IntPtr GetIUnknownForObjectInternal (object o);
-
-		public static IntPtr GetIUnknownForObject (object o)
-		{
-			IntPtr pUnk = GetIUnknownForObjectInternal (o);
-			// Internal method does not AddRef
-			AddRef (pUnk);
-			return pUnk;
-		}
-
 		[MonoTODO]
 		public static IntPtr GetIUnknownForObjectInContext (object o)
 		{
@@ -474,25 +493,48 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		private extern static IntPtr GetIUnknownForObjectInternal (object o);
+
+#endif // !FULL_AOT_RUNTIME
+
+		public static IntPtr GetIUnknownForObject (object o)
+		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
+			IntPtr pUnk = GetIUnknownForObjectInternal (o);
+			// Internal method does not AddRef
+			AddRef (pUnk);
+			return pUnk;
+#endif
+		}
+
 		public static void GetNativeVariantForObject (object obj, IntPtr pDstNativeVariant)
 		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			Variant vt = new Variant();
 			vt.SetValue(obj);
 			Marshal.StructureToPtr(vt, pDstNativeVariant, false);
+#endif
 		}
 
 		public static void GetNativeVariantForObject<T> (T obj, IntPtr pDstNativeVariant) {
 			GetNativeVariantForObject ((object)obj, pDstNativeVariant);
 		}
 
-#if !MOBILE
+#if !MOBILE && !FULL_AOT_RUNTIME
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern object GetObjectForCCW (IntPtr pUnk);
 #endif
 
 		public static object GetObjectForIUnknown (IntPtr pUnk)
 		{
-#if !MOBILE
+#if MOBILE || FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			object obj = GetObjectForCCW (pUnk);
 			// was not a CCW
 			if (obj == null) {
@@ -500,24 +542,34 @@ namespace System.Runtime.InteropServices
 				obj = proxy.GetTransparentProxy ();
 			}
 			return obj;
-#else
-			throw new NotImplementedException ();
 #endif
 		}
 
 		public static object GetObjectForNativeVariant (IntPtr pSrcNativeVariant)
 		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			Variant vt = (Variant)Marshal.PtrToStructure(pSrcNativeVariant, typeof(Variant));
 			return vt.GetValue();
+#endif
 		}
 
-		public static T GetObjectForNativeVariant<T> (IntPtr pSrcNativeVariant) {
+		public static T GetObjectForNativeVariant<T> (IntPtr pSrcNativeVariant)
+		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			Variant vt = (Variant)Marshal.PtrToStructure(pSrcNativeVariant, typeof(Variant));
 			return (T)vt.GetValue();
+#endif
 		}
 
 		public static object[] GetObjectsForNativeVariants (IntPtr aSrcNativeVariant, int cVars)
 		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			if (cVars < 0)
 				throw new ArgumentOutOfRangeException ("cVars", "cVars cannot be a negative number.");
 			object[] objects = new object[cVars];
@@ -525,9 +577,14 @@ namespace System.Runtime.InteropServices
 				objects[i] = GetObjectForNativeVariant ((IntPtr)(aSrcNativeVariant.ToInt64 () +
 					i * SizeOf (typeof(Variant))));
 			return objects;
+#endif
 		}
 
-		public static T[] GetObjectsForNativeVariants<T> (IntPtr aSrcNativeVariant, int cVars) {
+		public static T[] GetObjectsForNativeVariants<T> (IntPtr aSrcNativeVariant, int cVars)
+		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			if (cVars < 0)
 				throw new ArgumentOutOfRangeException ("cVars", "cVars cannot be a negative number.");
 			T[] objects = new T[cVars];
@@ -535,14 +592,20 @@ namespace System.Runtime.InteropServices
 				objects[i] = GetObjectForNativeVariant<T> ((IntPtr)(aSrcNativeVariant.ToInt64 () +
 					i * SizeOf (typeof(Variant))));
 			return objects;
+#endif
 		}
 
 		[MonoTODO]
 		public static int GetStartComSlot (Type t)
 		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			throw new NotImplementedException ();
+#endif
 		}
 
+#if !FULL_AOT_RUNTIME
 		[MonoTODO]
 		[Obsolete ("This method has been deprecated")]
 		public static Thread GetThreadFromFiberCookie (int cookie)
@@ -569,12 +632,6 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
-		public static Type GetTypeFromCLSID (Guid clsid)
-		{
-			throw new NotImplementedException ();			
-		}
-
-#if !FULL_AOT_RUNTIME
 		[Obsolete]
 		[MonoTODO]
 		public static string GetTypeInfoName (UCOMITypeInfo pTI)
@@ -582,11 +639,6 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
-		public static string GetTypeInfoName (ITypeInfo typeInfo)
-		{
-			throw new NotImplementedException ();
-		}
-
 		[Obsolete]
 		[MonoTODO]
 		public static Guid GetTypeLibGuid (UCOMITypeLib pTLB)
@@ -638,12 +690,6 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
-		public static object GetUniqueObjectForIUnknown (IntPtr unknown)
-		{
-			throw new NotImplementedException ();
-		}
-#endif
-
 		[MonoTODO]
 		[Obsolete ("This method has been deprecated")]
 		public static IntPtr GetUnmanagedThunkForManagedMethodPtr (IntPtr pfnMethodToWrap, IntPtr pbSignature, int cbSignature)
@@ -651,16 +697,6 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
-#if !MOBILE
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		public extern static bool IsComObject (object o);
-#else
-		public static bool IsComObject (object o)
-		{
-			throw new NotImplementedException ();
-		}
-#endif		
-
 		[MonoTODO]
 		public static bool IsTypeVisibleFromCom (Type t)
 		{
@@ -672,6 +708,31 @@ namespace System.Runtime.InteropServices
 		{
 			throw new NotImplementedException ();
 		}
+#endif // !FULL_AOT_RUNTIME
+
+		public static Type GetTypeFromCLSID (Guid clsid)
+		{
+			throw new PlatformNotSupportedException ();
+		}
+
+		public static string GetTypeInfoName (ITypeInfo typeInfo)
+		{
+			throw new PlatformNotSupportedException ();
+		}
+
+		public static object GetUniqueObjectForIUnknown (IntPtr unknown)
+		{
+			throw new PlatformNotSupportedException ();
+		}
+
+#if !MOBILE
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		public extern static bool IsComObject (object o);
+#else
+		public static bool IsComObject (object o)
+		{
+			throw new PlatformNotSupportedException ();
+		}
 #endif
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
@@ -715,15 +776,8 @@ namespace System.Runtime.InteropServices
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static string PtrToStringUni (IntPtr ptr, int len);
 
-#if !MOBILE
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static string PtrToStringBSTR (IntPtr ptr);
-#else
-		public static string PtrToStringBSTR (IntPtr ptr)
-		{
-			throw new NotImplementedException ();
-		}
-#endif
 		
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		[ComVisible (true)]
@@ -788,7 +842,7 @@ namespace System.Runtime.InteropServices
 				return *(short*)addr;
 
 			short s;
-			String.memcpy ((byte*)&s, (byte*)ptr, 2);
+			Buffer.Memcpy ((byte*)&s, (byte*)ptr, 2);
 			return s;
 		}
 
@@ -800,7 +854,7 @@ namespace System.Runtime.InteropServices
 				return *(short*)addr;
 
 			short s;
-			String.memcpy ((byte*)&s, addr, 2);
+			Buffer.Memcpy ((byte*)&s, addr, 2);
 			return s;
 		}
 
@@ -820,7 +874,7 @@ namespace System.Runtime.InteropServices
 				return *(int*)addr;
 
 			int s;
-			String.memcpy ((byte*)&s, addr, 4);
+			Buffer.Memcpy ((byte*)&s, addr, 4);
 			return s;
 		}
 
@@ -833,7 +887,7 @@ namespace System.Runtime.InteropServices
 				return *(int*)addr;
 			else {
 				int s;
-				String.memcpy ((byte*)&s, addr, 4);
+				Buffer.Memcpy ((byte*)&s, addr, 4);
 				return s;
 			}
 		}
@@ -857,7 +911,7 @@ namespace System.Runtime.InteropServices
 				return *(long*)ptr;
 
 			long s;
-			String.memcpy ((byte*)&s, addr, 8);
+			Buffer.Memcpy ((byte*)&s, addr, 8);
 			return s;
 		}
 
@@ -869,7 +923,7 @@ namespace System.Runtime.InteropServices
 				return *(long*)addr;
 			
 			long s;
-			String.memcpy ((byte*)&s, addr, 8);
+			Buffer.Memcpy ((byte*)&s, addr, 8);
 			return s;
 		}
 
@@ -934,16 +988,22 @@ namespace System.Runtime.InteropServices
 #if !FULL_AOT_RUNTIME
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern static int ReleaseComObjectInternal (object co);
+#endif
 
 		public static int ReleaseComObject (object o)
 		{
+#if FULL_AOT_RUNTIME
+			throw new PlatformNotSupportedException ();
+#else
 			if (o == null)
 				throw new ArgumentException ("Value cannot be null.", "o");
 			if (!IsComObject (o))
 				throw new ArgumentException ("Value must be a Com object.", "o");
 			return ReleaseComObjectInternal (o);
+#endif
 		}
 
+#if !FULL_AOT_RUNTIME
 		[Obsolete]
 		[MonoTODO]
 		public static void ReleaseThreadCache()
@@ -975,6 +1035,21 @@ namespace System.Runtime.InteropServices
 			return SizeOf (structure.GetType ());
 		}
 
+		internal static uint SizeOfType (Type type)
+		{
+			return (uint) SizeOf (type);
+		}
+
+		internal static uint AlignedSizeOf<T> () where T : struct
+		{
+			uint size = SizeOfType (typeof (T));
+			if (size == 1 || size == 2)
+				return size;
+			if (IntPtr.Size == 8 && size == 4)
+				return size;
+			return (size + 3) & (~((uint)3));
+		}
+
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static IntPtr StringToBSTR (string s);
 
@@ -1032,25 +1107,21 @@ namespace System.Runtime.InteropServices
 		{
 			if (s == null)
 				throw new ArgumentNullException ("s");
-			int len = s.Length;
-			IntPtr ctm = AllocCoTaskMem ((len+1) * 2 + 4);
-			byte [] buffer = null;
-			WriteInt32 (ctm, 0, len*2);
-			try {
-				buffer = s.GetBuffer ();
 
-				for (int i = 0; i < len; i++)
-					WriteInt16 (ctm, 4 + (i * 2), (short) ((buffer [(i*2)] << 8) | (buffer [i*2+1])));
-				WriteInt16 (ctm, 4 + buffer.Length, 0);
-			} finally {
-				if (buffer != null)
-					for (int i = buffer.Length; i > 0; ){
-						i--;
-						buffer [i] = 0;
-					}
+			byte[] buffer = s.GetBuffer ();
+			int len = s.Length;
+			
+			// SecureString doesn't take endian-ness into account. 
+			// Therefore swap bytes here before we send it to c-side if little-endian.
+			if (BitConverter.IsLittleEndian) {
+				for (int i = 0; i < buffer.Length; i += 2) {
+					byte b = buffer[i];
+					buffer[i] = buffer[i + 1];
+					buffer[i + 1] = b;
+				}
 			}
-			return (IntPtr) ((long)ctm + 4);
-		}
+			return BufferToBSTR (buffer, len);
+        }
 
 		public static IntPtr SecureStringToCoTaskMemAnsi (SecureString s)
 		{
@@ -1137,6 +1208,10 @@ namespace System.Runtime.InteropServices
 				throw ex;
 		}
 
+
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		public extern static IntPtr BufferToBSTR (Array ptr, int slen);
+
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static IntPtr UnsafeAddrOfPinnedArrayElement (Array arr, int index);
 
@@ -1171,7 +1246,7 @@ namespace System.Runtime.InteropServices
 			if (((uint)addr & 1) == 0)
 				*(short*)addr = val;
 			else
-				String.memcpy (addr, (byte*)&val, 2);
+				Buffer.Memcpy (addr, (byte*)&val, 2);
 		}
 
 		public static unsafe void WriteInt16 (IntPtr ptr, int ofs, short val)
@@ -1181,7 +1256,7 @@ namespace System.Runtime.InteropServices
 			if (((uint)addr & 1) == 0)
 				*(short*)addr = val;
 			else {
-				String.memcpy (addr, (byte*)&val, 2);
+				Buffer.Memcpy (addr, (byte*)&val, 2);
 			}
 		}
 
@@ -1215,7 +1290,7 @@ namespace System.Runtime.InteropServices
 			if (((uint)addr & 3) == 0) 
 				*(int*)addr = val;
 			else {
-				String.memcpy (addr, (byte*)&val, 4);
+				Buffer.Memcpy (addr, (byte*)&val, 4);
 			}
 		}
 
@@ -1226,7 +1301,7 @@ namespace System.Runtime.InteropServices
 			if (((uint)addr & 3) == 0) 
 				*(int*)addr = val;
 			else {
-				String.memcpy (addr, (byte*)&val, 4);
+				Buffer.Memcpy (addr, (byte*)&val, 4);
 			}
 		}
 
@@ -1246,7 +1321,7 @@ namespace System.Runtime.InteropServices
 			if (((uint)addr & 7) == 0) 
 				*(long*)addr = val;
 			else 
-				String.memcpy (addr, (byte*)&val, 8);
+				Buffer.Memcpy (addr, (byte*)&val, 8);
 		}
 
 		public static unsafe void WriteInt64 (IntPtr ptr, int ofs, long val)
@@ -1258,7 +1333,7 @@ namespace System.Runtime.InteropServices
 			if (((uint)addr & 7) == 0) 
 				*(long*)addr = val;
 			else 
-				String.memcpy (addr, (byte*)&val, 8);
+				Buffer.Memcpy (addr, (byte*)&val, 8);
 		}
 
 		[MonoTODO]
@@ -1503,6 +1578,7 @@ namespace System.Runtime.InteropServices
 			return null;
 		}
 
+#if FEATURE_COMINTEROP
 		[DllImport ("oleaut32.dll", CharSet=CharSet.Unicode, EntryPoint = "SetErrorInfo")]
 		static extern int _SetErrorInfo (int dwReserved,
 			[MarshalAs(UnmanagedType.Interface)] IErrorInfo pIErrorInfo);
@@ -1551,7 +1627,7 @@ namespace System.Runtime.InteropServices
 			}
 			return retVal;
 		}
-
+#endif
 		public static Exception GetExceptionForHR (int errorCode)
 		{
 			return GetExceptionForHR (errorCode, IntPtr.Zero);
@@ -1559,7 +1635,7 @@ namespace System.Runtime.InteropServices
 
 		public static Exception GetExceptionForHR (int errorCode, IntPtr errorInfo)
 		{
-#if !MOBILE
+#if FEATURE_COMINTEROP
 			IErrorInfo info = null;
 			if (errorInfo != (IntPtr)(-1)) {
 				if (errorInfo == IntPtr.Zero) {
@@ -1571,7 +1647,7 @@ namespace System.Runtime.InteropServices
 				}
 			}
 
-			if (info is ManagedErrorInfo && ((ManagedErrorInfo) info).Exception.hresult == errorCode) {
+			if (info is ManagedErrorInfo && ((ManagedErrorInfo) info).Exception._HResult == errorCode) {
 				return ((ManagedErrorInfo) info).Exception;
 			}
 
@@ -1598,13 +1674,11 @@ namespace System.Runtime.InteropServices
 #endif
 		}
 
-#if !FULL_AOT_RUNTIME
 		public static int FinalReleaseComObject (object o)
 		{
 			while (ReleaseComObject (o) != 0);
 			return 0;
 		}
-#endif
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private static extern Delegate GetDelegateForFunctionPointerInternal (IntPtr ptr, Type t);
@@ -1644,5 +1718,9 @@ namespace System.Runtime.InteropServices
 			
 			return GetFunctionPointerForDelegateInternal ((Delegate)(object)d);
 		}
+
+		internal static void SetLastWin32Error (int error)
+		{
+		}
 	}
 }
