@@ -10,6 +10,7 @@
 
 using System;
 using System.Collections;
+using System.Collections.Generic;
 using System.IO;
 using System.Runtime.Serialization;
 using System.Runtime.Serialization.Formatters.Binary;
@@ -167,6 +168,8 @@ namespace MonoTests.System.IO
 					info = new DirectoryInfo ("/test/");
 					Assert.AreEqual ("test", info.Name, "#4");
 				} else {
+					Directory.SetCurrentDirectory (@"C:\");
+
 					info = new DirectoryInfo (@"c:");
 					Assert.AreEqual (@"C:\", info.Name, "#4");
 
@@ -216,6 +219,8 @@ namespace MonoTests.System.IO
 					Assert.IsNotNull (info.Parent, "#7a");
 					Assert.AreEqual ("/", info.Parent.FullName, "#7b");
 				} else {
+					Directory.SetCurrentDirectory (@"C:\");
+
 					info = new DirectoryInfo (@"c:");
 					Assert.IsNull (info.Parent, "#4");
 
@@ -417,6 +422,8 @@ namespace MonoTests.System.IO
 				di = new DirectoryInfo ("/test/");
 				Assert.AreEqual ("/test/", di.FullName, "#D4");
 			} else {
+				Directory.SetCurrentDirectory (@"C:\");
+
 				di = new DirectoryInfo (@"c:");
 				Assert.AreEqual (@"C:\", di.FullName, "#D1");
 
@@ -515,9 +522,7 @@ namespace MonoTests.System.IO
 				Assert.AreEqual (0, info.GetDirectories ("test[12]*").Length, "#6");
 				Assert.AreEqual (2, info.GetDirectories ("test2*0").Length, "#7");
 				Assert.AreEqual (4, info.GetDirectories ("*test*").Length, "#8");
-#if NET_2_0
 				Assert.AreEqual (6, info.GetDirectories ("*", SearchOption.AllDirectories).Length, "#9");
-#endif
 			} finally {
 				DeleteDir (path);
 			}
@@ -571,7 +576,6 @@ namespace MonoTests.System.IO
 			Assert.AreEqual(directoryToBeLookedFor, directoriesFound[0].Name, "The name of the directory found should match the expected one.");
 		}
 
-#if NET_2_0
 		[Test] // GetDirectories (String, SearchOption)
 		public void GetDirectories3_SearchPattern_Null ()
 		{
@@ -586,7 +590,6 @@ namespace MonoTests.System.IO
 				Assert.AreEqual ("searchPattern", ex.ParamName, "#5");
 			}
 		}
-#endif
 
 		[Test] // GetFiles ()
 		public void GetFiles1 ()
@@ -691,7 +694,6 @@ namespace MonoTests.System.IO
 			}
 		}
 
-#if NET_2_0
 		[Test] // GetFiles (String, SearchOption)
 		public void GetFiles3_SearchPattern_Null ()
 		{
@@ -706,7 +708,6 @@ namespace MonoTests.System.IO
 				Assert.AreEqual ("searchPattern", ex.ParamName, "#5");
 			}
 		}
-#endif
 
 		[Test]
 		public void GetFileSystemInfos2_SearchPattern_Null ()
@@ -1059,6 +1060,21 @@ namespace MonoTests.System.IO
 			Assert.AreEqual (TempFolder + DSC + "ToString.Test", info.ToString ());
 		}
 
+		[Test]
+		public void EnumerateFileSystemInfosTest ()
+		{
+			var dirInfo = new DirectoryInfo (TempFolder);
+			dirInfo.CreateSubdirectory ("1").CreateSubdirectory ("a");
+			dirInfo.CreateSubdirectory ("2").CreateSubdirectory ("b");
+
+			var l = new List<string> ();
+			foreach (var info in dirInfo.EnumerateFileSystemInfos ("*", SearchOption.AllDirectories))
+				l.Add (info.Name);
+
+			l.Sort ();
+			Assert.AreEqual ("1,2,a,b", string.Join (",", l), "#1");
+		}
+
 #if !MOBILE
 		[Test]
 		public void Serialization ()
@@ -1111,7 +1127,7 @@ namespace MonoTests.System.IO
 			try {
 				Directory.CreateDirectory (path);
 				Directory.CreateDirectory (dir);
-				Mono.Unix.UnixSymbolicLinkInfo li = new Mono.Unix.UnixSymbolicLinkInfo (link);
+				global::Mono.Unix.UnixSymbolicLinkInfo li = new global::Mono.Unix.UnixSymbolicLinkInfo (link);
 				li.CreateSymbolicLinkTo (dir);
 
 				DirectoryInfo info = new DirectoryInfo (path);
