@@ -25,10 +25,6 @@ namespace MonoTests.System.IO
 	{
 		string TempFolder = Path.Combine (Path.GetTempPath (), "MonoTests.System.IO.Tests");
 		static readonly char DSC = Path.DirectorySeparatorChar;
-		static bool MacOSX = false;
-
-		[DllImport ("libc")]
-		static extern int uname (IntPtr buf);
 
 		[TearDown]
 		public void TearDown ()
@@ -44,13 +40,6 @@ namespace MonoTests.System.IO
 				Directory.Delete (TempFolder, true);
 
 			Directory.CreateDirectory (TempFolder);
-#if !MOBILE			
-			// from XplatUI.cs
-			IntPtr buf = Marshal.AllocHGlobal (8192);
-			if (uname (buf) == 0)
-				MacOSX = Marshal.PtrToStringAnsi (buf) == "Darwin";
-			Marshal.FreeHGlobal (buf);
-#endif
 		}
 
 		public void TestCtr ()
@@ -1728,5 +1717,30 @@ namespace MonoTests.System.IO
 			}
 		}
 #endif
+
+		[Test] // Covers #11699
+		public void ReadWriteFileLength ()
+		{
+			int bufferSize = 128;
+			int readLength = 1;
+			int writeLength = bufferSize + 1;
+
+			string path = TempFolder + DSC + "readwritefilelength.tmp";
+
+			try {
+				File.WriteAllBytes (path, new byte [readLength + 1]);
+
+				using (var file = new FileStream (path, FileMode.Open, FileAccess.ReadWrite, FileShare.Read,
+												 bufferSize, FileOptions.SequentialScan))
+				{
+					file.Read (new byte [readLength], 0, readLength);
+					file.Write (new byte [writeLength], 0, writeLength);
+
+					Assert.AreEqual (readLength + writeLength, file.Length);
+				}
+			} finally {
+				DeleteFile (path);
+			}
+		}
 	}
 }
