@@ -6,6 +6,7 @@
 //
 // (C) 2002, 2003 Motus Technologies Inc. (http://www.motus.com)
 // Copyright (C) 2004-2007,2011 Novell, Inc (http://www.novell.com)
+// Copyright 2011 Xamarin Inc (http://www.xamarin.com).
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the
@@ -103,7 +104,7 @@ public class CryptoConfigTest {
 		CreateFromName ("System.Security.Cryptography.HMACSHA1", "System.Security.Cryptography.HMACSHA1");
 		CreateFromName ("MACTripleDES", "System.Security.Cryptography.MACTripleDES");
 		CreateFromName ("System.Security.Cryptography.MACTripleDES", "System.Security.Cryptography.MACTripleDES");
-#if NET_2_0
+
 		// new HMAC - new base class doesn't return anything with it's short name
 		Assert.IsNull (CryptoConfig.CreateFromName ("HMAC"), "HMAC");
 		CreateFromName ("System.Security.Cryptography.HMAC", "System.Security.Cryptography.HMACSHA1");
@@ -122,13 +123,13 @@ public class CryptoConfigTest {
 		CreateFromName ("RIPEMD-160", "System.Security.Cryptography.RIPEMD160Managed");
 		CreateFromName ("System.Security.Cryptography.RIPEMD160", "System.Security.Cryptography.RIPEMD160Managed");
 		// x.509 stuff
-#if !TARGET_JVM //TargetJvmNotWorking - this algorithm should be added to System
+#if !MOBILE
 		CreateFromName ("X509Chain", "System.Security.Cryptography.X509Certificates.X509Chain");
-#endif
-#endif
+
 		// note: CryptoConfig can create any object !
 		CreateFromName ("System.Security.Cryptography.CryptoConfig", "System.Security.Cryptography.CryptoConfig");
 		CreateFromName ("System.IO.MemoryStream", "System.IO.MemoryStream");
+#endif
 		// non existing algo should return null (without exception)
 		Assert.IsNull (CryptoConfig.CreateFromName ("NonExistingAlgorithm"), "NonExistingAlgorithm");
 	}
@@ -136,15 +137,20 @@ public class CryptoConfigTest {
 	// additional names (URL) used for XMLDSIG (System.Security.Cryptography.Xml)
 	// URL taken from http://www.w3.org/TR/2002/REC-xmldsig-core-20020212/
 	[Test]
+#if MOBILE
+	[Ignore ("System.Security.dll is not part of Moonlight, MonoTouch and Mono for Android")]
+#endif
 	public void CreateFromURL () 
 	{
 		// URL used in SignatureMethod element
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig#dsa-sha1", "System.Security.Cryptography.DSASignatureDescription");
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig#rsa-sha1", "System.Security.Cryptography.RSAPKCS1SHA1SignatureDescription");
+		CreateFromName ("http://www.w3.org/2001/04/xmldsig-more#rsa-sha256", "System.Security.Cryptography.RSAPKCS1SHA256SignatureDescription");
+		CreateFromName ("http://www.w3.org/2001/04/xmldsig-more#rsa-sha384", "System.Security.Cryptography.RSAPKCS1SHA384SignatureDescription");
+		CreateFromName ("http://www.w3.org/2001/04/xmldsig-more#rsa-sha512", "System.Security.Cryptography.RSAPKCS1SHA512SignatureDescription");
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig#hmac-sha1", null);
 		// URL used in DigestMethod element 
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig#sha1", "System.Security.Cryptography.SHA1CryptoServiceProvider");
-#if !TARGET_JVM //TargetJvmNotWorking - algorithms from System.Security assembly
 		// URL used in Canonicalization or Transform elements 
 		CreateFromName ("http://www.w3.org/TR/2001/REC-xml-c14n-20010315", "System.Security.Cryptography.Xml.XmlDsigC14NTransform");
 		CreateFromName ("http://www.w3.org/TR/2001/REC-xml-c14n-20010315#WithComments", "System.Security.Cryptography.Xml.XmlDsigC14NWithCommentsTransform");
@@ -165,7 +171,6 @@ public class CryptoConfigTest {
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig# KeyValue/DSAKeyValue", "System.Security.Cryptography.Xml.DSAKeyValue");
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig# KeyValue/RSAKeyValue", "System.Security.Cryptography.Xml.RSAKeyValue");
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig# RetrievalMethod", "System.Security.Cryptography.Xml.KeyInfoRetrievalMethod");
-#if NET_2_0
 		CreateFromName ("http://www.w3.org/2001/04/xmlenc#sha256", "System.Security.Cryptography.SHA256Managed");
 		CreateFromName ("http://www.w3.org/2001/04/xmlenc#sha384", null);
 		CreateFromName ("http://www.w3.org/2001/04/xmlenc#sha512", "System.Security.Cryptography.SHA512Managed");
@@ -174,48 +179,30 @@ public class CryptoConfigTest {
 		CreateFromName ("http://www.w3.org/2001/04/xmldsig-more#hmac-sha384", "System.Security.Cryptography.HMACSHA384");
 		CreateFromName ("http://www.w3.org/2001/04/xmldsig-more#hmac-sha512", "System.Security.Cryptography.HMACSHA512");
 		CreateFromName ("http://www.w3.org/2001/04/xmldsig-more#hmac-ripemd160", "System.Security.Cryptography.HMACRIPEMD160");
-#endif
-#endif
 	}
 
 	[Test]
 	public void CreateFromName_UpperCase () 
 	{
 		CreateFromName ("SHA", "System.Security.Cryptography.SHA1CryptoServiceProvider");
-#if NET_2_0
 		CreateFromName ("SYSTEM.SECURITY.CRYPTOGRAPHY.TRIPLEDES", "System.Security.Cryptography.TripleDESCryptoServiceProvider");  
 		CreateFromName ("HTTP://WWW.W3.ORG/2000/09/XMLDSIG#DSA-SHA1", "System.Security.Cryptography.DSASignatureDescription");
-#else
-		CreateFromName ("SYSTEM.SECURITY.CRYPTOGRAPHY.TRIPLEDES", null);  
-		CreateFromName ("HTTP://WWW.W3.ORG/2000/09/XMLDSIG#DSA-SHA1", null);
-#endif
 	}
 
 	[Test]
 	public void CreateFromName_LowerCase () 
 	{
-#if NET_2_0
 		CreateFromName ("sha", "System.Security.Cryptography.SHA1CryptoServiceProvider");
 		CreateFromName ("system.security.cryptography.tripledes", "System.Security.Cryptography.TripleDESCryptoServiceProvider");  
-#else
-		CreateFromName ("sha", null);
-		CreateFromName ("system.security.cryptography.tripledes", null);  
-#endif
 		CreateFromName ("http://www.w3.org/2000/09/xmldsig#dsa-sha1", "System.Security.Cryptography.DSASignatureDescription");
 	}
 
 	[Test]
 	public void CreateFromName_MixedCase () 
 	{
-#if NET_2_0
 		CreateFromName ("ShA", "System.Security.Cryptography.SHA1CryptoServiceProvider");
 		CreateFromName ("SyStEm.SeCuRiTy.CrYpToGrApHy.TrIpLeDeS", "System.Security.Cryptography.TripleDESCryptoServiceProvider");  
 		CreateFromName ("hTtP://wWw.W3.oRg/2000/09/xMlDsIg#dSa-sHa1", "System.Security.Cryptography.DSASignatureDescription");
-#else
-		CreateFromName ("ShA", null);
-		CreateFromName ("SyStEm.SeCuRiTy.CrYpToGrApHy.TrIpLeDeS", null);  
-		CreateFromName ("hTtP://wWw.W3.oRg/2000/09/xMlDsIg#dSa-sHa1", null);
-#endif
 	}
 
 	// Tests created using "A Layer Man Guide to ASN.1" from RSA, page 19-20
@@ -226,11 +213,7 @@ public class CryptoConfigTest {
 	static byte[] oidmd5withRSAEncryption = { 0x06, 0x09, 0x2A, 0x86, 0x48, 0x86, 0xF7, 0x0D, 0x01, 0x01, 0x04 };
 
 	[Test]
-#if NET_2_0
 	[ExpectedException (typeof (ArgumentNullException))]
-#else
-	[ExpectedException (typeof (NullReferenceException))]
-#endif
 	public void EncodeOIDNull () 
 	{
 		byte[] o = CryptoConfig.EncodeOID (null);
@@ -323,50 +306,35 @@ public class CryptoConfigTest {
 		MapNameToOID ("System.Security.Cryptography.SHA1", "1.3.14.3.2.26");
 //		MapNameToOID ("System.Security.Cryptography.HashAlgorithm", "1.3.14.3.2.26");
 		MapNameToOID ("System.Security.Cryptography.SHA1CryptoServiceProvider", "1.3.14.3.2.26");
+		MapNameToOID ("System.Security.Cryptography.SHA1Cng", "1.3.14.3.2.26");
 		MapNameToOID ("System.Security.Cryptography.SHA1Managed", "1.3.14.3.2.26");
 		MapNameToOID ("MD5", "1.2.840.113549.2.5");
 		MapNameToOID ("System.Security.Cryptography.MD5", "1.2.840.113549.2.5");
 		MapNameToOID ("System.Security.Cryptography.MD5CryptoServiceProvider", "1.2.840.113549.2.5");
-#if NET_2_0
 		MapNameToOID ("SHA256", "2.16.840.1.101.3.4.2.1");
+		MapNameToOID ("System.Security.Cryptography.SHA256CryptoServiceProvider", "2.16.840.1.101.3.4.2.1");
+		MapNameToOID ("System.Security.Cryptography.SHA256Cng", "2.16.840.1.101.3.4.2.1");
 		MapNameToOID ("System.Security.Cryptography.SHA256", "2.16.840.1.101.3.4.2.1");
 		MapNameToOID ("System.Security.Cryptography.SHA256Managed", "2.16.840.1.101.3.4.2.1");
 		MapNameToOID ("SHA384", "2.16.840.1.101.3.4.2.2");
+		MapNameToOID ("System.Security.Cryptography.SHA384CryptoServiceProvider", "2.16.840.1.101.3.4.2.2");
+		MapNameToOID ("System.Security.Cryptography.SHA384Cng", "2.16.840.1.101.3.4.2.2");
 		MapNameToOID ("System.Security.Cryptography.SHA384", "2.16.840.1.101.3.4.2.2");
 		MapNameToOID ("System.Security.Cryptography.SHA384Managed", "2.16.840.1.101.3.4.2.2");
 		MapNameToOID ("SHA512", "2.16.840.1.101.3.4.2.3");
+		MapNameToOID ("System.Security.Cryptography.SHA512CryptoServiceProvider", "2.16.840.1.101.3.4.2.3");
+		MapNameToOID ("System.Security.Cryptography.SHA512Cng", "2.16.840.1.101.3.4.2.3");
 		MapNameToOID ("System.Security.Cryptography.SHA512", "2.16.840.1.101.3.4.2.3");
 		MapNameToOID ("System.Security.Cryptography.SHA512Managed", "2.16.840.1.101.3.4.2.3");
 		MapNameToOID ("RIPEMD160", "1.3.36.3.2.1");
 		MapNameToOID ("System.Security.Cryptography.RIPEMD160", "1.3.36.3.2.1");
 		MapNameToOID ("System.Security.Cryptography.RIPEMD160Managed", "1.3.36.3.2.1");
-#else
-		MapNameToOID ("SHA256", "2.16.840.1.101.3.4.1");
-//		MapNameToOID ("SHA-256", "2.16.840.1.101.3.4.1");
-		MapNameToOID ("System.Security.Cryptography.SHA256", "2.16.840.1.101.3.4.1");
-		MapNameToOID ("System.Security.Cryptography.SHA256Managed", "2.16.840.1.101.3.4.1");
-		MapNameToOID ("SHA384", "2.16.840.1.101.3.4.2");
-//		MapNameToOID ("SHA-384", "2.16.840.1.101.3.4.2");
-		MapNameToOID ("System.Security.Cryptography.SHA384", "2.16.840.1.101.3.4.2");
-		MapNameToOID ("System.Security.Cryptography.SHA384Managed", "2.16.840.1.101.3.4.2");
-		MapNameToOID ("SHA512", "2.16.840.1.101.3.4.3");
-//		MapNameToOID ("SHA-512", "2.16.840.1.101.3.4.3");
-		MapNameToOID ("System.Security.Cryptography.SHA512", "2.16.840.1.101.3.4.3");
-		MapNameToOID ("System.Security.Cryptography.SHA512Managed", "2.16.840.1.101.3.4.3");
-#endif
 		// LAMESPEC: only documentated in ".NET Framework Security" book
 		MapNameToOID ("TripleDESKeyWrap", "1.2.840.113549.1.9.16.3.6");
-#if NET_2_0
 		// new OID defined in Fx 2.0
 		MapNameToOID ("DES", "1.3.14.3.2.7");
 		MapNameToOID ("TripleDES", "1.2.840.113549.3.7");
 		MapNameToOID ("RC2", "1.2.840.113549.3.2");
-#else
-		// no OID defined before Fx 2.0
-		MapNameToOID ("DES", null);
-		MapNameToOID ("TripleDES", null);
-		MapNameToOID ("RC2", null);
-#endif
 		MapNameToOID ("RSA", null);
 		MapNameToOID ("DSA", null);
 		MapNameToOID ("3DES", null);
@@ -385,7 +353,6 @@ public class CryptoConfigTest {
 		MapNameToOID ("RandomNumberGenerator", null);
 		MapNameToOID ("System.Security.Cryptography.RandomNumberGenerator", null);
 		MapNameToOID ("System.Security.Cryptography.KeyedHashAlgorithm", null);
-#if NET_2_0
 		MapNameToOID ("HMAC", null);
 		MapNameToOID ("System.Security.Cryptography.HMAC", null);
 		MapNameToOID ("HMACMD5", null);
@@ -398,7 +365,6 @@ public class CryptoConfigTest {
 		MapNameToOID ("System.Security.Cryptography.HMACSHA384", null);
 		MapNameToOID ("HMACSHA512", null);
 		MapNameToOID ("System.Security.Cryptography.HMACSHA512", null);
-#endif
 		MapNameToOID ("HMACSHA1", null);
 		MapNameToOID ("System.Security.Cryptography.HMACSHA1", null);
 		MapNameToOID ("MACTripleDES", null);
@@ -411,35 +377,21 @@ public class CryptoConfigTest {
 	public void MapNameToOID_UpperCase () 
 	{
 		MapNameToOID ("SHA1", "1.3.14.3.2.26");
-#if NET_2_0
 		MapNameToOID ("SYSTEM.SECURITY.CRYPTOGRAPHY.MD5CRYPTOSERVICEPROVIDER", "1.2.840.113549.2.5");
-#else
-		MapNameToOID ("SYSTEM.SECURITY.CRYPTOGRAPHY.MD5CRYPTOSERVICEPROVIDER", null);
-#endif
 	}
 
 	[Test]
 	public void MapNameToOID_LowerCase () 
 	{
-#if NET_2_0
 		MapNameToOID ("sha1", "1.3.14.3.2.26");
 		MapNameToOID ("system.security.cryptography.md5cryptoserviceprovider", "1.2.840.113549.2.5");
-#else
-		MapNameToOID ("sha1", null);
-		MapNameToOID ("system.security.cryptography.md5cryptoserviceprovider", null);
-#endif
 	}
 
 	[Test]
 	public void MapNameToOID_MixedCase () 
 	{
-#if NET_2_0
 		MapNameToOID ("sHa1", "1.3.14.3.2.26");
 		MapNameToOID ("SySteM.SeCuRiTy.CrYpToGrApHy.Md5cRyPtOsErViCePrOvIdEr", "1.2.840.113549.2.5");
-#else
-		MapNameToOID ("sHa1", null);
-		MapNameToOID ("SySteM.SeCuRiTy.CrYpToGrApHy.Md5cRyPtOsErViCePrOvIdEr", null);
-#endif
 	}
 
 	[Test]
