@@ -6,6 +6,7 @@
  *
  * (C) 2002-2011 Novell, Inc.
  * Copyright 2011 Xamarin Inc
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 
 #include <config.h>
@@ -54,7 +55,19 @@
 #endif
 #endif
 
-#if defined(PLATFORM_MACOSX) || defined(__OpenBSD__) || defined(__FreeBSD__)
+#if defined(PLATFORM_MACOSX)
+#define USE_OSX_LOADER
+#endif
+
+#if ( defined(__OpenBSD__) || defined(__FreeBSD__) ) && defined(HAVE_LINK_H)
+#define USE_BSD_LOADER
+#endif
+
+#if defined(__HAIKU__)
+#define USE_HAIKU_LOADER
+#endif
+
+#if defined(USE_OSX_LOADER) || defined(USE_BSD_LOADER)
 #include <sys/proc.h>
 #include <sys/sysctl.h>
 #  if !defined(__OpenBSD__)
@@ -82,9 +95,7 @@
 
 #include <mono/io-layer/wapi.h>
 #include <mono/io-layer/wapi-private.h>
-#include <mono/io-layer/handles-private.h>
 #include <mono/io-layer/process-private.h>
-#include <mono/io-layer/threads.h>
 #include <mono/io-layer/io-trace.h>
 #include <mono/utils/strenc.h>
 #include <mono/utils/mono-path.h>
@@ -96,6 +107,9 @@
 #include <mono/utils/mono-proclib.h>
 #include <mono/utils/mono-once.h>
 #include <mono/utils/mono-logger-internals.h>
+#include <mono/utils/w32handle.h>
+
+#define STILL_ACTIVE STATUS_PENDING
 
 /* The process' environment strings */
 #if defined(__APPLE__)
@@ -114,22 +128,28 @@ static char *mono_environ[1] = { NULL };
 extern char **environ;
 #endif
 
-static guint32 process_wait (gpointer handle, guint32 timeout, gboolean alertable);
+static guint32 process_wait (gpointer handle, guint32 timeout, gboolean *alerted);
 static void process_close (gpointer handle, gpointer data);
+static void process_details (gpointer data);
+static const gchar* process_typename (void);
+static gsize process_typesize (void);
 static gboolean is_pid_valid (pid_t pid);
 
-#if !(defined(PLATFORM_MACOSX) || defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__HAIKU__))
+#if !(defined(USE_OSX_LOADER) || defined(USE_BSD_LOADER) || defined(USE_HAIKU_LOADER))
 static FILE *
 open_process_map (int pid, const char *mode);
 #endif
 
-struct _WapiHandleOps _wapi_process_ops = {
+static MonoW32HandleOps _wapi_process_ops = {
 	process_close,		/* close_shared */
 	NULL,				/* signal */
 	NULL,				/* own */
 	NULL,				/* is_owned */
 	process_wait,			/* special_wait */
-	NULL				/* prewait */	
+	NULL,				/* prewait */
+	process_details,	/* details */
+	process_typename,	/* typename */
+	process_typesize,	/* typesize */
 };
 
 #if HAVE_SIGACTION
@@ -165,7 +185,7 @@ lookup_process_handle (gpointer handle)
 	WapiHandle_process *process_data;
 	gboolean ret;
 
-	ret = _wapi_lookup_handle (handle, WAPI_HANDLE_PROCESS,
+	ret = mono_w32handle_lookup (handle, MONO_W32HANDLE_PROCESS,
 							   (gpointer *)&process_data);
 	if (!ret)
 		return NULL;
@@ -392,10 +412,10 @@ is_managed_binary (const char *filename)
 	off_t new_offset;
 	unsigned char buffer[8];
 	off_t file_size, optional_header_offset;
-	off_t pe_header_offset;
+	off_t pe_header_offset, clr_header_offset;
 	gboolean managed = FALSE;
 	int num_read;
-	guint32 first_word, second_word;
+	guint32 first_word, second_word, magic_number;
 	
 	/* If we are unable to open the file, then we definitely
 	 * can't say that it is managed. The child mono process
@@ -460,13 +480,34 @@ is_managed_binary (const char *filename)
 	if ((num_read != 2) || ((buffer[0] | (buffer[1] << 8)) < 216))
 		goto leave;
 
+	optional_header_offset = pe_header_offset + 24;
+
+	/* Read the PE magic number */
+	new_offset = lseek (file, optional_header_offset, SEEK_SET);
+	
+	if (new_offset != optional_header_offset)
+		goto leave;
+
+	num_read = read (file, buffer, 2);
+
+	if (num_read != 2)
+		goto leave;
+
+	magic_number = (buffer[0] | (buffer[1] << 8));
+	
+	if (magic_number == 0x10B)  // PE32
+		clr_header_offset = 208;
+	else if (magic_number == 0x20B)  // PE32+
+		clr_header_offset = 224;
+	else
+		goto leave;
+
 	/* Read the CLR header address and size fields. These will be
 	 * zero if the binary is not managed.
 	 */
-	optional_header_offset = pe_header_offset + 24;
-	new_offset = lseek (file, optional_header_offset + 208, SEEK_SET);
+	new_offset = lseek (file, optional_header_offset + clr_header_offset, SEEK_SET);
 
-	if (new_offset != optional_header_offset + 208)
+	if (new_offset != optional_header_offset + clr_header_offset)
 		goto leave;
 
 	num_read = read (file, buffer, 8);
@@ -564,7 +605,6 @@ gboolean CreateProcess (const gunichar2 *appname, const gunichar2 *cmdline,
 	GError *gerr = NULL;
 	int in_fd, out_fd, err_fd;
 	pid_t pid = 0;
-	int thr_ret;
 	int startup_pipe [2] = {-1, -1};
 	int dummy;
 	struct MonoProcess *mono_process;
@@ -882,8 +922,8 @@ gboolean CreateProcess (const gunichar2 *appname, const gunichar2 *cmdline,
 
 	process_set_defaults (&process_handle);
 	
-	handle = _wapi_handle_new (WAPI_HANDLE_PROCESS, &process_handle);
-	if (handle == _WAPI_HANDLE_INVALID) {
+	handle = mono_w32handle_new (MONO_W32HANDLE_PROCESS, &process_handle);
+	if (handle == INVALID_HANDLE_VALUE) {
 		g_warning ("%s: error creating process handle", __func__);
 
 		ret = FALSE;
@@ -958,19 +998,14 @@ gboolean CreateProcess (const gunichar2 *appname, const gunichar2 *cmdline,
 		MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s: new process startup not synchronized. We may not notice if the newly created process exits immediately.", __func__);
 	}
 
-	thr_ret = _wapi_handle_lock_shared_handles ();
-	g_assert (thr_ret == 0);
-	
-	pid = fork ();
-	if (pid == -1) {
-		/* Error */
+	switch (pid = fork ()) {
+	case -1: /* Error */ {
 		SetLastError (ERROR_OUTOFMEMORY);
 		ret = FALSE;
 		fork_failed = TRUE;
-		goto cleanup;
-	} else if (pid == 0) {
-		/* Child */
-		
+		break;
+	}
+	case 0: /* Child */ {
 		if (startup_pipe [0] != -1) {
 			/* Wait until the parent has updated it's internal data */
 			ssize_t _i G_GNUC_UNUSED = read (startup_pipe [0], &dummy, 1);
@@ -991,7 +1026,7 @@ gboolean CreateProcess (const gunichar2 *appname, const gunichar2 *cmdline,
 		}
 		
 		/* Close all file descriptors */
-		for (i = wapi_getdtablesize () - 1; i > 2; i--)
+		for (i = mono_w32handle_fd_reserve - 1; i > 2; i--)
 			close (i);
 
 #ifdef DEBUG_ENABLED
@@ -1015,59 +1050,51 @@ gboolean CreateProcess (const gunichar2 *appname, const gunichar2 *cmdline,
 		
 		/* set error */
 		_exit (-1);
+
+		break;
 	}
-	/* parent */
-	
-	process_handle_data = lookup_process_handle (handle);
-	if (!process_handle_data) {
-		g_warning ("%s: error looking up process handle %p", __func__,
-			   handle);
-		_wapi_handle_unref (handle);
-		goto cleanup;
+	default: /* Parent */ {
+		process_handle_data = lookup_process_handle (handle);
+		if (!process_handle_data) {
+			g_warning ("%s: error looking up process handle %p", __func__, handle);
+			mono_w32handle_unref (handle);
+		} else {
+			process_handle_data->id = pid;
+
+			/* Add our mono_process into the linked list of mono_processes */
+			mono_process = (struct MonoProcess *) g_malloc0 (sizeof (struct MonoProcess));
+			mono_process->pid = pid;
+			mono_process->handle_count = 1;
+			mono_os_sem_init (&mono_process->exit_sem, 0);
+
+			/* Keep the process handle artificially alive until the process
+			 * exits so that the information in the handle isn't lost. */
+			mono_w32handle_ref (handle);
+			mono_process->handle = handle;
+
+			process_handle_data->mono_process = mono_process;
+
+			mono_os_mutex_lock (&mono_processes_mutex);
+			mono_process->next = mono_processes;
+			mono_processes = mono_process;
+			mono_os_mutex_unlock (&mono_processes_mutex);
+
+			if (process_info != NULL) {
+				process_info->hProcess = handle;
+				process_info->dwProcessId = pid;
+
+				/* FIXME: we might need to handle the thread info some day */
+				process_info->hThread = INVALID_HANDLE_VALUE;
+				process_info->dwThreadId = 0;
+			}
+		}
+
+		break;
 	}
-	
-	process_handle_data->id = pid;
-
-	/* Add our mono_process into the linked list of mono_processes */
-	mono_process = (struct MonoProcess *) g_malloc0 (sizeof (struct MonoProcess));
-	mono_process->pid = pid;
-	mono_process->handle_count = 1;
-	if (mono_os_sem_init (&mono_process->exit_sem, 0) != 0) {
-		/* If we can't create the exit semaphore, we just don't add anything
-		 * to our list of mono processes. Waiting on the process will return 
-		 * immediately. */
-		g_warning ("%s: could not create exit semaphore for process.", strerror (errno));
-		g_free (mono_process);
-	} else {
-		/* Keep the process handle artificially alive until the process
-		 * exits so that the information in the handle isn't lost. */
-		_wapi_handle_ref (handle);
-		mono_process->handle = handle;
-
-		process_handle_data->mono_process = mono_process;
-
-		mono_os_mutex_lock (&mono_processes_mutex);
-		mono_process->next = mono_processes;
-		mono_processes = mono_process;
-		mono_os_mutex_unlock (&mono_processes_mutex);
 	}
-	
-	if (process_info != NULL) {
-		process_info->hProcess = handle;
-		process_info->dwProcessId = pid;
-
-		/* FIXME: we might need to handle the thread info some
-		 * day
-		 */
-		process_info->hThread = INVALID_HANDLE_VALUE;
-		process_info->dwThreadId = 0;
-	}
-
-cleanup:
-	_wapi_handle_unlock_shared_handles ();
 
 	if (fork_failed)
-		_wapi_handle_unref (handle);
+		mono_w32handle_unref (handle);
 
 	if (startup_pipe [1] != -1) {
 		/* Write 1 byte, doesn't matter what */
@@ -1125,20 +1152,22 @@ process_set_name (WapiHandle_process *process_handle)
 }
 
 void
-wapi_processes_init (void)
+_wapi_processes_init (void)
 {
-	pid_t pid = _wapi_getpid ();
+	pid_t pid = wapi_getpid ();
 	WapiHandle_process process_handle = {0};
 
-	_wapi_handle_register_capabilities (WAPI_HANDLE_PROCESS,
-		(WapiHandleCapability)(WAPI_HANDLE_CAP_WAIT | WAPI_HANDLE_CAP_SPECIAL_WAIT));
+	mono_w32handle_register_ops (MONO_W32HANDLE_PROCESS, &_wapi_process_ops);
+
+	mono_w32handle_register_capabilities (MONO_W32HANDLE_PROCESS,
+		(MonoW32HandleCapability)(MONO_W32HANDLE_CAP_WAIT | MONO_W32HANDLE_CAP_SPECIAL_WAIT));
 	
 	process_handle.id = pid;
 
 	process_set_defaults (&process_handle);
 	process_set_name (&process_handle);
 
-	current_process = _wapi_handle_new (WAPI_HANDLE_PROCESS,
+	current_process = mono_w32handle_new (MONO_W32HANDLE_PROCESS,
 					    &process_handle);
 	g_assert (current_process);
 
@@ -1148,7 +1177,7 @@ wapi_processes_init (void)
 gpointer
 _wapi_process_duplicate (void)
 {
-	_wapi_handle_ref (current_process);
+	mono_w32handle_ref (current_process);
 	
 	return current_process;
 }
@@ -1204,9 +1233,9 @@ process_open_compare (gpointer handle, gpointer user_data)
 	 * unsignalled
 	 */
 	if (checking_pid == wanted_pid &&
-	    !_wapi_handle_issignalled (handle)) {
+	    !mono_w32handle_issignalled (handle)) {
 		/* If the handle is blown away in the window between
-		 * returning TRUE here and _wapi_search_handle pinging
+		 * returning TRUE here and mono_w32handle_search pinging
 		 * the timestamp, the search will continue
 		 */
 		return TRUE;
@@ -1234,7 +1263,7 @@ OpenProcess (guint32 req_access G_GNUC_UNUSED, gboolean inherit G_GNUC_UNUSED, g
 	
 	MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s: looking for process %d", __func__, pid);
 
-	handle = _wapi_search_handle (WAPI_HANDLE_PROCESS,
+	handle = mono_w32handle_search (MONO_W32HANDLE_PROCESS,
 				      process_open_compare,
 				      GUINT_TO_POINTER (pid), NULL, TRUE);
 	if (handle == 0) {
@@ -1252,7 +1281,7 @@ OpenProcess (guint32 req_access G_GNUC_UNUSED, gboolean inherit G_GNUC_UNUSED, g
 		}
 	}
 
-	/* _wapi_search_handle () already added a ref */
+	/* mono_w32handle_search () already added a ref */
 	return handle;
 }
 
@@ -1261,6 +1290,7 @@ GetExitCodeProcess (gpointer process, guint32 *code)
 {
 	WapiHandle_process *process_handle;
 	guint32 pid = -1;
+	gboolean alerted;
 	
 	if (!code)
 		return FALSE;
@@ -1274,7 +1304,8 @@ GetExitCodeProcess (gpointer process, guint32 *code)
 			*code = STILL_ACTIVE;
 			return TRUE;
 		} else {
-			return FALSE;
+			*code = -1;
+			return TRUE;
 		}
 	}
 
@@ -1285,7 +1316,7 @@ GetExitCodeProcess (gpointer process, guint32 *code)
 		return FALSE;
 	}
 
-	if (process_handle->id == _wapi_getpid ()) {
+	if (process_handle->id == wapi_getpid ()) {
 		*code = STILL_ACTIVE;
 		return TRUE;
 	}
@@ -1296,9 +1327,9 @@ GetExitCodeProcess (gpointer process, guint32 *code)
 	/* Make sure any process exit has been noticed, before
 	 * checking if the process is signalled.  Fixes bug 325463.
 	 */
-	process_wait (process, 0, TRUE);
+	process_wait (process, 0, &alerted);
 	
-	if (_wapi_handle_issignalled (process))
+	if (mono_w32handle_issignalled (process))
 		*code = process_handle->exitstatus;
 	else
 		*code = STILL_ACTIVE;
@@ -1344,7 +1375,7 @@ GetProcessTimes (gpointer process, WapiFileTime *create_time,
 	/* A process handle is only signalled if the process has
 	 * exited.  Otherwise exit_time isn't set
 	 */
-	if (_wapi_handle_issignalled (process))
+	if (mono_w32handle_issignalled (process))
 		*exit_time = process_handle->exit_time;
 
 #ifdef HAVE_GETRUSAGE
@@ -1403,7 +1434,7 @@ static gint find_procmodule (gconstpointer a, gconstpointer b)
 	}
 }
 
-#ifdef PLATFORM_MACOSX
+#if defined(USE_OSX_LOADER)
 #include <mach-o/dyld.h>
 #include <mach-o/getsect.h>
 
@@ -1458,7 +1489,7 @@ static GSList *load_modules (void)
 	
 	return(ret);
 }
-#elif defined(__OpenBSD__) || defined(__FreeBSD__)
+#elif defined(USE_BSD_LOADER)
 #include <link.h>
 static int load_modules_callback (struct dl_phdr_info *info, size_t size, void *ptr)
 {
@@ -1466,7 +1497,7 @@ static int load_modules_callback (struct dl_phdr_info *info, size_t size, void *
 	    + sizeof (info->dlpi_phnum))
 		return (-1);
 
-	struct dl_phdr_info *cpy = calloc(1, sizeof(struct dl_phdr_info));
+	struct dl_phdr_info *cpy = g_calloc (1, sizeof(struct dl_phdr_info));
 	if (!cpy)
 		return (-1);
 
@@ -1502,7 +1533,7 @@ static GSList *load_modules (void)
 		MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s: inode=%d, filename=%s, address_start=%p, address_end=%p", __func__,
 				   mod->inode, mod->filename, mod->address_start, mod->address_end);
 
-		free(info);
+		g_free (info);
 
 		if (g_slist_find_custom (ret, mod, find_procmodule) == NULL) {
 			ret = g_slist_prepend (ret, mod);
@@ -1517,7 +1548,7 @@ static GSList *load_modules (void)
 
 	return(ret);
 }
-#elif defined(__HAIKU__)
+#elif defined(USE_HAIKU_LOADER)
 
 static GSList *load_modules (void)
 {
@@ -1713,7 +1744,7 @@ static gboolean match_procname_to_modulename (char *procname, char *modulename)
 	return result;
 }
 
-#if !(defined(PLATFORM_MACOSX) || defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__HAIKU__))
+#if !(defined(USE_OSX_LOADER) || defined(USE_BSD_LOADER) || defined(USE_HAIKU_LOADER))
 static FILE *
 open_process_map (int pid, const char *mode)
 {
@@ -1742,7 +1773,7 @@ gboolean EnumProcessModules (gpointer process, gpointer *modules,
 			     guint32 size, guint32 *needed)
 {
 	WapiHandle_process *process_handle;
-#if !defined(__OpenBSD__) && !defined(PLATFORM_MACOSX) && !defined(__FreeBSD__)
+#if !defined(USE_OSX_LOADER) && !defined(USE_BSD_LOADER)
 	FILE *fp;
 #endif
 	GSList *mods = NULL;
@@ -1778,7 +1809,7 @@ gboolean EnumProcessModules (gpointer process, gpointer *modules,
 		proc_name = g_strdup (process_handle->proc_name);
 	}
 	
-#if defined(PLATFORM_MACOSX) || defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__HAIKU__)
+#if defined(USE_OSX_LOADER) || defined(USE_BSD_LOADER) || defined(USE_HAIKU_LOADER)
 	mods = load_modules ();
 	if (!proc_name) {
 		modules[0] = NULL;
@@ -1835,11 +1866,11 @@ gboolean EnumProcessModules (gpointer process, gpointer *modules,
 static char *
 get_process_name_from_proc (pid_t pid)
 {
-#if defined(__OpenBSD__) || defined(__FreeBSD__)
+#if defined(USE_BSD_LOADER)
 	int mib [6];
 	size_t size;
 	struct kinfo_proc *pi;
-#elif defined(PLATFORM_MACOSX)
+#elif defined(USE_OSX_LOADER)
 #if !(!defined (__mono_ppc__) && defined (TARGET_OSX))
 	size_t size;
 	struct kinfo_proc *pi;
@@ -1866,7 +1897,7 @@ get_process_name_from_proc (pid_t pid)
 		fclose (fp);
 	}
 	g_free (filename);
-#elif defined(PLATFORM_MACOSX)
+#elif defined(USE_OSX_LOADER)
 #if !defined (__mono_ppc__) && defined (TARGET_OSX)
 	/* No proc name on OSX < 10.5 nor ppc nor iOS */
 	memset (buf, '\0', sizeof(buf));
@@ -1901,12 +1932,12 @@ get_process_name_from_proc (pid_t pid)
 	if (sysctl(mib, 4, NULL, &size, NULL, 0) < 0)
 		return(ret);
 
-	if ((pi = malloc(size)) == NULL)
+	if ((pi = g_malloc (size)) == NULL)
 		return(ret);
 
 	if (sysctl (mib, 4, pi, &size, NULL, 0) < 0) {
 		if (errno == ENOMEM) {
-			free(pi);
+			g_free (pi);
 			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s: Didn't allocate enough memory for kproc info", __func__);
 		}
 		return(ret);
@@ -1915,9 +1946,10 @@ get_process_name_from_proc (pid_t pid)
 	if (strlen (pi->kp_proc.p_comm) > 0)
 		ret = g_strdup (pi->kp_proc.p_comm);
 
-	free(pi);
+	g_free (pi);
 #endif
-#elif defined(__FreeBSD__)
+#elif defined(USE_BSD_LOADER)
+#if defined(__FreeBSD__)
 	mib [0] = CTL_KERN;
 	mib [1] = KERN_PROC;
 	mib [2] = KERN_PROC_PID;
@@ -1927,12 +1959,12 @@ get_process_name_from_proc (pid_t pid)
 		return(ret);
 	}
 
-	if ((pi = malloc(size)) == NULL)
+	if ((pi = g_malloc (size)) == NULL)
 		return(ret);
 
 	if (sysctl (mib, 4, pi, &size, NULL, 0) < 0) {
 		if (errno == ENOMEM) {
-			free(pi);
+			g_free (pi);
 			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s: Didn't allocate enough memory for kproc info", __func__);
 		}
 		return(ret);
@@ -1940,7 +1972,7 @@ get_process_name_from_proc (pid_t pid)
 
 	if (strlen (pi->ki_comm) > 0)
 		ret = g_strdup (pi->ki_comm);
-	free(pi);
+	g_free (pi);
 #elif defined(__OpenBSD__)
 	mib [0] = CTL_KERN;
 	mib [1] = KERN_PROC;
@@ -1955,7 +1987,7 @@ retry:
 		return(ret);
 	}
 
-	if ((pi = malloc(size)) == NULL)
+	if ((pi = g_malloc (size)) == NULL)
 		return(ret);
 
 	mib[5] = (int)(size / sizeof(struct kinfo_proc));
@@ -1963,22 +1995,18 @@ retry:
 	if ((sysctl (mib, 6, pi, &size, NULL, 0) < 0) ||
 		(size != sizeof (struct kinfo_proc))) {
 		if (errno == ENOMEM) {
-			free(pi);
+			g_free (pi);
 			goto retry;
 		}
 		return(ret);
 	}
 
-#if defined(__OpenBSD__)
 	if (strlen (pi->p_comm) > 0)
 		ret = g_strdup (pi->p_comm);
-#elif defined(__FreeBSD__)
-	if (strlen (pi->ki_comm) > 0)
-		ret = g_strdup (pi->ki_comm);
-#endif
 
-	free(pi);
-#elif defined(__HAIKU__)
+	g_free (pi);
+#endif
+#elif defined(USE_HAIKU_LOADER)
 	image_info imageInfo;
 	int32 cookie = 0;
 
@@ -2082,7 +2110,7 @@ get_module_name (gpointer process, gpointer module,
 	char *procname_ext = NULL;
 	glong len;
 	gsize bytes;
-#if !defined(__OpenBSD__) && !defined(PLATFORM_MACOSX) && !defined(__FreeBSD__)
+#if !defined(USE_OSX_LOADER) && !defined(USE_BSD_LOADER)
 	FILE *fp;
 #endif
 	GSList *mods = NULL;
@@ -2116,7 +2144,7 @@ get_module_name (gpointer process, gpointer module,
 	}
 
 	/* Look up the address in /proc/<pid>/maps */
-#if defined(PLATFORM_MACOSX) || defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__HAIKU__)
+#if defined(USE_OSX_LOADER) || defined(USE_BSD_LOADER) || defined(USE_HAIKU_LOADER)
 	mods = load_modules ();
 #else
 	fp = open_process_map (pid, "r");
@@ -2269,7 +2297,7 @@ GetModuleInformation (gpointer process, gpointer module,
 {
 	WapiHandle_process *process_handle;
 	pid_t pid;
-#if !defined(__OpenBSD__) && !defined(PLATFORM_MACOSX) && !defined(__FreeBSD__)
+#if !defined(USE_OSX_LOADER) && !defined(USE_BSD_LOADER)
 	FILE *fp;
 #endif
 	GSList *mods = NULL;
@@ -2300,7 +2328,7 @@ GetModuleInformation (gpointer process, gpointer module,
 		proc_name = g_strdup (process_handle->proc_name);
 	}
 
-#if defined(PLATFORM_MACOSX) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__HAIKU__)
+#if defined(USE_OSX_LOADER) || defined(USE_BSD_LOADER) || defined(USE_HAIKU_LOADER)
 	mods = load_modules ();
 #else
 	/* Look up the address in /proc/<pid>/maps */
@@ -2585,7 +2613,7 @@ mono_processes_cleanup (void)
 			mp->handle = NULL;
 			mono_os_mutex_unlock (&mono_processes_mutex);
 			if (unref_handle)
-				_wapi_handle_unref (unref_handle);
+				mono_w32handle_unref (unref_handle);
 		}
 	}
 
@@ -2654,6 +2682,23 @@ process_close (gpointer handle, gpointer data)
 	mono_processes_cleanup ();
 }
 
+static void process_details (gpointer data)
+{
+	WapiHandle_process *process_handle = (WapiHandle_process *) data;
+	g_print ("id: %d, exited: %s, exitstatus: %d",
+		process_handle->id, process_handle->exited ? "true" : "false", process_handle->exitstatus);
+}
+
+static const gchar* process_typename (void)
+{
+	return "Process";
+}
+
+static gsize process_typesize (void)
+{
+	return sizeof (WapiHandle_process);
+}
+
 #if HAVE_SIGACTION
 MONO_SIGNAL_HANDLER_FUNC (static, mono_sigchld_signal_handler, (int _dummy, siginfo_t *info, void *context))
 {
@@ -2661,8 +2706,6 @@ MONO_SIGNAL_HANDLER_FUNC (static, mono_sigchld_signal_handler, (int _dummy, sigi
 	int pid;
 	struct MonoProcess *p;
 
-	MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "SIG CHILD handler for pid: %i\n", info->si_pid);
-
 	do {
 		do {
 			pid = waitpid (-1, &status, WNOHANG);
@@ -2671,8 +2714,6 @@ MONO_SIGNAL_HANDLER_FUNC (static, mono_sigchld_signal_handler, (int _dummy, sigi
 		if (pid <= 0)
 			break;
 
-		MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "child ended: %i", pid);
-
 		/*
 		 * This can run concurrently with the code in the rest of this module.
 		 */
@@ -2690,8 +2731,6 @@ MONO_SIGNAL_HANDLER_FUNC (static, mono_sigchld_signal_handler, (int _dummy, sigi
 			p->freeable = TRUE;
 		}
 	} while (1);
-
-	MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "SIG CHILD handler: done looping.");
 }
 
 #endif
@@ -2711,13 +2750,12 @@ process_add_sigchld_handler (void)
 }
 
 static guint32
-process_wait (gpointer handle, guint32 timeout, gboolean alertable)
+process_wait (gpointer handle, guint32 timeout, gboolean *alerted)
 {
 	WapiHandle_process *process_handle;
 	pid_t pid G_GNUC_UNUSED, ret;
 	int status;
-	guint32 start;
-	guint32 now;
+	gint64 start, now;
 	struct MonoProcess *mp;
 
 	/* FIXME: We can now easily wait on processes that aren't our own children,
@@ -2726,6 +2764,9 @@ process_wait (gpointer handle, guint32 timeout, gboolean alertable)
 
 	MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u)", __func__, handle, timeout);
 
+	if (alerted)
+		*alerted = FALSE;
+
 	process_handle = lookup_process_handle (handle);
 	if (!process_handle) {
 		g_warning ("%s: error looking up process handle %p", __func__, handle);
@@ -2745,7 +2786,34 @@ process_wait (gpointer handle, guint32 timeout, gboolean alertable)
 	/* We don't need to lock mono_processes here, the entry
 	 * has a handle_count > 0 which means it will not be freed. */
 	mp = process_handle->mono_process;
-	g_assert (mp);
+	if (!mp) {
+		pid_t res;
+
+		if (pid == mono_process_current_pid ()) {
+			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): waiting on current process", __func__, handle, timeout);
+			return WAIT_TIMEOUT;
+		}
+
+		/* This path is used when calling Process.HasExited, so
+		 * it is only used to poll the state of the process, not
+		 * to actually wait on it to exit */
+		g_assert (timeout == 0);
+
+		MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): waiting on non-child process", __func__, handle, timeout);
+
+		res = waitpid (pid, &status, WNOHANG);
+		if (res == 0) {
+			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): non-child process WAIT_TIMEOUT", __func__, handle, timeout);
+			return WAIT_TIMEOUT;
+		}
+		if (res > 0) {
+			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): non-child process waited successfully", __func__, handle, timeout);
+			return WAIT_OBJECT_0;
+		}
+
+		MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): non-child process WAIT_FAILED, error : %s (%d))", __func__, handle, timeout, g_strerror (errno), errno);
+		return WAIT_FAILED;
+	}
 
 	start = mono_msec_ticks ();
 	now = start;
@@ -2754,26 +2822,20 @@ process_wait (gpointer handle, guint32 timeout, gboolean alertable)
 		if (timeout != INFINITE) {
 			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): waiting on semaphore for %li ms...", 
 				   __func__, handle, timeout, (timeout - (now - start)));
-			ret = mono_os_sem_timedwait (&mp->exit_sem, (timeout - (now - start)), alertable ? MONO_SEM_FLAGS_ALERTABLE : MONO_SEM_FLAGS_NONE);
+			ret = mono_os_sem_timedwait (&mp->exit_sem, (timeout - (now - start)), alerted ? MONO_SEM_FLAGS_ALERTABLE : MONO_SEM_FLAGS_NONE);
 		} else {
 			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): waiting on semaphore forever...", 
 				   __func__, handle, timeout);
-			ret = mono_os_sem_wait (&mp->exit_sem, alertable ? MONO_SEM_FLAGS_ALERTABLE : MONO_SEM_FLAGS_NONE);
+			ret = mono_os_sem_wait (&mp->exit_sem, alerted ? MONO_SEM_FLAGS_ALERTABLE : MONO_SEM_FLAGS_NONE);
 		}
 
-		if (ret == -1 && errno != EINTR && errno != ETIMEDOUT) {
-			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): sem_timedwait failure: %s", 
-				   __func__, handle, timeout, g_strerror (errno));
-			/* Should we return a failure here? */
-		}
-
-		if (ret == 0) {
+		if (ret == MONO_SEM_TIMEDWAIT_RET_SUCCESS) {
 			/* Success, process has exited */
 			mono_os_sem_post (&mp->exit_sem);
 			break;
 		}
 
-		if (timeout == 0) {
+		if (ret == MONO_SEM_TIMEDWAIT_RET_TIMEDOUT) {
 			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): WAIT_TIMEOUT (timeout = 0)", __func__, handle, timeout);
 			return WAIT_TIMEOUT;
 		}
@@ -2784,8 +2846,9 @@ process_wait (gpointer handle, guint32 timeout, gboolean alertable)
 			return WAIT_TIMEOUT;
 		}
 		
-		if (alertable && _wapi_thread_cur_apc_pending ()) {
+		if (alerted && ret == MONO_SEM_TIMEDWAIT_RET_ALERTED) {
 			MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): WAIT_IO_COMPLETION", __func__, handle, timeout);
+			*alerted = TRUE;
 			return WAIT_IO_COMPLETION;
 		}
 	}
@@ -2793,9 +2856,6 @@ process_wait (gpointer handle, guint32 timeout, gboolean alertable)
 	/* Process must have exited */
 	MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): Waited successfully", __func__, handle, timeout);
 
-	ret = _wapi_handle_lock_shared_handles ();
-	g_assert (ret == 0);
-
 	status = mp ? mp->status : 0;
 	if (WIFSIGNALED (status))
 		process_handle->exitstatus = 128 + WTERMSIG (status);
@@ -2808,9 +2868,7 @@ process_wait (gpointer handle, guint32 timeout, gboolean alertable)
 	MONO_TRACE (G_LOG_LEVEL_DEBUG, MONO_TRACE_IO_LAYER, "%s (%p, %u): Setting pid %d signalled, exit status %d",
 		   __func__, handle, timeout, process_handle->id, process_handle->exitstatus);
 
-	_wapi_handle_set_signal_state (handle, TRUE, TRUE);
-
-	_wapi_handle_unlock_shared_handles ();
+	mono_w32handle_set_signal_state (handle, TRUE, TRUE);
 
 	return WAIT_OBJECT_0;
 }
