@@ -50,9 +50,6 @@ typedef WapiSecurityAttributes SECURITY_ATTRIBUTES;
 typedef WapiSecurityAttributes *LPSECURITY_ATTRIBUTES;
 typedef WapiOverlapped *LPOVERLAPPED;
 typedef WapiOverlappedCB LPOVERLAPPED_COMPLETION_ROUTINE;
-typedef WapiThreadStart LPTHREAD_START_ROUTINE;
-typedef WapiCriticalSection CRITICAL_SECTION;
-typedef WapiCriticalSection *LPCRITICAL_SECTION;
 typedef WapiFileTime FILETIME;
 typedef WapiFileTime *LPFILETIME;
 typedef WapiSystemTime SYSTEMTIME;
@@ -63,8 +60,8 @@ typedef WapiWSABuf WSABUF;
 typedef WapiWSABuf *LPWSABUF;
 typedef WapiLargeInteger LARGE_INTEGER;
 typedef WapiLargeInteger *PLARGE_INTEGER;
-typedef WapiSystemInfo SYSTEM_INFO;
-typedef WapiSystemInfo *LPSYSTEM_INFO;
+typedef WapiULargeInteger ULARGE_INTEGER;
+typedef WapiULargeInteger *PULARGE_INTEGER;
 typedef WapiFloatingSaveArea FLOATING_SAVE_AREA;
 typedef WapiFloatingSaveArea *PFLOATING_SAVE_AREA;
 typedef WapiContext CONTEXT;
@@ -105,7 +102,6 @@ typedef WapiImageResourceDirectoryEntry IMAGE_RESOURCE_DIRECTORY_ENTRY;
 typedef WapiImageResourceDirectoryEntry *PIMAGE_RESOURCE_DIRECTORY_ENTRY;
 typedef WapiImageResourceDataEntry IMAGE_RESOURCE_DATA_ENTRY;
 typedef WapiImageResourceDataEntry *PIMAGE_RESOURCE_DATA_ENTRY;
-typedef WapiApcProc PAPCFUNC;
 typedef WapiShellExecuteInfo SHELLEXECUTEINFO;
 typedef WapiShellExecuteInfo *LPSHELLEXECUTEINFO;
 typedef WapiTransmitFileBuffers TRANSMIT_FILE_BUFFERS;
