@@ -525,6 +525,7 @@ mono_init_internal (const char *filename, const char *exe_filename, const char *
 #endif
 
 #ifndef HOST_WIN32
+	mono_w32handle_init ();
 	wapi_init ();
 #endif
 
@@ -1166,6 +1167,12 @@ mono_domain_free (MonoDomain *domain, gboolean force)
 	g_slist_free (domain->domain_assemblies);
 	domain->domain_assemblies = NULL;
 
+	/* 
+	 * Send this after the assemblies have been unloaded and the domain is still in a 
+	 * usable state.
+	 */
+	mono_profiler_appdomain_event (domain, MONO_PROFILE_END_UNLOAD);
+
 	if (free_domain_hook)
 		free_domain_hook (domain);
 
@@ -1239,10 +1246,6 @@ mono_domain_free (MonoDomain *domain, gboolean force)
 		g_hash_table_destroy (domain->generic_virtual_cases);
 		domain->generic_virtual_cases = NULL;
 	}
-	if (domain->generic_virtual_thunks) {
-		g_hash_table_destroy (domain->generic_virtual_thunks);
-		domain->generic_virtual_thunks = NULL;
-	}
 	if (domain->ftnptrs_hash) {
 		g_hash_table_destroy (domain->ftnptrs_hash);
 		domain->ftnptrs_hash = NULL;
@@ -1483,8 +1486,10 @@ mono_context_get_domain_id (MonoAppContext *context)
 void
 mono_domain_add_class_static_data (MonoDomain *domain, MonoClass *klass, gpointer data, guint32 *bitmap)
 {
-	/* The first entry in the array is the index of the next free slot
-	 * and the total size of the array
+	/* Note [Domain Static Data Array]:
+	 *
+	 * Entry 0 in the array is the index of the next free slot.
+	 * Entry 1 is the total size of the array.
 	 */
 	int next;
 	if (domain->static_data_array) {
@@ -2047,3 +2052,24 @@ mono_domain_unlock (MonoDomain *domain)
 {
 	mono_locks_coop_release (&domain->lock, DomainLock);
 }
+
+GPtrArray*
+mono_domain_get_assemblies (MonoDomain *domain, gboolean refonly)
+{
+	GSList *tmp;
+	GPtrArray *assemblies;
+	MonoAssembly *ass;
+
+	assemblies = g_ptr_array_new ();
+	mono_domain_assemblies_lock (domain);
+	for (tmp = domain->domain_assemblies; tmp; tmp = tmp->next) {
+		ass = (MonoAssembly *)tmp->data;
+		if (refonly != ass->ref_only)
+			continue;
+		if (ass->corlib_internal)
+			continue;
+		g_ptr_array_add (assemblies, ass);
+	}
+	mono_domain_assemblies_unlock (domain);
+	return assemblies;
+}
