@@ -6,6 +6,7 @@
  *
  * Copyright 2003 Ximian, Inc (http://www.ximian.com)
  * Copyright 2004-2009 Novell, Inc (http://www.novell.com)
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 
 #include <config.h>
@@ -356,7 +357,8 @@ mon_finalize (MonoThreadsSync *mon)
 	LOCK_DEBUG (g_message ("%s: Finalizing sync %p", __func__, mon));
 
 	if (mon->entry_sem != NULL) {
-		CloseHandle (mon->entry_sem);
+		mono_coop_sem_destroy (mon->entry_sem);
+		g_free (mon->entry_sem);
 		mon->entry_sem = NULL;
 	}
 	/* If this isn't empty then something is seriously broken - it
@@ -392,7 +394,7 @@ mon_new (gsize id)
 					if (new_->wait_list) {
 						/* Orphaned events left by aborted threads */
 						while (new_->wait_list) {
-							LOCK_DEBUG (g_message (G_GNUC_PRETTY_FUNCTION ": (%d): Closing orphaned event %d", mono_thread_info_get_small_id (), new->wait_list->data));
+							LOCK_DEBUG (g_message (G_GNUC_PRETTY_FUNCTION ": (%d): Closing orphaned event %d", mono_thread_info_get_small_id (), new_->wait_list->data));
 							CloseHandle (new_->wait_list->data);
 							new_->wait_list = g_slist_remove (new_->wait_list, new_->wait_list->data);
 						}
@@ -628,18 +630,19 @@ mono_object_hash (MonoObject* obj)
 #endif
 }
 
-static void
+static gboolean
 mono_monitor_ensure_owned (LockWord lw, guint32 id)
 {
 	if (lock_word_is_flat (lw)) {
 		if (lock_word_get_owner (lw) == id)
-			return;
+			return TRUE;
 	} else if (lock_word_is_inflated (lw)) {
 		if (mon_status_get_owner (lock_word_get_inflated_lock (lw)->status) == id)
-			return;
+			return TRUE;
 	}
 
 	mono_set_pending_exception (mono_get_exception_synchronization_lock ("Object synchronization method was called from an unsynchronized block of code."));
+	return FALSE;
 }
 
 /*
@@ -678,7 +681,7 @@ mono_monitor_exit_inflated (MonoObject *obj)
 			tmp_status = InterlockedCompareExchange ((gint32*)&mon->status, new_status, old_status);
 			if (tmp_status == old_status) {
 				if (have_waiters)
-					ReleaseSemaphore (mon->entry_sem, 1, NULL);
+					mono_coop_sem_post (mon->entry_sem);
 				break;
 			}
 			old_status = tmp_status;
@@ -742,10 +745,10 @@ mono_monitor_try_enter_inflated (MonoObject *obj, guint32 ms, gboolean allow_int
 	LockWord lw;
 	MonoThreadsSync *mon;
 	HANDLE sem;
-	guint32 then = 0, now, delta;
+	gint64 then = 0, now, delta;
 	guint32 waitms;
-	guint32 ret;
 	guint32 new_status, old_status, tmp_status;
+	MonoSemTimedwaitRet wait_ret;
 	MonoInternalThread *thread;
 	gboolean interrupted = FALSE;
 
@@ -837,11 +840,12 @@ retry_contended:
 	 */
 	if (mon->entry_sem == NULL) {
 		/* Create the semaphore */
-		sem = CreateSemaphore (NULL, 0, 0x7fffffff, NULL);
-		g_assert (sem != NULL);
+		sem = g_new0 (MonoCoopSem, 1);
+		mono_coop_sem_init (sem, 0);
 		if (InterlockedCompareExchangePointer ((gpointer*)&mon->entry_sem, sem, NULL) != NULL) {
 			/* Someone else just put a handle here */
-			CloseHandle (sem);
+			mono_coop_sem_destroy (sem);
+			g_free (sem);
 		}
 	}
 
@@ -877,12 +881,10 @@ retry_contended:
 	mono_thread_set_state (thread, ThreadState_WaitSleepJoin);
 
 	/*
-	 * We pass TRUE instead of allow_interruption since we have to check for the
+	 * We pass ALERTABLE instead of allow_interruption since we have to check for the
 	 * StopRequested case below.
 	 */
-	MONO_PREPARE_BLOCKING;
-	ret = WaitForSingleObjectEx (mon->entry_sem, waitms, TRUE);
-	MONO_FINISH_BLOCKING;
+	wait_ret = mono_coop_sem_timedwait (mon->entry_sem, waitms, MONO_SEM_FLAGS_ALERTABLE);
 
 	mono_thread_clr_state (thread, ThreadState_WaitSleepJoin);
 	
@@ -890,23 +892,18 @@ retry_contended:
 	mono_perfcounters->thread_queue_len--;
 #endif
 
-	if (ret == WAIT_IO_COMPLETION && !allow_interruption) {
+	if (wait_ret == MONO_SEM_TIMEDWAIT_RET_ALERTED && !allow_interruption) {
 		interrupted = TRUE;
 		/* 
 		 * We have to obey a stop/suspend request even if 
 		 * allow_interruption is FALSE to avoid hangs at shutdown.
 		 */
-		if (!mono_thread_test_state (mono_thread_internal_current (), (MonoThreadState)(ThreadState_StopRequested | ThreadState_SuspendRequested))) {
+		if (!mono_thread_test_state (mono_thread_internal_current (), (MonoThreadState)(ThreadState_StopRequested | ThreadState_SuspendRequested | ThreadState_AbortRequested))) {
 			if (ms != INFINITE) {
 				now = mono_msec_ticks ();
-				if (now < then) {
-					LOCK_DEBUG (g_message ("%s: wrapped around! now=0x%x then=0x%x", __func__, now, then));
 
-					now += (0xffffffff - then);
-					then = 0;
-
-					LOCK_DEBUG (g_message ("%s: wrap rejig: now=0x%x then=0x%x delta=0x%x", __func__, now, then, now-then));
-				}
+				/* it should not overflow before ~30k years */
+				g_assert (now >= then);
 
 				delta = now - then;
 				if (delta >= ms) {
@@ -918,11 +915,11 @@ retry_contended:
 			/* retry from the top */
 			goto retry_contended;
 		}
-	} else if (ret == WAIT_OBJECT_0) {
+	} else if (wait_ret == MONO_SEM_TIMEDWAIT_RET_SUCCESS) {
 		interrupted = FALSE;
 		/* retry from the top */
 		goto retry_contended;
-	} else if (ret == WAIT_TIMEOUT) {
+	} else if (wait_ret == MONO_SEM_TIMEDWAIT_RET_TIMEDOUT) {
 		/* we're done */
 	}
 
@@ -931,10 +928,10 @@ retry_contended:
 
 	mono_profiler_monitor_event (obj, MONO_PROFILER_MONITOR_FAIL);
 
-	if (ret == WAIT_IO_COMPLETION) {
+	if (wait_ret == MONO_SEM_TIMEDWAIT_RET_ALERTED) {
 		LOCK_DEBUG (g_message ("%s: (%d) interrupted waiting, returning -1", __func__, id));
 		return -1;
-	} else if (ret == WAIT_TIMEOUT) {
+	} else if (wait_ret == MONO_SEM_TIMEDWAIT_RET_TIMEDOUT) {
 		LOCK_DEBUG (g_message ("%s: (%d) timed out waiting, returning FALSE", __func__, id));
 		return 0;
 	} else {
@@ -955,11 +952,6 @@ mono_monitor_try_enter_internal (MonoObject *obj, guint32 ms, gboolean allow_int
 
 	LOCK_DEBUG (g_message("%s: (%d) Trying to lock object %p (%d ms)", __func__, id, obj, ms));
 
-	if (G_UNLIKELY (!obj)) {
-		mono_set_pending_exception (mono_get_exception_argument_null ("obj"));
-		return FALSE;
-	}
-
 	lw.sync = obj->synchronisation;
 
 	if (G_LIKELY (lock_word_is_free (lw))) {
@@ -1005,18 +997,31 @@ mono_monitor_try_enter_internal (MonoObject *obj, guint32 ms, gboolean allow_int
 gboolean 
 mono_monitor_enter (MonoObject *obj)
 {
+	if (G_UNLIKELY (!obj)) {
+		mono_set_pending_exception (mono_get_exception_argument_null ("obj"));
+		return FALSE;
+	}
 	return mono_monitor_try_enter_internal (obj, INFINITE, FALSE) == 1;
 }
 
 gboolean 
 mono_monitor_enter_fast (MonoObject *obj)
 {
+	if (G_UNLIKELY (!obj)) {
+		/* don't set pending exn on the fast path, just return
+		 * FALSE and let the slow path take care of it. */
+		return FALSE;
+	}
 	return mono_monitor_try_enter_internal (obj, 0, FALSE) == 1;
 }
 
 gboolean
 mono_monitor_try_enter (MonoObject *obj, guint32 ms)
 {
+	if (G_UNLIKELY (!obj)) {
+		mono_set_pending_exception (mono_get_exception_argument_null ("obj"));
+		return FALSE;
+	}
 	return mono_monitor_try_enter_internal (obj, ms, FALSE) == 1;
 }
 
@@ -1034,7 +1039,8 @@ mono_monitor_exit (MonoObject *obj)
 
 	lw.sync = obj->synchronisation;
 
-	mono_monitor_ensure_owned (lw, mono_thread_info_get_small_id ());
+	if (!mono_monitor_ensure_owned (lw, mono_thread_info_get_small_id ()))
+		return;
 
 	if (G_UNLIKELY (lock_word_is_inflated (lw)))
 		mono_monitor_exit_inflated (obj);
@@ -1075,29 +1081,24 @@ mono_monitor_threads_sync_members_offset (int *status_offset, int *nest_offset)
 	*nest_offset = ENCODE_OFF_SIZE (MONO_STRUCT_OFFSET (MonoThreadsSync, nest), sizeof (ts.nest));
 }
 
-gboolean 
-ves_icall_System_Threading_Monitor_Monitor_try_enter (MonoObject *obj, guint32 ms)
-{
-	gint32 res;
-
-	do {
-		res = mono_monitor_try_enter_internal (obj, ms, TRUE);
-		if (res == -1)
-			mono_thread_interruption_checkpoint ();
-	} while (res == -1);
-	
-	return res == 1;
-}
-
 void
 ves_icall_System_Threading_Monitor_Monitor_try_enter_with_atomic_var (MonoObject *obj, guint32 ms, char *lockTaken)
 {
 	gint32 res;
+	if (G_UNLIKELY (!obj)) {
+		mono_set_pending_exception (mono_get_exception_argument_null ("obj"));
+		return;
+	}
 	do {
 		res = mono_monitor_try_enter_internal (obj, ms, TRUE);
 		/*This means we got interrupted during the wait and didn't got the monitor.*/
-		if (res == -1)
-			mono_thread_interruption_checkpoint ();
+		if (res == -1) {
+			MonoException *exc = mono_thread_interruption_checkpoint ();
+			if (exc) {
+				mono_set_pending_exception (exc);
+				return;
+			}
+		}
 	} while (res == -1);
 	/*It's safe to do it from here since interruption would happen only on the wrapper.*/
 	*lockTaken = res == 1;
@@ -1126,6 +1127,8 @@ mono_monitor_enter_v4_fast (MonoObject *obj, char *lock_taken)
 {
 	if (*lock_taken == 1)
 		return FALSE;
+	if (G_UNLIKELY (!obj))
+		return FALSE;
 	gint32 res = mono_monitor_try_enter_internal (obj, 0, TRUE);
 	*lock_taken = res == 1;
 	return res == 1;
@@ -1184,7 +1187,8 @@ ves_icall_System_Threading_Monitor_Monitor_pulse (MonoObject *obj)
 	id = mono_thread_info_get_small_id ();
 	lw.sync = obj->synchronisation;
 
-	mono_monitor_ensure_owned (lw, id);
+	if (!mono_monitor_ensure_owned (lw, id))
+		return;
 
 	if (!lock_word_is_inflated (lw)) {
 		/* No threads waiting. A wait would have inflated the lock */
@@ -1215,7 +1219,8 @@ ves_icall_System_Threading_Monitor_Monitor_pulse_all (MonoObject *obj)
 	id = mono_thread_info_get_small_id ();
 	lw.sync = obj->synchronisation;
 
-	mono_monitor_ensure_owned (lw, id);
+	if (!mono_monitor_ensure_owned (lw, id))
+		return;
 
 	if (!lock_word_is_inflated (lw)) {
 		/* No threads waiting. A wait would have inflated the lock */
@@ -1251,7 +1256,8 @@ ves_icall_System_Threading_Monitor_Monitor_wait (MonoObject *obj, guint32 ms)
 
 	lw.sync = obj->synchronisation;
 
-	mono_monitor_ensure_owned (lw, id);
+	if (!mono_monitor_ensure_owned (lw, id))
+		return FALSE;
 
 	if (!lock_word_is_inflated (lw)) {
 		mono_monitor_inflate_owned (obj, id);
@@ -1261,7 +1267,8 @@ ves_icall_System_Threading_Monitor_Monitor_wait (MonoObject *obj, guint32 ms)
 	mon = lock_word_get_inflated_lock (lw);
 
 	/* Do this WaitSleepJoin check before creating the event handle */
-	mono_thread_current_check_pending_interrupt ();
+	if (mono_thread_current_check_pending_interrupt ())
+		return FALSE;
 	
 	event = CreateEvent (NULL, FALSE, FALSE, NULL);
 	if (event == NULL) {
@@ -1271,7 +1278,11 @@ ves_icall_System_Threading_Monitor_Monitor_wait (MonoObject *obj, guint32 ms)
 	
 	LOCK_DEBUG (g_message ("%s: (%d) queuing handle %p", __func__, mono_thread_info_get_small_id (), event));
 
-	mono_thread_current_check_pending_interrupt ();
+	/* This looks superfluous */
+	if (mono_thread_current_check_pending_interrupt ()) {
+		CloseHandle (event);
+		return FALSE;
+	}
 	
 	mono_thread_set_state (thread, ThreadState_WaitSleepJoin);
 
@@ -1290,42 +1301,22 @@ ves_icall_System_Threading_Monitor_Monitor_wait (MonoObject *obj, guint32 ms)
 	 * is private to this thread.  Therefore even if the event was
 	 * signalled before we wait, we still succeed.
 	 */
-	MONO_PREPARE_BLOCKING;
+	MONO_ENTER_GC_SAFE;
 	ret = WaitForSingleObjectEx (event, ms, TRUE);
-	MONO_FINISH_BLOCKING;
+	MONO_EXIT_GC_SAFE;
 
 	/* Reset the thread state fairly early, so we don't have to worry
 	 * about the monitor error checking
 	 */
 	mono_thread_clr_state (thread, ThreadState_WaitSleepJoin);
-	
-	if (mono_thread_interruption_requested ()) {
-		/* 
-		 * Can't remove the event from wait_list, since the monitor is not locked by
-		 * us. So leave it there, mon_new () will delete it when the mon structure
-		 * is placed on the free list.
-		 * FIXME: The caller expects to hold the lock after the wait returns, but it
-		 * doesn't happen in this case:
-		 * http://connect.microsoft.com/VisualStudio/feedback/ViewFeedback.aspx?FeedbackID=97268
-		 */
-		return FALSE;
-	}
 
 	/* Regain the lock with the previous nest count */
 	do {
 		regain = mono_monitor_try_enter_inflated (obj, INFINITE, TRUE, id);
-		if (regain == -1) 
-			mono_thread_interruption_checkpoint ();
+		/* We must regain the lock before handling interruption requests */
 	} while (regain == -1);
 
-	if (regain == 0) {
-		/* Something went wrong, so throw a
-		 * SynchronizationLockException
-		 */
-		CloseHandle (event);
-		mono_set_pending_exception (mono_get_exception_synchronization_lock ("Failed to regain lock"));
-		return FALSE;
-	}
+	g_assert (regain == 1);
 
 	mon->nest = nest;
 
@@ -1335,9 +1326,9 @@ ves_icall_System_Threading_Monitor_Monitor_wait (MonoObject *obj, guint32 ms)
 		/* Poll the event again, just in case it was signalled
 		 * while we were trying to regain the monitor lock
 		 */
-		MONO_PREPARE_BLOCKING;
+		MONO_ENTER_GC_SAFE;
 		ret = WaitForSingleObjectEx (event, 0, FALSE);
-		MONO_FINISH_BLOCKING;
+		MONO_EXIT_GC_SAFE;
 	}
 
 	/* Pulse will have popped our event from the queue if it signalled
