@@ -9,18 +9,7 @@
  * Copyright 2010 Novell, Inc (http://www.novell.com)
  * Copyright (C) 2012 Xamarin Inc
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License 2.0 as published by the Free Software Foundation;
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
- *
- * You should have received a copy of the GNU Library General Public
- * License 2.0 along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 
 #include "config.h"
@@ -45,14 +34,18 @@
 gboolean
 sgen_resume_thread (SgenThreadInfo *info)
 {
-	return thread_resume (info->client_info.info.native_handle) == KERN_SUCCESS;
+	kern_return_t ret;
+	do {
+		ret = thread_resume (info->client_info.info.native_handle);
+	} while (ret == KERN_ABORTED);
+	return ret == KERN_SUCCESS;
 }
 
 gboolean
 sgen_suspend_thread (SgenThreadInfo *info)
 {
-	mach_msg_type_number_t num_state;
-	thread_state_t state;
+	mach_msg_type_number_t num_state, num_fpstate;
+	thread_state_t state, fpstate;
 	kern_return_t ret;
 	ucontext_t ctx;
 	mcontext_t mctx;
@@ -60,17 +53,22 @@ sgen_suspend_thread (SgenThreadInfo *info)
 	gpointer stack_start;
 
 	state = (thread_state_t) alloca (mono_mach_arch_get_thread_state_size ());
+	fpstate = (thread_state_t) alloca (mono_mach_arch_get_thread_fpstate_size ());
 	mctx = (mcontext_t) alloca (mono_mach_arch_get_mcontext_size ());
 
-	ret = thread_suspend (info->client_info.info.native_handle);
+	do {
+		ret = thread_suspend (info->client_info.info.native_handle);
+	} while (ret == KERN_ABORTED);
 	if (ret != KERN_SUCCESS)
 		return FALSE;
 
-	ret = mono_mach_arch_get_thread_state (info->client_info.info.native_handle, state, &num_state);
+	do {
+		ret = mono_mach_arch_get_thread_states (info->client_info.info.native_handle, state, &num_state, fpstate, &num_fpstate);
+	} while (ret == KERN_ABORTED);
 	if (ret != KERN_SUCCESS)
 		return FALSE;
 
-	mono_mach_arch_thread_state_to_mcontext (state, mctx);
+	mono_mach_arch_thread_states_to_mcontext (state, fpstate, mctx);
 	ctx.uc_mcontext = mctx;
 
 	info->client_info.stopped_domain = mono_thread_info_tls_get (info, TLS_KEY_DOMAIN);
@@ -81,11 +79,7 @@ sgen_suspend_thread (SgenThreadInfo *info)
 	if (stack_start >= info->client_info.stack_start_limit && stack_start <= info->client_info.stack_end) {
 		info->client_info.stack_start = stack_start;
 
-#ifdef USE_MONO_CTX
 		mono_sigctx_to_monoctx (&ctx, &info->client_info.ctx);
-#else
-		ARCH_COPY_SIGCTX_REGS (&info->client_info.regs, &ctx);
-#endif
 	} else {
 		g_assert (!info->client_info.stack_start);
 	}
@@ -128,7 +122,9 @@ sgen_thread_handshake (BOOL suspend)
 			if (!sgen_suspend_thread (info))
 				continue;
 		} else {
-			ret = thread_resume (info->client_info.info.native_handle);
+			do {
+				ret = thread_resume (info->client_info.info.native_handle);
+			} while (ret == KERN_ABORTED);
 			if (ret != KERN_SUCCESS)
 				continue;
 		}
