@@ -17,12 +17,10 @@ WRAPPER(XDOMAIN_INVOKE, "xdomain-invoke")
 WRAPPER(XDOMAIN_DISPATCH, "xdomain-dispatch")
 WRAPPER(LDFLD, "ldfld")
 WRAPPER(STFLD, "stfld")
-WRAPPER(LDFLD_REMOTE, "ldfld-remote")
-WRAPPER(STFLD_REMOTE, "stfld-remote")
 WRAPPER(SYNCHRONIZED, "synchronized")
 WRAPPER(DYNAMIC_METHOD, "dynamic-method")
 WRAPPER(ISINST, "isinst")
-WRAPPER(CASTCLASS, "cancast")
+WRAPPER(CASTCLASS, "castclass")
 WRAPPER(PROXY_ISINST, "proxy_isinst")
 WRAPPER(STELEMREF, "stelemref")
 WRAPPER(UNBOX, "unbox")
@@ -32,5 +30,5 @@ WRAPPER(UNKNOWN, "unknown")
 WRAPPER(COMINTEROP_INVOKE, "cominterop-invoke")
 WRAPPER(COMINTEROP, "cominterop")
 WRAPPER(ALLOC, "alloc")
-WRAPPER(STATIC_RGCTX_INVOKE, "static-rgctx-invoke")
+
 
