@@ -41,23 +41,24 @@ static GSList *cached_info_list;
 /* Statistics */
 static int unwind_info_size;
 
-#define unwind_lock() mono_mutex_lock (&unwind_mutex)
-#define unwind_unlock() mono_mutex_unlock (&unwind_mutex)
+#define unwind_lock() mono_os_mutex_lock (&unwind_mutex)
+#define unwind_unlock() mono_os_mutex_unlock (&unwind_mutex)
 
 #ifdef TARGET_AMD64
 static int map_hw_reg_to_dwarf_reg [] = { 0, 2, 1, 3, 7, 6, 4, 5, 8, 9, 10, 11, 12, 13, 14, 15, 16 };
-#define NUM_REGS AMD64_NREG
+#define NUM_DWARF_REGS AMD64_NREG
 #define DWARF_DATA_ALIGN (-8)
 #define DWARF_PC_REG (mono_hw_reg_to_dwarf_reg (AMD64_RIP))
 #elif defined(TARGET_ARM)
 // http://infocenter.arm.com/help/topic/com.arm.doc.ihi0040a/IHI0040A_aadwarf.pdf
-/* Assign d8..d15 to hregs 16..24 */
+/* Assign d8..d15 to hregs 16..24 (dwarf regs 264..271) */
 static int map_hw_reg_to_dwarf_reg [] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 264, 265, 266, 267, 268, 269, 270, 271 };
-#define NUM_REGS 272
+#define NUM_DWARF_REGS 272
 #define DWARF_DATA_ALIGN (-4)
 #define DWARF_PC_REG (mono_hw_reg_to_dwarf_reg (ARMREG_LR))
+#define IS_DOUBLE_REG(dwarf_reg) (((dwarf_reg) >= 264) && ((dwarf_reg) <= 271))
 #elif defined(TARGET_ARM64)
-#define NUM_REGS 96
+#define NUM_DWARF_REGS 96
 #define DWARF_DATA_ALIGN (-8)
 /* LR */
 #define DWARF_PC_REG 30
@@ -74,21 +75,25 @@ static int map_hw_reg_to_dwarf_reg [] = {
  */
 static int map_hw_reg_to_dwarf_reg [] = { 0, 1, 2, 3, 5, 4, 6, 7, 8 };
 /* + 1 is for IP */
-#define NUM_REGS X86_NREG + 1
+#define NUM_DWARF_REGS (X86_NREG + 1)
 #define DWARF_DATA_ALIGN (-4)
 #define DWARF_PC_REG (mono_hw_reg_to_dwarf_reg (X86_NREG))
 #elif defined (TARGET_POWERPC)
 // http://refspecs.linuxfoundation.org/ELF/ppc64/PPC-elf64abi-1.9.html
-static int map_hw_reg_to_dwarf_reg [] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 
+static int map_hw_reg_to_dwarf_reg [ppc_lr + 1] = { 0, 1, 2, 3, 4, 5, 6, 7, 8,
 										  9, 10, 11, 12, 13, 14, 15, 16,
 										  17, 18, 19, 20, 21, 22, 23, 24,
 										  25, 26, 27, 28, 29, 30, 31 };
-#define NUM_REGS 110
 #define DWARF_DATA_ALIGN (-(gint32)sizeof (mgreg_t))
+#if _CALL_ELF == 2
+#define DWARF_PC_REG 65
+#else
 #define DWARF_PC_REG 108
+#endif
+#define NUM_DWARF_REGS (DWARF_PC_REG + 1)
 #elif defined (TARGET_S390X)
 static int map_hw_reg_to_dwarf_reg [] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
-#define NUM_REGS 16
+#define NUM_DWARF_REGS 16
 #define DWARF_DATA_ALIGN (-8)
 #define DWARF_PC_REG (mono_hw_reg_to_dwarf_reg (14))
 #elif defined (TARGET_MIPS)
@@ -99,19 +104,36 @@ static int map_hw_reg_to_dwarf_reg [32] = {
 	16, 17, 18, 19, 20, 21, 22, 23,
 	24, 25, 26, 27, 28, 29, 30, 31
 };
-#define NUM_REGS 32
+#define NUM_DWARF_REGS 32
 #define DWARF_DATA_ALIGN (-(gint32)sizeof (mgreg_t))
 #define DWARF_PC_REG (mono_hw_reg_to_dwarf_reg (mips_ra))
 #else
 static int map_hw_reg_to_dwarf_reg [16];
-#define NUM_REGS 16
+#define NUM_DWARF_REGS 16
 #define DWARF_DATA_ALIGN 0
 #define DWARF_PC_REG -1
 #endif
 
-static gboolean dwarf_reg_to_hw_reg_inited;
+#define NUM_HW_REGS (sizeof (map_hw_reg_to_dwarf_reg) / sizeof (int))
 
-static int map_dwarf_reg_to_hw_reg [NUM_REGS];
+#ifndef IS_DOUBLE_REG
+#define IS_DOUBLE_REG(dwarf_reg) (dwarf_reg ? 0 : 0)
+#endif
+
+static gboolean dwarf_reg_to_hw_reg_inited;
+static gboolean hw_reg_to_dwarf_reg_inited;
+
+static int map_dwarf_reg_to_hw_reg [NUM_DWARF_REGS];
+
+static void
+init_hw_reg_map (void)
+{
+#ifdef TARGET_POWERPC
+	map_hw_reg_to_dwarf_reg [ppc_lr] = DWARF_PC_REG;
+#endif
+	mono_memory_barrier ();
+	hw_reg_to_dwarf_reg_inited = TRUE;
+}
 
 /*
  * mono_hw_reg_to_dwarf_reg:
@@ -121,14 +143,10 @@ static int map_dwarf_reg_to_hw_reg [NUM_REGS];
 int
 mono_hw_reg_to_dwarf_reg (int reg)
 {
-#ifdef TARGET_POWERPC
-	if (reg == ppc_lr)
-		return 108;
-	else
-		g_assert (reg < NUM_REGS);
-#endif
+	if (!hw_reg_to_dwarf_reg_inited)
+		init_hw_reg_map ();
 
-	if (NUM_REGS == 0) {
+	if (NUM_HW_REGS == 0) {
 		g_assert_not_reached ();
 		return -1;
 	} else {
@@ -137,19 +155,15 @@ mono_hw_reg_to_dwarf_reg (int reg)
 }
 
 static void
-init_reg_map (void)
+init_dwarf_reg_map (void)
 {
 	int i;
 
-	g_assert (NUM_REGS > 0);
-	for (i = 0; i < sizeof (map_hw_reg_to_dwarf_reg) / sizeof (int); ++i) {
+	g_assert (NUM_HW_REGS > 0);
+	for (i = 0; i < NUM_HW_REGS; ++i) {
 		map_dwarf_reg_to_hw_reg [mono_hw_reg_to_dwarf_reg (i)] = i;
 	}
 
-#ifdef TARGET_POWERPC
-	map_dwarf_reg_to_hw_reg [DWARF_PC_REG] = ppc_lr;
-#endif
-
 	mono_memory_barrier ();
 	dwarf_reg_to_hw_reg_inited = TRUE;
 }
@@ -158,7 +172,7 @@ int
 mono_dwarf_reg_to_hw_reg (int reg)
 {
 	if (!dwarf_reg_to_hw_reg_inited)
-		init_reg_map ();
+		init_dwarf_reg_map ();
 
 	return map_dwarf_reg_to_hw_reg [reg];
 }
@@ -333,13 +347,15 @@ mono_print_unwind_info (guint8 *unwind_info, int unwind_info_len)
 }
 
 /*
- * mono_unwind_ops_encode:
+ * mono_unwind_ops_encode_full:
  *
  *   Encode the unwind ops in UNWIND_OPS into the compact DWARF encoding.
  * Return a pointer to malloc'ed memory.
+ * If ENABLE_EXTENSIONS is FALSE, avoid encoding the mono extension
+ * opcode (DW_CFA_mono_advance_loc).
  */
 guint8*
-mono_unwind_ops_encode (GSList *unwind_ops, guint32 *out_len)
+mono_unwind_ops_encode_full (GSList *unwind_ops, guint32 *out_len, gboolean enable_extensions)
 {
 	GSList *l;
 	MonoUnwindOp *op;
@@ -354,7 +370,7 @@ mono_unwind_ops_encode (GSList *unwind_ops, guint32 *out_len)
 	for (; l; l = l->next) {
 		int reg;
 
-		op = l->data;
+		op = (MonoUnwindOp *)l->data;
 
 		/* Convert the register from the hw encoding to the dwarf encoding */
 		reg = mono_hw_reg_to_dwarf_reg (op->reg);
@@ -366,15 +382,17 @@ mono_unwind_ops_encode (GSList *unwind_ops, guint32 *out_len)
 
 		/* Emit an advance_loc if neccesary */
 		while (op->when > loc) {
-			if (op->when - loc > 65536) {
+			if (op->when - loc >= 65536) {
 				*p ++ = DW_CFA_advance_loc4;
-				*(guint32*)p = (guint32)(op->when - loc);
+				guint32 v = (guint32)(op->when - loc);
+				memcpy (p, &v, 4);
 				g_assert (read32 (p) == (guint32)(op->when - loc));
 				p += 4;
 				loc = op->when;
-			} else if (op->when - loc > 256) {
+			} else if (op->when - loc >= 256) {
 				*p ++ = DW_CFA_advance_loc2;
-				*(guint16*)p = (guint16)(op->when - loc);
+				guint16 v = (guint16)(op->when - loc);
+				memcpy (p, &v, 2);
 				g_assert (read16 (p) == (guint32)(op->when - loc));
 				p += 2;
 				loc = op->when;
@@ -425,6 +443,8 @@ mono_unwind_ops_encode (GSList *unwind_ops, guint32 *out_len)
 			*p ++ = op->op;
 			break;
 		case DW_CFA_mono_advance_loc:
+			if (!enable_extensions)
+				break;
 			/* Only one location is supported */
 			g_assert (op->val == 0);
 			*p ++ = op->op;
@@ -437,11 +457,17 @@ mono_unwind_ops_encode (GSList *unwind_ops, guint32 *out_len)
 	
 	g_assert (p - buf < 4096);
 	*out_len = p - buf;
-	res = g_malloc (p - buf);
+	res = (guint8 *)g_malloc (p - buf);
 	memcpy (res, buf, p - buf);
 	return res;
 }
 
+guint8*
+mono_unwind_ops_encode (GSList *unwind_ops, guint32 *out_len)
+{
+	return mono_unwind_ops_encode_full (unwind_ops, out_len, TRUE);
+}
+
 #if 0
 #define UNW_DEBUG(stmt) do { stmt; } while (0)
 #else
@@ -462,8 +488,8 @@ print_dwarf_state (int cfa_reg, int cfa_offset, int ip, int nregs, Loc *location
 }
 
 typedef struct {
-	Loc locations [NUM_REGS];
-	guint8 reg_saved [NUM_REGS];
+	Loc locations [NUM_HW_REGS];
+	guint8 reg_saved [NUM_HW_REGS];
 	int cfa_reg, cfa_offset;
 } UnwindState;
 
@@ -480,13 +506,13 @@ typedef struct {
 void
 mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len, 
 				   guint8 *start_ip, guint8 *end_ip, guint8 *ip, guint8 **mark_locations,
-				   mgreg_t *regs, int nregs,
+				   mono_unwind_reg_t *regs, int nregs,
 				   mgreg_t **save_locations, int save_locations_len,
 				   guint8 **out_cfa)
 {
-	Loc locations [NUM_REGS];
-	guint8 reg_saved [NUM_REGS];
-	int i, pos, reg, cfa_reg = -1, cfa_offset = 0, offset;
+	Loc locations [NUM_HW_REGS];
+	guint8 reg_saved [NUM_HW_REGS];
+	int pos, reg, hwreg, cfa_reg = -1, cfa_offset = 0, offset;
 	guint8 *p;
 	guint8 *cfa_val;
 	UnwindState state_stack [1];
@@ -511,11 +537,11 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 			p ++;
 			break;
 		case DW_CFA_offset:
-			reg = *p & 0x3f;
+			hwreg = mono_dwarf_reg_to_hw_reg (*p & 0x3f);
 			p ++;
-			reg_saved [reg] = TRUE;
-			locations [reg].loc_type = LOC_OFFSET;
-			locations [reg].offset = decode_uleb128 (p, &p) * DWARF_DATA_ALIGN;
+			reg_saved [hwreg] = TRUE;
+			locations [hwreg].loc_type = LOC_OFFSET;
+			locations [hwreg].offset = decode_uleb128 (p, &p) * DWARF_DATA_ALIGN;
 			break;
 		case 0: {
 			int ext_op = *p;
@@ -533,23 +559,25 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 				break;
 			case DW_CFA_offset_extended_sf:
 				reg = decode_uleb128 (p, &p);
+				hwreg = mono_dwarf_reg_to_hw_reg (reg);
 				offset = decode_sleb128 (p, &p);
-				g_assert (reg < NUM_REGS);
-				reg_saved [reg] = TRUE;
-				locations [reg].loc_type = LOC_OFFSET;
-				locations [reg].offset = offset * DWARF_DATA_ALIGN;
+				g_assert (reg < NUM_DWARF_REGS);
+				reg_saved [hwreg] = TRUE;
+				locations [hwreg].loc_type = LOC_OFFSET;
+				locations [hwreg].offset = offset * DWARF_DATA_ALIGN;
 				break;
 			case DW_CFA_offset_extended:
 				reg = decode_uleb128 (p, &p);
+				hwreg = mono_dwarf_reg_to_hw_reg (reg);
 				offset = decode_uleb128 (p, &p);
-				g_assert (reg < NUM_REGS);
-				reg_saved [reg] = TRUE;
-				locations [reg].loc_type = LOC_OFFSET;
-				locations [reg].offset = offset * DWARF_DATA_ALIGN;
+				g_assert (reg < NUM_DWARF_REGS);
+				reg_saved [hwreg] = TRUE;
+				locations [hwreg].loc_type = LOC_OFFSET;
+				locations [hwreg].offset = offset * DWARF_DATA_ALIGN;
 				break;
 			case DW_CFA_same_value:
-				reg = decode_uleb128 (p, &p);
-				locations [reg].loc_type = LOC_SAME;
+				hwreg = mono_dwarf_reg_to_hw_reg (decode_uleb128 (p, &p));
+				locations [hwreg].loc_type = LOC_SAME;
 				break;
 			case DW_CFA_advance_loc1:
 				pos += *p;
@@ -598,13 +626,16 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 
 	g_assert (cfa_reg != -1);
 	cfa_val = (guint8*)regs [mono_dwarf_reg_to_hw_reg (cfa_reg)] + cfa_offset;
-	for (i = 0; i < NUM_REGS; ++i) {
-		if (reg_saved [i] && locations [i].loc_type == LOC_OFFSET) {
-			int hreg = mono_dwarf_reg_to_hw_reg (i);
-			g_assert (hreg < nregs);
-			regs [hreg] = *(mgreg_t*)(cfa_val + locations [i].offset);
-			if (save_locations && hreg < save_locations_len)
-				save_locations [hreg] = (mgreg_t*)(cfa_val + locations [i].offset);
+	for (hwreg = 0; hwreg < NUM_HW_REGS; ++hwreg) {
+		if (reg_saved [hwreg] && locations [hwreg].loc_type == LOC_OFFSET) {
+			int dwarfreg = mono_hw_reg_to_dwarf_reg (hwreg);
+			g_assert (hwreg < nregs);
+			if (IS_DOUBLE_REG (dwarfreg))
+				regs [hwreg] = *(guint64*)(cfa_val + locations [hwreg].offset);
+			else
+				regs [hwreg] = *(mgreg_t*)(cfa_val + locations [hwreg].offset);
+			if (save_locations && hwreg < save_locations_len)
+				save_locations [hwreg] = (mgreg_t*)(cfa_val + locations [hwreg].offset);
 		}
 	}
 
@@ -614,7 +645,7 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 void
 mono_unwind_init (void)
 {
-	mono_mutex_init_recursive (&unwind_mutex);
+	mono_os_mutex_init_recursive (&unwind_mutex);
 
 	mono_counters_register ("Unwind info size", MONO_COUNTER_JIT | MONO_COUNTER_INT, &unwind_info_size);
 }
@@ -624,7 +655,7 @@ mono_unwind_cleanup (void)
 {
 	int i;
 
-	mono_mutex_destroy (&unwind_mutex);
+	mono_os_mutex_destroy (&unwind_mutex);
 
 	if (!cached_info)
 		return;
@@ -634,8 +665,12 @@ mono_unwind_cleanup (void)
 
 		g_free (cached);
 	}
-
 	g_free (cached_info);
+
+	for (GSList *cursor = cached_info_list; cursor != NULL; cursor = cursor->next)
+		g_free (cursor->data);
+
+	g_slist_free (cached_info_list);
 }
 
 /*
@@ -670,31 +705,30 @@ mono_cache_unwind_info (guint8 *unwind_info, guint32 unwind_info_len)
 		}
 	}
 
-	info = g_malloc (sizeof (MonoUnwindInfo) + unwind_info_len);
+	info = (MonoUnwindInfo *)g_malloc (sizeof (MonoUnwindInfo) + unwind_info_len);
 	info->len = unwind_info_len;
 	memcpy (&info->info, unwind_info, unwind_info_len);
 
 	i = cached_info_next;
 	
 	if (cached_info_next >= cached_info_size) {
-		MonoUnwindInfo **old_table, **new_table;
+		MonoUnwindInfo **new_table;
 
 		/*
 		 * Avoid freeing the old table so mono_get_cached_unwind_info ()
 		 * doesn't need locks/hazard pointers.
 		 */
 
-		old_table = cached_info;
 		new_table = g_new0 (MonoUnwindInfo*, cached_info_size * 2);
 
 		memcpy (new_table, cached_info, cached_info_size * sizeof (MonoUnwindInfo*));
 
 		mono_memory_barrier ();
 
-		cached_info = new_table;
-
 		cached_info_list = g_slist_prepend (cached_info_list, cached_info);
 
+		cached_info = new_table;
+
 		cached_info_size *= 2;
 	}
 
@@ -864,11 +898,11 @@ decode_lsda (guint8 *lsda, guint8 *code, MonoJitExceptionInfo **ex_info, guint32
 	p = (guint8*)ALIGN_TO ((mgreg_t)p, 4);
 
 	if (ex_info) {
-		*ex_info = g_malloc0 (ncall_sites * sizeof (MonoJitExceptionInfo));
+		*ex_info = (MonoJitExceptionInfo *)g_malloc0 (ncall_sites * sizeof (MonoJitExceptionInfo));
 		*ex_info_len = ncall_sites;
 	}
 	if (type_info)
-		*type_info = g_malloc0 (ncall_sites * sizeof (gpointer));
+		*type_info = (gpointer *)g_malloc0 (ncall_sites * sizeof (gpointer));
 
 	for (i = 0; i < ncall_sites; ++i) {
 		int block_start_offset, block_size, landing_pad;
@@ -910,7 +944,7 @@ guint8*
 mono_unwind_decode_fde (guint8 *fde, guint32 *out_len, guint32 *code_len, MonoJitExceptionInfo **ex_info, guint32 *ex_info_len, gpointer **type_info, int *this_reg, int *this_offset)
 {
 	guint8 *p, *cie, *fde_current, *fde_aug = NULL, *code, *fde_cfi, *cie_cfi;
-	gint32 fde_len, cie_offset, pc_begin, pc_range, aug_len, fde_data_len;
+	gint32 fde_len, cie_offset, pc_begin, pc_range, aug_len;
 	gint32 cie_len, cie_id, cie_version, code_align, data_align, return_reg;
 	gint32 i, cie_aug_len, buf_len;
 	char *cie_aug_str;
@@ -921,6 +955,8 @@ mono_unwind_decode_fde (guint8 *fde, guint32 *out_len, guint32 *code_len, MonoJi
 	 * http://refspecs.freestandards.org/LSB_3.0.0/LSB-Core-generic/LSB-Core-generic/ehframechpt.html
 	 */
 
+	/* This is generated by JITDwarfEmitter::EmitEHFrame () */
+
 	*type_info = NULL;
 	*this_reg = -1;
 	*this_offset = -1;
@@ -1005,7 +1041,6 @@ mono_unwind_decode_fde (guint8 *fde, guint32 *out_len, guint32 *code_len, MonoJi
 		aug_len = 0;
 	}
 	fde_cfi = p;
-	fde_data_len = fde + 4 + fde_len - p;
 
 	if (code_len)
 		*code_len = pc_range;
@@ -1040,7 +1075,7 @@ mono_unwind_decode_fde (guint8 *fde, guint32 *out_len, guint32 *code_len, MonoJi
 	g_assert (return_reg == DWARF_PC_REG);
 
 	buf_len = (cie + cie_len + 4 - cie_cfi) + (fde + fde_len + 4 - fde_cfi);
-	buf = g_malloc0 (buf_len);
+	buf = (guint8 *)g_malloc0 (buf_len);
 
 	i = 0;
 	p = cie_cfi;
@@ -1066,7 +1101,7 @@ mono_unwind_decode_fde (guint8 *fde, guint32 *out_len, guint32 *code_len, MonoJi
 
 	*out_len = i;
 
-	return g_realloc (buf, i);
+	return (guint8 *)g_realloc (buf, i);
 }
 
 /*
@@ -1089,7 +1124,7 @@ mono_unwind_decode_llvm_mono_fde (guint8 *fde, int fde_len, guint8 *cie, guint8
 	res->this_reg = -1;
 	res->this_offset = -1;
 
-	/* fde points to data emitted by LLVM in DwarfException::EmitMonoEHFrame () */
+	/* fde points to data emitted by LLVM in DwarfMonoException::EmitMonoEHFrame () */
 	p = fde;
 	has_aug = *p;
 	p ++;
@@ -1140,7 +1175,7 @@ mono_unwind_decode_llvm_mono_fde (guint8 *fde, int fde_len, guint8 *cie, guint8
 	cie_cfi_len = p - cie_cfi;
 	fde_cfi_len = (fde + fde_len - fde_cfi);
 
-	buf = g_malloc0 (cie_cfi_len + fde_cfi_len);
+	buf = (guint8 *)g_malloc0 (cie_cfi_len + fde_cfi_len);
 	memcpy (buf, cie_cfi, cie_cfi_len);
 	memcpy (buf + cie_cfi_len, fde_cfi, fde_cfi_len);
 
@@ -1156,7 +1191,7 @@ mono_unwind_decode_llvm_mono_fde (guint8 *fde, int fde_len, guint8 *cie, guint8
 GSList*
 mono_unwind_get_cie_program (void)
 {
-#if defined(TARGET_AMD64) || defined(TARGET_X86) || defined(TARGET_POWERPC)
+#if defined(TARGET_AMD64) || defined(TARGET_X86) || defined(TARGET_POWERPC) || defined(TARGET_ARM)
 	return mono_arch_get_cie_program ();
 #else
 	return NULL;
