@@ -356,15 +356,15 @@ typedef struct {
 		"push {r0}\n"				\
 		"push {r1}\n"				\
 		"mov r0, %0\n"				\
-		"ldr r1, [sp, #4]\n"			\
-		"str r1, [r0]!\n"			\
-		"ldr r1, [sp, #0]\n"			\
-		"str r1, [r0]!\n"			\
+		"ldr r1, [sp, #4]\n"   		\
+		"str r1, [r0], #4\n"   		\
+		"ldr r1, [sp, #0]\n"	   	\
+		"str r1, [r0], #4\n"	   	\
 		"stmia r0!, {r2-r12}\n"		\
-		"str sp, [r0]!\n"			\
-		"str lr, [r0]!\n"			\
+		"str sp, [r0], #4\n"		\
+		"str lr, [r0], #4\n"		\
 		"mov r1, pc\n"				\
-		"str r1, [r0]!\n"			\
+		"str r1, [r0], #4\n"		\
 		"pop {r1}\n"				\
 		"pop {r0}\n"				\
 		:							\
@@ -386,6 +386,11 @@ typedef struct {
 	mgreg_t regs [32];
 	double fregs [32];
 	mgreg_t pc;
+	/*
+	 * fregs might not be initialized if this context was created from a
+	 * ucontext.
+	 */
+	mgreg_t has_fregs;
 } MonoContext;
 
 #define MONO_CONTEXT_SET_IP(ctx,ip) do { (ctx)->pc = (mgreg_t)ip; } while (0)
