@@ -6,7 +6,7 @@
 #include <glib.h>
 
 #include "mono-os-semaphore.h"
-#include "mono-threads.h"
+#include "mono-threads-api.h"
 
 G_BEGIN_DECLS
 
@@ -18,16 +18,16 @@ struct _MonoCoopSem {
 	MonoSemType s;
 };
 
-static inline gint
+static inline void
 mono_coop_sem_init (MonoCoopSem *sem, int value)
 {
-	return mono_os_sem_init (&sem->s, value);
+	mono_os_sem_init (&sem->s, value);
 }
 
-static inline gint
+static inline void
 mono_coop_sem_destroy (MonoCoopSem *sem)
 {
-	return mono_os_sem_destroy (&sem->s);
+	mono_os_sem_destroy (&sem->s);
 }
 
 static inline gint
@@ -35,33 +35,33 @@ mono_coop_sem_wait (MonoCoopSem *sem, MonoSemFlags flags)
 {
 	gint res;
 
-	MONO_PREPARE_BLOCKING;
+	MONO_ENTER_GC_SAFE;
 
 	res = mono_os_sem_wait (&sem->s, flags);
 
-	MONO_FINISH_BLOCKING;
+	MONO_EXIT_GC_SAFE;
 
 	return res;
 }
 
-static inline gint
+static inline MonoSemTimedwaitRet
 mono_coop_sem_timedwait (MonoCoopSem *sem, guint timeout_ms, MonoSemFlags flags)
 {
-	gint res;
+	MonoSemTimedwaitRet res;
 
-	MONO_PREPARE_BLOCKING;
+	MONO_ENTER_GC_SAFE;
 
 	res = mono_os_sem_timedwait (&sem->s, timeout_ms, flags);
 
-	MONO_FINISH_BLOCKING;
+	MONO_EXIT_GC_SAFE;
 
 	return res;
 }
 
-static inline gint
+static inline void
 mono_coop_sem_post (MonoCoopSem *sem)
 {
-	return mono_os_sem_post (&sem->s);
+	mono_os_sem_post (&sem->s);
 }
 
 G_END_DECLS
