@@ -1,6 +1,7 @@
 /*
  * Copyright 2006-2010 Novell
  * Copyright 2011 Xamarin Inc
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 
 #include <stdlib.h>
@@ -165,7 +166,7 @@ register_internal (const char *name, int type, void *addr, int size)
 		}
 	}
 
-	counter = (MonoCounter *) malloc (sizeof (MonoCounter));
+	counter = (MonoCounter *) g_malloc (sizeof (MonoCounter));
 	if (!counter) {
 		mono_os_mutex_unlock (&counters_mutex);
 		return;
@@ -560,7 +561,7 @@ dump_counter (MonoCounter *counter, FILE *outfile) {
 }
 
 static const char
-section_names [][10] = {
+section_names [][12] = {
 	"JIT",
 	"GC",
 	"Metadata",
@@ -568,6 +569,8 @@ section_names [][10] = {
 	"Security",
 	"Runtime",
 	"System",
+	"", // MONO_COUNTER_PERFCOUNTERS - not used.
+	"Profiler",
 };
 
 static void
@@ -645,8 +648,8 @@ mono_counters_cleanup (void)
 	while (counter) {
 		MonoCounter *tmp = counter;
 		counter = counter->next;
-		free ((void*)tmp->name);
-		free (tmp);
+		g_free ((void*)tmp->name);
+		g_free (tmp);
 	}
 
 	mono_os_mutex_unlock (&counters_mutex);
