@@ -4,40 +4,83 @@
  * Authors:
  *	Rodrigo Kumpera    (rkumpera@novell.com)
  * Copyright 2009 Novell, Inc (http://www.novell.com)
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 #include <glib.h>
 
+#include <config.h>
 #include "mono-error.h"
 #include "mono-error-internals.h"
 
 #include <mono/metadata/exception.h>
-#include <mono/metadata/object-internals.h>
-#include <mono/metadata/class-internals.h>
+#include <mono/metadata/exception-internals.h>
 #include <mono/metadata/debug-helpers.h>
+#include <mono/metadata/object-internals.h>
 
-#define mono_internal_error_get_message(E) ((E)->full_message ? (E)->full_message : (E)->message)
+#define set_error_messagev() do { \
+	if (!(error->full_message = g_strdup_vprintf (msg_format, args))) \
+			error->flags |= MONO_ERROR_INCOMPLETE; \
+} while (0)
 
 #define set_error_message() do { \
 	va_list args; \
 	va_start (args, msg_format); \
-	if (g_vsnprintf (error->message, sizeof (error->message), msg_format, args) >= sizeof (error->message)) {\
-		va_end (args); \
-		va_start (args, msg_format); \
-		if (!(error->full_message = g_strdup_vprintf (msg_format, args))) \
-			error->flags |= MONO_ERROR_INCOMPLETE; \
-	} \
+	set_error_messagev();	     \
 	va_end (args); \
 } while (0)
 
+static void
+mono_error_set_generic_errorv (MonoError *oerror, const char *name_space, const char *name, const char *msg_format, va_list args);
+
+static gboolean
+is_managed_exception (MonoErrorInternal *error)
+{
+	return (error->error_code == MONO_ERROR_EXCEPTION_INSTANCE);
+}
+
 static void
 mono_error_prepare (MonoErrorInternal *error)
 {
+	/* mono_error_set_* after a mono_error_cleanup without an intervening init */
+	g_assert (error->error_code != MONO_ERROR_CLEANUP_CALLED_SENTINEL);
 	if (error->error_code != MONO_ERROR_NONE)
 		return;
 
-	error->type_name = error->assembly_name = error->member_name = error->full_message = error->exception_name_space = error->exception_name = NULL;
-	error->klass = NULL;
-	error->message [0] = 0;
+	error->type_name = error->assembly_name = error->member_name = error->full_message = error->exception_name_space = error->exception_name = error->full_message_with_fields = error->first_argument = NULL;
+	error->exn.klass = NULL;
+}
+
+static MonoClass*
+get_class (MonoErrorInternal *error)
+{
+	MonoClass *klass = NULL;
+	if (is_managed_exception (error))
+		klass = mono_object_class (mono_gchandle_get_target (error->exn.instance_handle));
+	else
+		klass = error->exn.klass;
+	return klass;
+}
+
+static const char*
+get_type_name (MonoErrorInternal *error)
+{
+	if (error->type_name)
+		return error->type_name;
+	MonoClass *klass = get_class (error);
+	if (klass)
+		return klass->name;
+	return "<unknown type>";
+}
+
+static const char*
+get_assembly_name (MonoErrorInternal *error)
+{
+	if (error->assembly_name)
+		return error->assembly_name;
+	MonoClass *klass = get_class (error);
+	if (klass && klass->image)
+		return klass->image->name;
+	return "<unknown assembly>";
 }
 
 void
@@ -50,6 +93,13 @@ mono_error_init_flags (MonoError *oerror, unsigned short flags)
 	error->flags = flags;
 }
 
+/**
+ * mono_error_init:
+ * @error: Pointer to MonoError struct to initialize
+ *
+ * Any function which takes a MonoError for purposes of reporting an error
+ * is required to call either this or mono_error_init_flags on entry.
+ */
 void
 mono_error_init (MonoError *error)
 {
@@ -60,11 +110,30 @@ void
 mono_error_cleanup (MonoError *oerror)
 {
 	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
-	if (error->error_code == MONO_ERROR_NONE)
+	short int orig_error_code = error->error_code;
+	gboolean free_strings = error->flags & MONO_ERROR_FREE_STRINGS;
+	gboolean has_instance_handle = is_managed_exception (error);
+
+	/* Two cleanups in a row without an intervening init. */
+	g_assert (orig_error_code != MONO_ERROR_CLEANUP_CALLED_SENTINEL);
+
+	/* Mark it as cleaned up. */
+	error->error_code = MONO_ERROR_CLEANUP_CALLED_SENTINEL;
+	error->flags = 0;
+
+	if (orig_error_code == MONO_ERROR_NONE)
 		return;
 
+
+	if (has_instance_handle)
+		mono_gchandle_free (error->exn.instance_handle);
+
+
 	g_free ((char*)error->full_message);
-	if (!(error->flags & MONO_ERROR_FREE_STRINGS)) //no memory was allocated
+	g_free ((char*)error->full_message_with_fields);
+	error->full_message = NULL;
+	error->full_message_with_fields = NULL;
+	if (!free_strings) //no memory was allocated
 		return;
 
 	g_free ((char*)error->type_name);
@@ -72,6 +141,10 @@ mono_error_cleanup (MonoError *oerror)
 	g_free ((char*)error->member_name);
 	g_free ((char*)error->exception_name_space);
 	g_free ((char*)error->exception_name);
+	g_free ((char*)error->first_argument);
+	error->type_name = error->assembly_name = error->member_name = error->exception_name_space = error->exception_name = error->first_argument = NULL;
+	error->exn.klass = NULL;
+
 }
 
 gboolean
@@ -80,6 +153,15 @@ mono_error_ok (MonoError *error)
 	return error->error_code == MONO_ERROR_NONE;
 }
 
+void
+mono_error_assert_ok_pos (MonoError *error, const char* filename, int lineno)
+{
+	if (mono_error_ok (error))
+		return;
+
+	g_error ("%s:%d: %s\n", filename, lineno, mono_error_get_message (error));
+}
+
 unsigned short
 mono_error_get_error_code (MonoError *error)
 {
@@ -94,7 +176,16 @@ mono_error_get_message (MonoError *oerror)
 	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
 	if (error->error_code == MONO_ERROR_NONE)
 		return NULL;
-	return mono_internal_error_get_message (error);
+	if (error->full_message_with_fields)
+		return error->full_message_with_fields;
+
+	error->full_message_with_fields = g_strdup_printf ("%s assembly:%s type:%s member:%s",
+		error->full_message,
+		get_assembly_name (error),
+		get_type_name (error),
+		error->member_name ? error->member_name : "<none>");
+
+	return error->full_message_with_fields ? error->full_message_with_fields : error->full_message;
 }
 
 /*
@@ -119,6 +210,7 @@ mono_error_dup_strings (MonoError *oerror, gboolean dup_strings)
 		DUP_STR (member_name);
 		DUP_STR (exception_name_space);
 		DUP_STR (exception_name);
+		DUP_STR (first_argument);
 	}
 #undef DUP_STR
 }
@@ -163,7 +255,9 @@ mono_error_set_class (MonoError *oerror, MonoClass *klass)
 {
 	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
 
-	error->klass = klass;	
+	if (is_managed_exception (error))
+		return;
+	error->exn.klass = klass;	
 }
 
 static void
@@ -195,7 +289,7 @@ mono_error_set_assembly_load_simple (MonoError *oerror, const char *assembly_nam
 	if (refection_only)
 		mono_error_set_assembly_load (oerror, assembly_name, "Cannot resolve dependency to assembly because it has not been preloaded. When using the ReflectionOnly APIs, dependent assemblies must be pre-loaded or loaded on demand through the ReflectionOnlyAssemblyResolve event.");
 	else
-		mono_error_set_assembly_load (oerror, assembly_name, "Could not load file or assembly or one of its dependencies.");
+		mono_error_set_assembly_load (oerror, assembly_name, "Could not load file or assembly '%s' or one of its dependencies.", assembly_name);
 }
 
 void
@@ -273,155 +367,111 @@ mono_error_set_bad_image (MonoError *oerror, MonoImage *image, const char *msg_f
 }
 
 void
-mono_error_set_generic_error (MonoError *oerror, const char * name_space, const char *name, const char *msg_format, ...)
+mono_error_set_generic_errorv (MonoError *oerror, const char *name_space, const char *name, const char *msg_format, va_list args)
 {
 	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
 	mono_error_prepare (error);
 
 	error->error_code = MONO_ERROR_GENERIC;
 	mono_error_set_corlib_exception (oerror, name_space, name);
+	set_error_messagev ();
+}
+
+void
+mono_error_set_generic_error (MonoError *oerror, const char * name_space, const char *name, const char *msg_format, ...)
+{
+	va_list args;
+	va_start (args, msg_format);
+	mono_error_set_generic_errorv (oerror, name_space, name, msg_format, args);
+	va_end (args);
+}
+
+/**
+ * mono_error_set_not_implemented:
+ *
+ * System.NotImplementedException
+ */
+void
+mono_error_set_not_implemented (MonoError *oerror, const char *msg_format, ...)
+{
+	va_list args;
+	va_start (args, msg_format);
+	mono_error_set_generic_errorv (oerror, "System", "NotImplementedException", msg_format, args);
+	va_end (args);
+}
+
+/**
+ * mono_error_set_execution_engine:
+ *
+ * System.ExecutionEngineException
+ */
+void
+mono_error_set_execution_engine (MonoError *oerror, const char *msg_format, ...)
+{
+	va_list args;
+	va_start (args, msg_format);
+	mono_error_set_generic_errorv (oerror, "System", "ExecutionEngineException", msg_format, args);
+	va_end (args);
+}
+
+/**
+ * mono_error_set_not_supported:
+ *
+ * System.NotSupportedException
+ */
+void
+mono_error_set_not_supported (MonoError *oerror, const char *msg_format, ...)
+{
+	va_list args;
+	va_start (args, msg_format);
+	mono_error_set_generic_errorv (oerror, "System", "NotSupportedException", msg_format, args);
+	va_end (args);
+}
+
+/**
+ * mono_error_set_invalid_operation:
+ *
+ * System.InvalidOperationException
+ */
+void
+mono_error_set_invalid_operation (MonoError *oerror, const char *msg_format, ...)
+{
+	va_list args;
+	va_start (args, msg_format);
+	mono_error_set_generic_errorv (oerror, "System", "InvalidOperationException", msg_format, args);
+	va_end (args);
+}
+
+void
+mono_error_set_invalid_program (MonoError *oerror, const char *msg_format, ...)
+{
+	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
+
+	mono_error_prepare (error);
+	error->error_code = MONO_ERROR_INVALID_PROGRAM;
+
 	set_error_message ();
 }
 
 void
-mono_error_set_from_loader_error (MonoError *oerror)
+mono_error_set_exception_instance (MonoError *oerror, MonoException *exc)
 {
-	MonoLoaderError *loader_error = mono_loader_get_last_error ();
 	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
-	gboolean dup_strings = TRUE;
 
 	mono_error_prepare (error);
-
-	if (!loader_error) {
-		mono_error_set_generic_error (oerror, "System", "ExecutionEngineException", "Runtime tried to produce a mono-error from an empty loader-error");
-		return;
-	}
-
-	switch (loader_error->exception_type) {
-	case MONO_EXCEPTION_NONE:
-		mono_error_set_generic_error (oerror, "System", "ExecutionEngineException", "Runtime tried to produce a mono-error from a non-error loader-error");
-		break;
-	case MONO_EXCEPTION_SECURITY_LINKDEMAND:
-	case MONO_EXCEPTION_SECURITY_INHERITANCEDEMAND:
-		mono_error_set_generic_error (oerror, "System.Security", "SecurityException", "Failed for unknown reasons.");
-		break;
-
-	case MONO_EXCEPTION_INVALID_PROGRAM:
-		mono_error_set_generic_error (oerror, "System", "InvalidProgramException", "Failed for unknown reasons.");
-		break;
-
-	case MONO_EXCEPTION_UNVERIFIABLE_IL:
-		mono_error_set_generic_error (oerror, "System.Security", "VerificationException", "Failed for unknown reasons.");
-		break;
-
-	case MONO_EXCEPTION_MISSING_METHOD:
-		error->error_code = MONO_ERROR_MISSING_METHOD;
-		mono_error_set_type_name (oerror, loader_error->class_name);
-		mono_error_set_member_name (oerror, loader_error->member_name);
-		g_snprintf (error->message, sizeof (error->message), "Failed for unknown reasons.");
-		break;
-
-	case MONO_EXCEPTION_MISSING_FIELD:
-		mono_error_set_field_load (oerror, loader_error->klass, loader_error->member_name, "Failed for unknown reasons.");
-		break;
-
-	case MONO_EXCEPTION_TYPE_LOAD:
-		mono_error_set_type_load_name (oerror, g_strdup (loader_error->class_name), g_strdup (loader_error->assembly_name), "Failed for unknown reasons.");
-		dup_strings = FALSE;
-		break;
-	
-	case MONO_EXCEPTION_FILE_NOT_FOUND:
-		mono_error_set_assembly_load_simple (oerror, loader_error->assembly_name, loader_error->ref_only);
-		break;
-
-	case MONO_EXCEPTION_METHOD_ACCESS:
-		mono_error_set_generic_error (oerror, "System", "MethodAccessException", "Failed for unknown reasons.");
-		break;
-
-	case MONO_EXCEPTION_FIELD_ACCESS:
-		mono_error_set_generic_error (oerror, "System", "FieldAccessException", "Failed for unknown reasons.");
-		break;
-
-	case MONO_EXCEPTION_OBJECT_SUPPLIED:
-	case MONO_EXCEPTION_GENERIC_SHARING_FAILED:
-		mono_error_set_generic_error (oerror, "System", "ExecutionEngineException", "Runtime tried to produce a mono-error from JIT internal error %d", loader_error->exception_type);
-		break;
-
-	case MONO_EXCEPTION_BAD_IMAGE:
-		mono_error_set_bad_image_name (oerror, "<unknown>", "%s", loader_error->msg);
-		break;
-
-	case MONO_EXCEPTION_OUT_OF_MEMORY:
-		mono_error_set_out_of_memory (oerror, "Failed for unknown reasons.");
-		break;
-
-	default:
-		mono_error_set_generic_error (oerror, "System", "ExecutionEngineException", "Runtime tried to produce an unknown loader-error %d", loader_error->exception_type);
-		break;
-	}
-
-	mono_error_dup_strings (oerror, dup_strings);
-	mono_loader_clear_error ();
-}
-
-static const char*
-get_type_name (MonoErrorInternal *error)
-{
-	if (error->type_name)
-		return error->type_name;
-	if (error->klass)
-		return error->klass->name;
-	return "<unknown type>";
-}
-
-static const char*
-get_assembly_name (MonoErrorInternal *error)
-{
-	if (error->assembly_name)
-		return error->assembly_name;
-	if (error->klass && error->klass->image)
-		return error->klass->image->name;
-	return "<unknown assembly>";
-}
-
-void
-mono_loader_set_error_from_mono_error (MonoError *oerror)
-{
-	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
-
-	switch (error->error_code) {
-	case MONO_ERROR_MISSING_METHOD:
-		mono_loader_set_error_method_load (get_type_name (error), error->member_name);
-		break;
-	case MONO_ERROR_MISSING_FIELD:
-		mono_loader_set_error_field_load (error->klass, error->member_name);
-		break;
-	case MONO_ERROR_TYPE_LOAD:
-		mono_loader_set_error_type_load (get_type_name (error), get_assembly_name (error));
-		break;
-	case MONO_ERROR_FILE_NOT_FOUND:
-		/* XXX can't recover if it's ref only or not */
-		mono_loader_set_error_assembly_load (get_assembly_name (error), FALSE);
-		break;
-	case MONO_ERROR_BAD_IMAGE:
-		mono_loader_set_error_bad_image (g_strdup (mono_internal_error_get_message (error)));
-		break;
-	default:
-		mono_loader_set_error_bad_image (g_strdup_printf ("Non translatable error: %s", mono_internal_error_get_message (error)));
-	}
+	error->error_code = MONO_ERROR_EXCEPTION_INSTANCE;
+	error->exn.instance_handle = mono_gchandle_new (exc ? &exc->object : NULL, FALSE);
 }
 
 void
 mono_error_set_out_of_memory (MonoError *oerror, const char *msg_format, ...)
 {
 	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
-	va_list args;
 	mono_error_prepare (error);
 
 	error->error_code = MONO_ERROR_OUT_OF_MEMORY;
-	va_start (args, msg_format);
-	g_vsnprintf (error->message, sizeof (error->message), msg_format, args);
-	va_end (args);
+
+	set_error_message ();
 }
 
 void
@@ -431,7 +481,19 @@ mono_error_set_argument (MonoError *oerror, const char *argument, const char *ms
 	mono_error_prepare (error);
 
 	error->error_code = MONO_ERROR_ARGUMENT;
-	error->type_name = argument; /*use the first available string slot*/
+	error->first_argument = argument;
+
+	set_error_message ();
+}
+
+void
+mono_error_set_argument_null (MonoError *oerror, const char *argument, const char *msg_format, ...)
+{
+	MonoErrorInternal *error = (MonoErrorInternal*)oerror;
+	mono_error_prepare (error);
+
+	error->error_code = MONO_ERROR_ARGUMENT_NULL;
+	error->first_argument = argument;
 
 	set_error_message ();
 }
@@ -460,11 +522,14 @@ get_type_name_as_mono_string (MonoErrorInternal *error, MonoDomain *domain, Mono
 	if (error->type_name) {
 		res = mono_string_new (domain, error->type_name);
 		
-	} else if (error->klass) {
-		char *name = mono_type_full_name (&error->klass->byval_arg);
-		if (name) {
-			res = mono_string_new (domain, name);
-			g_free (name);
+	} else {
+		MonoClass *klass = get_class (error);
+		if (klass) {
+			char *name = mono_type_full_name (&klass->byval_arg);
+			if (name) {
+				res = mono_string_new (domain, name);
+				g_free (name);
+			}
 		}
 	}
 	if (!res)
@@ -475,7 +540,7 @@ get_type_name_as_mono_string (MonoErrorInternal *error, MonoDomain *domain, Mono
 static void
 set_message_on_exception (MonoException *exception, MonoErrorInternal *error, MonoError *error_out)
 {
-	MonoString *msg = mono_string_new (mono_domain_get (), mono_internal_error_get_message (error));
+	MonoString *msg = mono_string_new (mono_domain_get (), error->full_message);
 	if (msg)
 		MONO_OBJECT_SETREF (exception, message, msg);
 	else
@@ -499,7 +564,7 @@ mono_error_prepare_exception (MonoError *oerror, MonoError *error_out)
 		return NULL;
 
 	case MONO_ERROR_MISSING_METHOD:
-		if ((error->type_name || error->klass) && error->member_name) {
+		if ((error->type_name || error->exn.klass) && error->member_name) {
 			type_name = get_type_name_as_mono_string (error, domain, error_out);
 			if (!mono_error_ok (error_out))
 				break;
@@ -510,16 +575,16 @@ mono_error_prepare_exception (MonoError *oerror, MonoError *error_out)
 				break;
 			}
 
-			exception = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "MissingMethodException", type_name, method_name);
+			exception = mono_exception_from_name_two_strings_checked (mono_defaults.corlib, "System", "MissingMethodException", type_name, method_name, error_out);
 			if (exception)
 				set_message_on_exception (exception, error, error_out);
 		} else {
-		 	exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "MissingMethodException", mono_internal_error_get_message (error));
+			exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "MissingMethodException", error->full_message);
 		}
 		break;
 
 	case MONO_ERROR_MISSING_FIELD:
-		if ((error->type_name || error->klass) && error->member_name) {
+		if ((error->type_name || error->exn.klass) && error->member_name) {
 			type_name = get_type_name_as_mono_string (error, domain, error_out);
 			if (!mono_error_ok (error_out))
 				break;
@@ -530,16 +595,16 @@ mono_error_prepare_exception (MonoError *oerror, MonoError *error_out)
 				break;
 			}
 			
-			exception = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "MissingFieldException", type_name, field_name);
+			exception = mono_exception_from_name_two_strings_checked (mono_defaults.corlib, "System", "MissingFieldException", type_name, field_name, error_out);
 			if (exception)
 				set_message_on_exception (exception, error, error_out);
 		} else {
-		 	exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "MissingFieldException", mono_internal_error_get_message (error));
+			exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "MissingFieldException", error->full_message);
 		}
 		break;
 
 	case MONO_ERROR_TYPE_LOAD:
-		if (error->type_name || error->assembly_name) {
+		if ((error->type_name && error->assembly_name) || error->exn.klass) {
 			type_name = get_type_name_as_mono_string (error, domain, error_out);
 			if (!mono_error_ok (error_out))
 				break;
@@ -552,18 +617,18 @@ mono_error_prepare_exception (MonoError *oerror, MonoError *error_out)
 				}
 			}
 
-			exception = mono_exception_from_name_two_strings (mono_get_corlib (), "System", "TypeLoadException", type_name, assembly_name);
-			if (exception)
+			exception = mono_exception_from_name_two_strings_checked (mono_get_corlib (), "System", "TypeLoadException", type_name, assembly_name, error_out);
+			if (exception && error->full_message != NULL && strcmp (error->full_message, ""))
 				set_message_on_exception (exception, error, error_out);
 		} else {
-		 	exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "TypeLoadException", mono_internal_error_get_message (error));
+			exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "TypeLoadException", error->full_message);
 		}
 		break;
 
 	case MONO_ERROR_FILE_NOT_FOUND:
 	case MONO_ERROR_BAD_IMAGE:
 		if (error->assembly_name) {
-			msg = mono_string_new (domain, mono_internal_error_get_message (error));
+			msg = mono_string_new (domain, error->full_message);
 			if (!msg) {
 				mono_error_set_out_of_memory (error_out, "Could not allocate message");
 				break;
@@ -578,14 +643,14 @@ mono_error_prepare_exception (MonoError *oerror, MonoError *error_out)
 			}
 
 			if (error->error_code == MONO_ERROR_FILE_NOT_FOUND)
-				exception = mono_exception_from_name_two_strings (mono_get_corlib (), "System.IO", "FileNotFoundException", msg, assembly_name);
+				exception = mono_exception_from_name_two_strings_checked (mono_get_corlib (), "System.IO", "FileNotFoundException", msg, assembly_name, error_out);
 			else
-				exception = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "BadImageFormatException", msg, assembly_name);
+				exception = mono_exception_from_name_two_strings_checked (mono_defaults.corlib, "System", "BadImageFormatException", msg, assembly_name, error_out);
 		} else {
 			if (error->error_code == MONO_ERROR_FILE_NOT_FOUND)
-				exception = mono_exception_from_name_msg (mono_get_corlib (), "System.IO", "FileNotFoundException", mono_internal_error_get_message (error));
+				exception = mono_exception_from_name_msg (mono_get_corlib (), "System.IO", "FileNotFoundException", error->full_message);
 			else
-				exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "BadImageFormatException", mono_internal_error_get_message (error));
+				exception = mono_exception_from_name_msg (mono_defaults.corlib, "System", "BadImageFormatException", error->full_message);
 		}
 		break;
 
@@ -594,19 +659,23 @@ mono_error_prepare_exception (MonoError *oerror, MonoError *error_out)
 		break;
 
 	case MONO_ERROR_ARGUMENT:
-		exception = mono_get_exception_argument (error->type_name, mono_internal_error_get_message (error));
+		exception = mono_get_exception_argument (error->first_argument, error->full_message);
+		break;
+
+	case MONO_ERROR_ARGUMENT_NULL:
+		exception = mono_get_exception_argument_null (error->first_argument);
 		break;
 
 	case MONO_ERROR_NOT_VERIFIABLE: {
 		char *type_name = NULL, *message;
-		if (error->klass) {
-			type_name = mono_type_get_full_name (error->klass);
+		if (error->exn.klass) {
+			type_name = mono_type_get_full_name (error->exn.klass);
 			if (!type_name) {
 				mono_error_set_out_of_memory (error_out, "Could not allocate message");
 				break;
 			}
 		}
-		message = g_strdup_printf ("Error in %s:%s %s", type_name, error->member_name, mono_internal_error_get_message (error));
+		message = g_strdup_printf ("Error in %s:%s %s", type_name, error->member_name, error->full_message);
 		if (!message) {
 			g_free (type_name);
 			mono_error_set_out_of_memory (error_out, "Could not allocate message");
@@ -619,13 +688,28 @@ mono_error_prepare_exception (MonoError *oerror, MonoError *error_out)
 	}
 	case MONO_ERROR_GENERIC:
 		if (!error->exception_name_space || !error->exception_name)
-			mono_error_set_generic_error (error_out, "System", "ExecutionEngineException", "MonoError with generic error but no exception name was supplied");
+			mono_error_set_execution_engine (error_out, "MonoError with generic error but no exception name was supplied");
 		else
-			exception = mono_exception_from_name_msg (mono_defaults.corlib, error->exception_name_space, error->exception_name, mono_internal_error_get_message (error));
+			exception = mono_exception_from_name_msg (mono_defaults.corlib, error->exception_name_space, error->exception_name, error->full_message);
 		break;
 
+	case MONO_ERROR_EXCEPTION_INSTANCE:
+		exception = (MonoException*) mono_gchandle_get_target (error->exn.instance_handle);
+		break;
+
+	case MONO_ERROR_CLEANUP_CALLED_SENTINEL:
+		mono_error_set_execution_engine (error_out, "MonoError reused after mono_error_cleanup");
+		break;
+
+	case MONO_ERROR_INVALID_PROGRAM: {
+		gboolean lacks_message = error->flags & MONO_ERROR_INCOMPLETE;
+		if (lacks_message)
+			return mono_exception_from_name_msg (mono_defaults.corlib, "System", "InvalidProgramException", "");
+		else
+			return mono_exception_from_name_msg (mono_defaults.corlib, "System", "InvalidProgramException", error->full_message);
+	}
 	default:
-		mono_error_set_generic_error (error_out, "System", "ExecutionEngineException", "Invalid error-code %d", error->error_code);
+		mono_error_set_execution_engine (error_out, "Invalid error-code %d", error->error_code);
 	}
 
 	if (!mono_error_ok (error_out))
@@ -662,18 +746,9 @@ mono_error_convert_to_exception (MonoError *target_error)
 	return ex;
 }
 
-
-/*
-Raises the exception of @error.
-Does nothing if @error has a success error code.
-Aborts in case of a double fault. This happens when it can't recover from an error caused by trying
-to construct the first exception object.
-The error object @error is cleaned up. 
-*/
 void
-mono_error_raise_exception (MonoError *target_error)
+mono_error_move (MonoError *dest, MonoError *src)
 {
-	MonoException *ex = mono_error_convert_to_exception (target_error);
-	if (ex)
-		mono_raise_exception (ex);	
+	memcpy (dest, src, sizeof (MonoErrorInternal));
+	mono_error_init (src);
 }
